/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: hci_mgr.c,v 1.17 2002/08/28 12:29:12 kds Exp $

   HCI Manager

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
                Imre Deak <ext-imre.deak@nokia.com>
*/		

/* The following prevents "kernel_version" from being set in this file. */
#define __NO_VERSION__

#include <linux/config.h>

/* Module related headers, non-module drivers should not include */
#include <linux/module.h>

/* Standard driver includes */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <linux/proc_fs.h>
#include <linux/notifier.h>
#include <linux/poll.h>
#include <linux/file.h>
#include <linux/utsname.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 7)
#include <linux/completion.h>
#endif

#include <asm/uaccess.h>
#include <asm/bitops.h>

#include <linux/in.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/if_arp.h>
#include <linux/rtnetlink.h>
#include <net/pkt_sched.h>


/* Local Includes */
#define FILEBIT	DBHCIMGR

#include <affix/types.h>
#include <affix/btlist.h>
#include <affix/btdebug.h>
#include <affix/hci.h>
#include <affix/bluetooth.h>
#include <affix/btcore.h>
#include <affix/hci_mgr.h>


#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>

DECLARE_MUTEX_LOCKED(exit_sema);
static volatile int	pid = -1, mgr_exit = 0;
btlist_head_t		btdevs;	/* Neighbor Devices list */
struct PIN_Code		affix_pin_code = { {{0, 0, 0, 0, 0, 0}}, 0, ""};
int			affix_ctl_mode	= 0;


hci_con *__hcc_lookup_hdch(int hd, CHANDLE chandle)
{
	hci_con	*con;
	
	btl_for_each (con, hcicons) {
		if (STATE(con) != DEAD)
			if ((con->hci && con->hci->hd == hd) && con->chandle == chandle)
			break;
	}
	return con;
}

hci_con *hcc_lookup_hdch(int hd, CHANDLE chandle)
{
	hci_con		*con;

	DBFENTER;

	btl_read_lock(&hcicons);
	con = __hcc_lookup_hdch(hd, chandle);
	if (con)
		hcc_hold(con);
	btl_read_unlock(&hcicons);

	DBFEXIT;
	return con;
}

hci_con *__hcc_lookup_hdbda(int hd, BD_ADDR *bda)
{
	hci_con	*con;

	btl_for_each (con, hcicons) {
		if (STATE(con) != DEAD && STATE(con) != CON_W4_DISCRSP)
			if ((con->hci && con->hci->hd == hd) && memcmp(&con->bda, bda, 6) == 0)
			break;
	}
	return con;
}

hci_con *hcc_lookup_hdbda(int hd, BD_ADDR *bda)
{
	hci_con		*con;

	DBFENTER;

	btl_read_lock(&hcicons);
	con = __hcc_lookup_hdbda(hd, bda);
	if (con)
		hcc_hold(con);
	btl_read_unlock(&hcicons);

	DBFEXIT;
	return con;
}


/*
  Driver Internal Commands
*/

int message_handler(struct hci_msg_hdr *hdr)
{
	int	err = 0, fd;

	DBFENTER;

	DBPRT(DBINIT, "Manager Command: %d\n", hdr->opcode);
	switch (hdr->opcode) {
 	case MGR_INIT_HCI: {
		__u32			cod;
		struct hci_init_req	*cmd = (void*)hdr;
		hci_struct		*hci = hci_lookup(cmd->hd);

		__u8	hci_version, lmp_version;
		__u16	hci_revision, manfid, lmp_subversion;

		if (hci == NULL) // device already gone
			return -ENODEV;

 		rtnl_lock();
		err = dev_open(hci->dev);
		rtnl_unlock();
		if (err) {
			hci_put(hci);
			return -EBUSY;
		}
		fd = __hci_open_id(hci->hd);
		if (fd < 0) {
			hci_put(hci);
			return -ENODEV;
		}

		err = HCI_ReadBDAddr(fd, &hci->bda);
		if (err)
			goto initerr;
		bdacpy(hci->dev->dev_addr, &hci->bda);

		HCI_ReadBufferSize(fd, &hci->acl_plen, &hci->sco_plen,
				   &hci->def_acl_num, &hci->def_sco_num);
		hci->acl_num = hci->def_acl_num;
		hci->sco_num = hci->def_sco_num;

		if (hci->acl_num) hci->flowmask |= HCI_ACL_ON;
		if (hci->sco_num) hci->flowmask |= HCI_SCO_ON;

		/* some general settings */
		HCI_WriteScanEnable(fd, 0x03);
		cod = COD_INFORMATION | COD_NETWORKING | COD_COMPUTER | COD_HANDPC;
		HCI_WriteClassOfDevice(fd, cod);
		down_read(&uts_sem);
		HCI_ChangeLocalName(fd, system_utsname.nodename);
		up_read(&uts_sem);

		HCI_ReadLocalVersionInformation(fd, &hci_version, &hci_revision, &lmp_version,
						&manfid, &lmp_subversion);
		hci->manfid = manfid;
#if 1
		/*  Removed when UART RI flow control was introduced */
		if (manfid == LMP_NOKIA)
			hci->select_packet = select_packet_ncc;
		else
			hci->select_packet = select_packet;
#else
		hci->select_packet = select_packet;
#endif
#ifdef DTL1_PKT_SCHED_FIX
		if (manfid == LMP_NOKIA)
			hci->fixit = 1;
		else
			hci->fixit = 0;
#endif
		hci_start(hci);
initerr:
		hci_close(fd);
		hci_put(hci);
	}
	break;

 	case MGR_CONNECT_REQ: {
		struct hci_connect_req	*cmd = (void*)hdr;
		INQUIRY_ITEM	dev;
		hci_con		*con;

		con = hcc_lookup_id(cmd->id);
		if (!con)
			break;

		if (con->hci == NULL) {
			hci_struct	*hci;

			/* select device */
			hci = hci_select();
			if (hci == NULL) {
				hcc_put(con);
				break;
			}
			hcc_bind(con, hci);
			hci_put(hci);
		}

		if (STATE(con) != CON_W4_LCONREQ) {
			hcc_put(con);
			break;
		}
		ENTERSTATE(con, CON_W4_CONRSP);

		if (con->link_type == LT_ACL) {
			fd = hci_open_id(con->hci->hd);
			if (fd < 0) {
				hcc_put(con);
				return -ENODEV;
			}

			dev.bda = con->bda;
			dev.PS_Repetition_Mode = 0x00;
			dev.PS_Mode = 0x00;
			dev.Clock_Offset = 0x00;
			con->hci->target_dev = dev;
			err = __HCI_CreateConnection(fd, &dev, con->hci->pkt_type & PT_ACL, con->hci->allow_switch);
			hci_close(fd);
		} else {
			hci_con		*acl;

			acl = hcc_lookup_acl(con->hci, &con->bda);
			if (!acl) {
				/* not connected FIXME:*/
				hcc_put(con);
				break;
			}
			
			fd = hci_open_id(acl->hci->hd);
			if (fd < 0) {
				hcc_put(con);
				return -ENODEV;
			}
			err = __HCI_AddSCOConnection(fd, acl->chandle, acl->hci->pkt_type & PT_SCO);
			hci_close(fd);
			hcc_put(acl);
		}

		hcc_put(con);
	}
	break;

 	case MGR_DISCONNECT_REQ: {
		struct hci_disconnect_req	*cmd = (void*)hdr;
		hci_con		*con;

		con = hcc_lookup_id(cmd->id);
		if (!con)
			break;

		if (STATE(con) != CON_W4_LDISCREQ) {
			hcc_put(con);
			break;
		}
		ENTERSTATE(con, CON_W4_DISCRSP);

		fd = hci_open_id(con->hci->hd);
		if (fd < 0) {
			hcc_put(con);
			return -ENODEV;

		}

		err = __HCI_Disconnect(fd, con->chandle, cmd->reason);
		hci_close(fd);
		hcc_put(con);
	}
	break;

 	case MGR_AUTH_REQ: 
	{
		struct hci_auth_req	*cmd = (void*)hdr;
		hci_con		*con;

		con = hcc_lookup_id(cmd->id);
		if (!con)
			break;

		fd = hci_open_id(con->hci->hd);
		if (fd < 0) {
			hcc_put(con);
			return -ENODEV;

		}
		err = __HCI_AuthenticationRequested(fd, con->chandle);
		hci_close(fd);
		hcc_put(con);
	}
	break;

	default:
	}

	return err;
}



/*
 * HCI events processing
 */
int connection_request(struct hci_event_arg *ep)
{
	struct Connection_Request_Event	*evt = (void*)ep->event;
	int		err = 0, fd;
	hci_struct	*hci;
	

	DBFENTER;

	hci = hci_lookup(ep->hd);
	if (!hci) {
		return -ENODEV;
	}
	
	fd = hci_open_id(ep->hd);
	if (fd < 0) {
		hci_put(hci);
		return fd;
	}
	
	err = __HCI_AcceptConnectionRequest(fd, &evt->bda, hci->role);

	hci_close(fd);
	hci_put(hci);

	DBFEXIT;
	return err;
}

int pin_code_request(struct hci_event_arg *ep)
{
	struct PIN_Code_Request_Event	*evt = (void*)ep->event;
	int		fd, err, ok = 0;
	btdev_struct	*btdev;

	DBFENTER;

	if (affix_ctl_mode & AFFIX_MODE_PIN)
		return 0;	// somebody takes care about pin

	fd = hci_open_id(ep->hd);
	if (fd < 0)
		return fd;

	btdev = btdev_lookup_bda(&evt->bda);
	if (btdev && (btdev->flags & NBT_PIN))
		ok = 1;

	if (ok) {
		err = HCI_PINCodeRequestReply(fd, &btdev->bda, btdev->PIN_Code_Length, btdev->PIN_Code);
	} else {
		if (affix_pin_code.Length == 0)
			err = HCI_PINCodeRequestNegativeReply(fd, &evt->bda);
		else
			err = HCI_PINCodeRequestReply(fd, &evt->bda, affix_pin_code.Length, affix_pin_code.Code);
	}
	if (btdev)
		btdev_put(btdev);
	hci_close(fd);

	DBFEXIT;
	return err;
}

int link_key_request(struct hci_event_arg *ep)
{
	struct Link_Key_Request_Event	*evt = (void*)ep->event;
	int		fd, err, ok = 0;
	btdev_struct	*btdev;

	DBFENTER;

	fd = hci_open_id(ep->hd);
	if (fd < 0)
		return fd;

	btdev = btdev_lookup_bda(&evt->bda);
	if (btdev && (btdev->flags & NBT_KEY))
		ok = 1;

	if (ok) {
		err = HCI_LinkKeyRequestReply(fd, &evt->bda, btdev->Link_Key);
	} else {
		err = HCI_LinkKeyRequestNegativeReply(fd, &evt->bda);
	}
	if (btdev)
		btdev_put(btdev);
	hci_close(fd);

	DBFEXIT;
	return err;
}

int event_handler(struct hci_event_arg *ep)
{
	int	err = 0;

	DBFENTER;

	DBPRT(DBINIT, "Manager has an event: 0x%02x\n", ep->event->EventCode);

	switch (ep->event->EventCode) {
		case HCI_E_CONNECTION_REQUEST:
			err = connection_request(ep);
			break;

		case HCI_E_PIN_CODE_REQUEST:
			pin_code_request(ep);
			break;

		case HCI_E_LINK_KEY_REQUEST:
			link_key_request(ep);
			break;

		default:
	}

	DBFEXIT;
	return err;
}

/*  This is like the standard daemonize call, but without destroying
 *  fs and files object of current.
 */
static void daemonize_keep_fs(void)
{
	int fd;
	
	exit_mm(current);
	current->session = 1;
	current->pgrp = 1;
        for (fd = 0; fd < 255; fd++)
		close(fd);
}

#ifdef CONFIG_AFFIX_UART
extern void init_btuart(void);
#endif
#ifdef CONFIG_AFFIX_UART_CS
extern void init_btuart_cs(void);
#endif
#ifdef CONFIG_AFFIX_USB
extern void btusb_init(void);
#endif

int affixd_thread(void *startup)
{
	int			ret = 0;
	struct task_struct	*curtask = current;
	struct pollfd		fds[2];
	__u8			msg[MSG_MAX_SIZE];
	__u8			event[EVENT_MAX_SIZE];
	struct hci_event_arg	ep;

	DBFENTER;

	lock_kernel();
	//daemonize();
	daemonize_keep_fs();
	strcpy(current->comm, "affixd");
	unlock_kernel();

#if 1
	spin_lock_irq(&curtask->sigmask_lock);
	sigfillset(&curtask->blocked);
	siginitsetinv(&current->blocked, sigmask(SIGTERM));
	recalc_sigpending(curtask);
	spin_unlock_irq(&curtask->sigmask_lock);
#endif
	
	/*  Open HCI manager */
	fds[0].fd = __hci_open_mgr();
	if (fds[0].fd < 0) {
		ret = fds[0].fd;
		fds[0].fd = fds[1].fd = 0;
		goto exit;
	}
	fds[0].events = POLLIN;
	
	fds[1].fd = __hci_open(NULL);
	if (fds[1].fd < 0) {
		ret = fds[1].fd;
		fds[1].fd = 0;
		goto exit;
	}
	fds[1].events = POLLIN;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 7)
	up(&exit_sema);
#else
	complete((struct completion *)startup);
#endif
	
	/* we can initialize drivers when we listen for Control Messages */
#ifdef CONFIG_AFFIX_UART
	init_btuart();
#endif
#ifdef CONFIG_AFFIX_UART_CS
	init_btuart_cs();
#endif
#ifdef CONFIG_AFFIX_USB
	btusb_init();
#endif
	ep.event = (void*)event;	/* set event buffer */

	hci_listen_event(fds[1].fd, ALL_EVENTS_MASK &
			~(COMMAND_COMPLETE_MASK | COMMAND_STATUS_MASK |
			  NUMBER_OF_COMPLETE_PACKETS_MASK | FLUSH_OCCURRED_MASK));

	for (;;) {
		ret = hci_poll(fds, 2, -1);
		if (ret > 0) {/* we have something */
			if ((fds[0].revents | fds[1].revents) & (POLLERR | POLLHUP | POLLNVAL))
				break;
			if (fds[1].revents) {
				if (fds[1].revents & POLLIN) { /* HCI event */
					hci_recv_event(fds[1].fd, &ep);
					event_handler(&ep);
				}
			}
			if (fds[0].revents) {
				if (fds[0].revents & POLLIN) {/* MSG */
					hci_recv_msg(fds[0].fd, msg);
					message_handler((void*)msg);
				}
			}
		}

		if (signal_pending(curtask)) {
			DBPRT(DBINIT, "Got a signal!!!\n");
			spin_lock_irq(&curtask->sigmask_lock);
			flush_signals(curtask);	// flush pending signals
			recalc_sigpending(curtask);
			spin_unlock_irq(&curtask->sigmask_lock);
			if (mgr_exit)
				break;
#if 0
			spin_lock_irq(&curtask->sigmask_lock);
			if (sigismember(&curtask->pending.signal, SIGTERM)) {
				sigdelset(&curtask->pending.signal, SIGTERM);
				// do simething
			}
			recalc_sigpending(curtask);
			spin_unlock_irq(&curtask->sigmask_lock);
#endif			
		}
	}

exit:
	if (fds[0].fd)
		hci_close(fds[0].fd);
	if (fds[1].fd)
		hci_close(fds[1].fd);
	
	up(&exit_sema);
	pid = -1;

	DBFEXIT;
	return ret;
}

int hci_start_manager(void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 7)
	static struct completion startup __initdata = COMPLETION_INITIALIZER(startup);
#endif

	DBFENTER;

	btl_head_init(&btdevs);
	mgr_exit = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 7)
	/* start thread here */
	pid = kernel_thread(affixd_thread, NULL, 0);
#else
	/* start thread here */
	pid = kernel_thread(affixd_thread, &startup, 0);
#endif
	
	if (pid < 0) {
		BTERRORF("Unable to create BTD thread\n");
		return -ENOMEM;
	}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 7)
	down(&exit_sema);
#else
	/* let's wait for completion */
	wait_for_completion(&startup);
#endif

	DBFEXIT;
	return 0;
}

void hci_stop_manager(void)
{
	DBFENTER;

	if (pid != -1) {
		mgr_exit = 1;
		kill_proc(pid, SIGTERM, 1);
		down(&exit_sema);
	}
	btdev_flush();

	DBFEXIT;
}


/*
  Messages (commands) for the HCI Manager
*/


int hci_init_req(hci_struct *hci)
{
	struct sk_buff		*skb;
	struct hci_init_req	*msg;

	DBFENTER;

	skb = alloc_skb(sizeof(*msg), GFP_ATOMIC);
	if (skb == NULL)
		return -ENOMEM;
	skb_put(skb, sizeof(*msg));

	msg = (void*)skb->data;
	msg->hdr.opcode = MGR_INIT_HCI;
	msg->hd = hci->hd;

	hci_msg_ready(skb);

	DBFEXIT;
	return 0;
}

int hci_connect_req(hci_con *con)
{
	struct sk_buff		*skb;
	struct hci_connect_req	*msg;

	DBFENTER;

	skb = alloc_skb(sizeof(*msg), GFP_ATOMIC);
	if (skb == NULL)
		return -ENOMEM;
	skb_put(skb, sizeof(*msg));

	msg = (void*)skb->data;
	msg->hdr.opcode = MGR_CONNECT_REQ;
	msg->id = con->id;

	hci_msg_ready(skb);

	DBFEXIT;
	return 0;
}

int hci_auth_req(hci_con *con)
{
	struct sk_buff		*skb;
	struct hci_auth_req	*msg;

	DBFENTER;

	skb = alloc_skb(sizeof(*msg), GFP_ATOMIC);
	if (skb == NULL)
		return -ENOMEM;
	skb_put(skb, sizeof(*msg));

	msg = (void*)skb->data;
	msg->hdr.opcode = MGR_AUTH_REQ;
	msg->id = con->id;

	hci_msg_ready(skb);

	DBFEXIT;
	return 0;
}


int hci_disconnect_req(hci_con *con, __u8 reason)
{
	struct sk_buff			*skb;
	struct hci_disconnect_req	*msg;

	DBFENTER;

	skb = alloc_skb(sizeof(*msg), GFP_ATOMIC);
	if (skb == NULL)
		return -ENOMEM;
	skb_put(skb, sizeof(*msg));

	msg = (void*)skb->data;
	msg->hdr.opcode = MGR_DISCONNECT_REQ;
	msg->id = con->id;
	msg->reason = reason;

	hci_msg_ready(skb);

	DBFEXIT;
	return 0;
}


/* Neighbor Device management */
btdev_struct *btdev_alloc(void)
{
	btdev_struct	*btdev;

	btdev = kmalloc(sizeof(*btdev), GFP_ATOMIC);
	if (btdev == NULL)
		return NULL;
	memset(btdev, 0, sizeof(*btdev));

	return btdev;
}

void btdev_free(btdev_struct *btdev)
{
	kfree(btdev);
}

btdev_struct *__btdev_lookup_bda(BD_ADDR *bda)
{
	btdev_struct	*btdev;

	btl_for_each (btdev, btdevs) {
		if (memcmp(&btdev->bda, bda, 6) == 0)
			break;
	}
	return btdev;
}

btdev_struct *btdev_lookup_bda(BD_ADDR *bda)
{
	btdev_struct	*btdev;

	DBFENTER;

	btl_read_lock(&btdevs);
	btdev = __btdev_lookup_bda(bda);
	if (btdev)
		btdev_hold(btdev);
	btl_read_unlock(&btdevs);

	DBFEXIT;
	return btdev;
}

btdev_struct *btdev_create(BD_ADDR *bda)
{
	btdev_struct	*btdev;

	DBFENTER;

	btl_write_lock(&btdevs);
	btdev = __btdev_lookup_bda(bda);
	if (btdev == NULL) {
		btdev = btdev_alloc();
		if (btdev) {
			memcpy(&btdev->bda, bda, 6);
			atomic_set(&btdev->refcnt, 1);
			__btl_add_tail(&btdevs, btdev);
		}
	}
	if (btdev)
		btdev_hold(btdev);
	btl_write_unlock(&btdevs);

	DBFEXIT;
	return btdev;
}

void __btdev_destroy(btdev_struct *btdev)
{
	if (atomic_read(&btdev->refcnt) == 0) {
		__btl_unlink(&btdevs, btdev);
		btdev_free(btdev);
	}
}

void btdev_destroy(btdev_struct *btdev)
{
	DBFENTER;

	btl_write_lock(&btdevs);
	__btdev_destroy(btdev);
	btl_write_unlock(&btdevs);

	DBFEXIT;
}

/*
  flush all entries in the device list
*/
void btdev_flush(void)
{
	btdev_struct	*btdev, *next;

	DBFENTER;

	btl_write_lock(&btdevs);
	btl_for_each_next (btdev, btdevs, next) {
		__btdev_put(btdev);
	}
	btl_write_unlock(&btdevs);

	DBFEXIT;
}


/* 
 * PIN Code management
 */

int affix_add_pin(struct PIN_Code *pin)
{
	btdev_struct	*btdev;

	DBFENTER;

	if (memcmp(&pin->bda, &affix_pin_code.bda, 6) == 0) {
		/* default */
		affix_pin_code = *pin;
	} else {
		btdev = btdev_create(&pin->bda);
		if (btdev == NULL)
			return -ENOMEM;

		btdev->PIN_Code_Length = pin->Length;
		memcpy(btdev->PIN_Code, pin->Code, 16);
		btdev->flags |= NBT_PIN;
	}
	DBFEXIT;
	return 0;
}

int affix_remove_pin(BD_ADDR *pbda)
{
	int		err = 0;
	btdev_struct	*btdev;
	BD_ADDR		bda;

	DBFENTER;

	btl_read_lock(&btdevs);
	if (pbda == NULL) {
		// remove all
		btl_for_each (btdev, btdevs)
			btdev->flags &= ~NBT_PIN;
	} else {
		err = copy_from_user(&bda, pbda, sizeof(BD_ADDR));
		if (err)
			goto exit;
		if (bda_zero(&bda))
			affix_pin_code.Length = 0;
		else {
			btdev = __btdev_lookup_bda(&bda);
			if (btdev == NULL) {
				err = -ENOENT;
				goto exit;
			}
			btdev->flags &= ~NBT_PIN;
		}
	}
exit:
	btl_read_unlock(&btdevs);
	DBFEXIT;
	return err;
}

int affix_remove_key(BD_ADDR *pbda)
{
	int		err = 0;
	btdev_struct	*btdev;
	BD_ADDR		bda;

	DBFENTER;

	btl_read_lock(&btdevs);
	if (pbda == NULL) {
		// remove all
		btl_for_each (btdev, btdevs)
			btdev->flags &= ~NBT_KEY;
	} else {
		err = copy_from_user(&bda, pbda, sizeof(BD_ADDR));
		if (err)
			goto exit;
		btdev = __btdev_lookup_bda(&bda);
		if (btdev == NULL) {
			err = -ENOENT;
			goto exit;
		}
		btdev->flags &= ~NBT_KEY;
	}
exit:
	btl_read_unlock(&btdevs);
	DBFEXIT;
	return err;
}

