/* 
   Affix - Bluetooth Protocol Stack for Linux

   The contents of this file are subject to the Mozilla Public
   License Version 1.1 (the "License"); you may not use this file
   except in compliance with the License. You may obtain a copy of
   the License at http://www.mozilla.org/MPL/

   Software distributed under the License is distributed on an "AS
   IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
   implied. See the License for the specific language governing
   rights and limitations under the License.

   The initial developer of the original code is David A. Hinds
   <dahinds@users.sourceforge.net>.  Portions created by David A. Hinds
   are Copyright (C) 1999 David A. Hinds.  All Rights Reserved.

   Alternatively, the contents of this file may be used under the
   terms of the GNU General Public License version 2 (the "GPL"), in which
   case the provisions of the GPL are applicable instead of the
   above.  If you wish to allow the use of your version of this file
   only under the terms of the GPL and not to allow others to use
   your version of this file under the MPL, indicate your decision
   by deleting the provisions above and replace them with the notice
   and other provisions required by the GPL.  If you do not delete
   the provisions above, a recipient may use your version of this
   file under either the MPL or the GPL.
*/

/* 
   $Id: btuart_cs.c,v 1.1.1.1 2002/07/11 12:19:05 kds Exp $

   btuart_cs - uart driver for pcmcia serial bluetooth cards
               based on serial_cs.c

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
                Imre Deak <ext-imre.deak@nokia.com>
*/

#include <linux/config.h>

#include <linux/module.h>

/* Standard driver includes */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/timer.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/major.h>
#include <linux/netdevice.h>
#include <asm/io.h>
#include <asm/system.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ciscode.h>
#include <pcmcia/ds.h>
#include <pcmcia/cisreg.h>

#define FILEBIT	DBUART
#include <affix/btdebug.h>

#if defined(CONFIG_AFFIX)
#include "btuart.h"
#else
#include "../uart/btuart.h"
#endif


/*====================================================================*/

/* Parameters that can be set with 'insmod' */

/* Bit map of interrupts to choose from */
//static u_int irq_mask = 0xdeb8;
static u_int irq_mask = 0xffff;
static int irq_list[4] = { -1 };

/* Enable the speaker? */
static int do_sound = 1;

MODULE_PARM(irq_mask, "i");
MODULE_PARM(irq_list, "1-4i");
//MODULE_PARM(do_sound, "i");

#define BTUARTCS_DEVNAME	"affix_uart_cs"


/* Table of multi-port card ID's */

typedef struct {
    u_short	manfid;
    u_short	prodid;
    int		multi;		/* 1 = multifunction, > 1 = # ports */
} multi_id_t;

static multi_id_t multi_id[] = {
    { MANFID_NOKIA, PRODID_DTL1, 1 },
    { MANFID_NOKIA, PRODID_DTL1_PROTO, 1 }
};
#define MULTI_COUNT (sizeof(multi_id)/sizeof(multi_id_t))

/* control structure for transport driver */
typedef struct btuartcs_device {
	int			magic;

	/* pcmcia/serial stuff */
	dev_link_t		link;
	int			ndev;
	int			multi;
	int			slave;
	int			manfid;
	int			prodid;
	dev_node_t		node[4];
	int			line[4];
	affix_uart_t		hw[4];
} btuartcs_device;



void btuart_config(dev_link_t *link);
void btuart_release(u_long arg);
int btuart_event(event_t event, int priority, event_callback_args_t *args);

static dev_info_t dev_info = BTUARTCS_DEVNAME;

dev_link_t *btuart_attach(void);
void btuart_detach(dev_link_t *);

dev_link_t	*dev_list = NULL;

/*====================================================================*/

static void cs_error(client_handle_t handle, int func, int ret)
{
    error_info_t err = { func, ret };
    CardServices(ReportError, handle, &err);
}

/*======================================================================

    btuart_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

======================================================================*/

dev_link_t *btuart_attach(void)
{
	btuartcs_device	*btuart;
	client_reg_t 	client_reg;
	dev_link_t 	*link;
	int 		i, ret;
    
	DBFENTER;

	/* Create new btuart device */
	btuart = kmalloc(sizeof(*btuart), GFP_KERNEL);
	if (!btuart) return NULL;
	memset(btuart, 0, sizeof(*btuart));

	/* initialize some attribues */
	btuart->magic = BTUARTLD_MAGIC;

	link = &btuart->link; link->priv = btuart;

	link->release.function = &btuart_release;
	link->release.data = (u_long)link;
	link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
	link->io.NumPorts1 = 8;
	link->irq.Attributes = IRQ_TYPE_EXCLUSIVE;
	link->irq.IRQInfo1 = IRQ_INFO2_VALID|IRQ_LEVEL_ID;
	if (irq_list[0] == -1)
		link->irq.IRQInfo2 = irq_mask;
	else
		for (i = 0; i < 4; i++)
			link->irq.IRQInfo2 |= 1 << irq_list[i];

	link->conf.Attributes = CONF_ENABLE_IRQ;
	link->conf.Vcc = 50;
	if (do_sound) {
		link->conf.Attributes |= CONF_ENABLE_SPKR;
		link->conf.Status = CCSR_AUDIO_ENA;
	}
	link->conf.IntType = INT_MEMORY_AND_IO;
    
	/* Register with Card Services */
	link->next = dev_list;
	dev_list = link;
	client_reg.dev_info = &dev_info;
	client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
	client_reg.EventMask =
		CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
		CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
		CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
	client_reg.event_handler = &btuart_event;
	client_reg.Version = 0x0210;
	client_reg.event_callback_args.client_data = link;
	
	ret = CardServices(RegisterClient, &link->handle, &client_reg);
	if (ret != CS_SUCCESS) {
		cs_error(link->handle, RegisterClient, ret);
		btuart_detach(link);
		return NULL;
	}
    
	DBFEXIT;

	return link;
}

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/

void btuart_detach(dev_link_t *link)
{
	btuartcs_device	*btuart = link->priv;
	dev_link_t	**linkp;
	long 		flags;
	int 		ret, i;

	DBFENTER;
    
	DBPRT(DBINIT, "irq?: %d, int?: %d\n", in_irq(), in_interrupt());

	/* Locate device structure */
	for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)
		if (*linkp == link)
			break;

	if (*linkp == NULL)
		return;

	save_flags(flags); cli();
	if (link->state & DEV_RELEASE_PENDING) {
		del_timer(&link->release);
		link->state &= ~DEV_RELEASE_PENDING;
	}
	restore_flags(flags);
	
	if (link->state & DEV_CONFIG) {
		btuart_release((u_long)link);
	}

	if (link->handle) {
		ret = CardServices(DeregisterClient, link->handle);
		if (ret != CS_SUCCESS)
			cs_error(link->handle, DeregisterClient, ret);
	}
    
	/* Unlink device structure, free bits */
	*linkp = link->next;

	for (i = 0; i < btuart->ndev;i++) 
		affix_uart_detach(btuart->hw[i].kdev);

	kfree(btuart);
	DBFEXIT;
}

/*====================================================================*/

int setup_btuart(btuartcs_device *btuart, ioaddr_t port, int irq)
{
	struct serial_struct	serial;
	int 			line;

	DBFENTER;

	memset(&serial, 0, sizeof(serial));
	serial.port = port;
	serial.irq = irq;
	serial.flags = ASYNC_SKIP_TEST;
	serial.flags |= (btuart->multi || btuart->slave) ? ASYNC_SHARE_IRQ : 0;
	line = register_serial(&serial);
	if (line < 0) {
		BTNOTICEF("affix_uart_cs: register_serial() at 0x%04x, "
				"irq %d failed\n", serial.port, serial.irq);
		return -1;
	}
	btuart->line[btuart->ndev] = line;
	sprintf(btuart->node[btuart->ndev].dev_name, "ttyS%d", line);
	btuart->node[btuart->ndev].major = TTY_MAJOR;
	btuart->node[btuart->ndev].minor = 0x40+line;
	if (btuart->ndev > 0)
		btuart->node[btuart->ndev-1].next = &btuart->node[btuart->ndev];

	// setup serial structure
	btuart->hw[btuart->ndev].port = port;
	btuart->hw[btuart->ndev].irq = irq;
	btuart->hw[btuart->ndev].kdev = mk_kdev(TTY_MAJOR, 0x40+line);
	btuart->hw[btuart->ndev].count = &btuart->link.open;
	btuart->hw[btuart->ndev].manfid = btuart->manfid;
	btuart->hw[btuart->ndev].prodid = btuart->prodid;
	btuart->hw[btuart->ndev].type = -1;
	btuart->hw[btuart->ndev].proto = 0;
	btuart->hw[btuart->ndev].flags = 0;
	btuart->hw[btuart->ndev].speed = 0;

	SET_MODULE_OWNER(&btuart->hw[btuart->ndev]);

	btuart->ndev++;

	return 0;
}

/*====================================================================*/

static int get_tuple(int fn, client_handle_t handle, tuple_t *tuple,
		     cisparse_t *parse)
{
	int i;
	i = CardServices(fn, handle, tuple);
	if (i != CS_SUCCESS) return CS_NO_MORE_ITEMS;
	i = CardServices(GetTupleData, handle, tuple);
	if (i != CS_SUCCESS) return i;
	return CardServices(ParseTuple, handle, tuple, parse);
}

#define first_tuple(a, b, c) get_tuple(GetFirstTuple, a, b, c)
#define next_tuple(a, b, c) get_tuple(GetNextTuple, a, b, c)

/*====================================================================*/

int simple_config(dev_link_t *link)
{
	static ioaddr_t base[5] = { 0x3f8, 0x2f8, 0x3e8, 0x2e8, 0x0 };
	client_handle_t handle = link->handle;
	btuartcs_device *btuart = link->priv;
	tuple_t tuple;
	u_char buf[256];
	cisparse_t parse;
	cistpl_cftable_entry_t *cf = &parse.cftable_entry;
	config_info_t config;
	int i, j, base2 = 0;

	DBFENTER;

	/* If the card is already configured, look up the port and irq */
	i = CardServices(GetConfigurationInfo, handle, &config);
	if ((i == CS_SUCCESS) &&
	    (config.Attributes & CONF_VALID_CLIENT)) {
		ioaddr_t port = 0;
		if ((config.BasePort2 != 0) && (config.NumPorts2 == 8)) {
			port = config.BasePort2;
			btuart->slave = 1;
		} else if ((btuart->manfid == MANFID_OSITECH) &&
			   (config.NumPorts1 == 0x40)) {
			port = config.BasePort1 + 0x28;
			btuart->slave = 1;
		}
		if (btuart->slave)
			return setup_btuart(btuart, port, config.AssignedIRQ);
	}
	link->conf.Vcc = config.Vcc;
    
	/* First pass: look for a config entry that looks normal. */
	tuple.TupleData = (cisdata_t *)buf;
	tuple.TupleOffset = 0; tuple.TupleDataMax = 255;
	tuple.Attributes = 0;
	tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
	i = first_tuple(handle, &tuple, &parse);
	while (i != CS_NO_MORE_ITEMS) {
		if (i != CS_SUCCESS) goto next_entry;
		//DBPRT(DBINIT, "1, index: %d, nwin: %d, len: %d\n", cf->index, cf->io.nwin, cf->io.win[0].len);
		if (cf->vpp1.present & (1<<CISTPL_POWER_VNOM))
			link->conf.Vpp1 = link->conf.Vpp2 =
				cf->vpp1.param[CISTPL_POWER_VNOM]/10000;
		if ((cf->io.nwin > 0) && (cf->io.win[0].len == 8) ) {
		  //if ((cf->io.nwin > 0) && (cf->io.win[0].len == 8) && (cf->io.win[0].base != 0)) {
			link->conf.ConfigIndex = cf->index;
			link->io.BasePort1 = cf->io.win[0].base;
			link->io.NumPorts1 = cf->io.win[0].len; /*yo*/
			link->io.IOAddrLines = cf->io.flags & CISTPL_IO_LINES_MASK;
			i = CardServices(RequestIO, link->handle, &link->io);
			
			//link->irq.IRQInfo1 = IRQ_LEVEL_ID;
			//link->irq.IRQInfo1 = 11;

			if (i == CS_SUCCESS) goto found_port;
		}
	next_entry:
		i = next_tuple(handle, &tuple, &parse);
	}
    
	/* Second pass: try to find an entry that isn't picky about
	   its base address, then try to grab any standard serial port
	   address, and finally try to get any free port. */
	i = first_tuple(handle, &tuple, &parse);
	while (i != CS_NO_MORE_ITEMS) {
		if ((i == CS_SUCCESS) && (cf->io.nwin > 0) &&
		    ((cf->io.flags & CISTPL_IO_LINES_MASK) <= 3)) {
			link->conf.ConfigIndex = cf->index;
			for (j = 0; j < 5; j++) {
				link->io.BasePort1 = base[j];
				link->io.IOAddrLines = base[j] ? 16 : 3;
				i = CardServices(RequestIO, link->handle, &link->io);
				if (i == CS_SUCCESS) goto found_port;
			}
		}
		i = next_tuple(handle, &tuple, &parse);
	}

	/* First, look for a generic full-sized window */
	link->io.NumPorts1 = btuart->multi * 8;
	i = first_tuple(handle, &tuple, &parse);
	while (i != CS_NO_MORE_ITEMS) {
		/* The quad port cards have bad CIS's, so just look for a
		   window larger than 8 ports and assume it will be right */
		if ((i == CS_SUCCESS) && (cf->io.nwin == 1) && (cf->io.win[0].len > 8)) {
			link->conf.ConfigIndex = cf->index;
			link->io.BasePort1 = cf->io.win[0].base;
			link->io.NumPorts1 = cf->io.win[0].len; /*yo*/
			link->io.IOAddrLines = cf->io.flags & CISTPL_IO_LINES_MASK;
			i = CardServices(RequestIO, link->handle, &link->io);
			base2 = link->io.BasePort1 + 8;
			if (i == CS_SUCCESS) break;
		}
		i = next_tuple(handle, &tuple, &parse);
	}

	/* If that didn't work, look for two windows */
	if (i != CS_SUCCESS) {
		link->io.NumPorts1 = link->io.NumPorts2 = 8;
		btuart->multi = 2;
		i = first_tuple(handle, &tuple, &parse);
		while (i != CS_NO_MORE_ITEMS) {
			if ((i == CS_SUCCESS) && (cf->io.nwin == 2)) {
				link->conf.ConfigIndex = cf->index;
				link->io.BasePort1 = cf->io.win[0].base;
				link->io.BasePort2 = cf->io.win[1].base;
				link->io.IOAddrLines = cf->io.flags & CISTPL_IO_LINES_MASK;
				i = CardServices(RequestIO, link->handle, &link->io);
				base2 = link->io.BasePort2;
				if (i == CS_SUCCESS) break;
			}
			i = next_tuple(handle, &tuple, &parse);
		}
	}
    

found_port:
	if (i != CS_SUCCESS) {
		cs_error(link->handle, RequestIO, i);
		return -1;
	}
    
	//DBPRT(DBINIT, "IRQ: %d\n", link->irq.AssignedIRQ);
	i = CardServices(RequestIRQ, link->handle, &link->irq);
	if (i != CS_SUCCESS) {
		cs_error(link->handle, RequestIRQ, i);
		link->irq.AssignedIRQ = 0;
	}

	if (btuart->multi && (btuart->manfid == MANFID_3COM))
		link->conf.ConfigIndex &= ~(0x08);
	i = CardServices(RequestConfiguration, link->handle, &link->conf);
	if (i != CS_SUCCESS) {
		cs_error(link->handle, RequestConfiguration, i);
		return -1;
	}

	setup_btuart(btuart, link->io.BasePort1, link->irq.AssignedIRQ);
	for (i = 0; i < btuart->multi-1; i++)
		setup_btuart(btuart, base2+(8*i), link->irq.AssignedIRQ);

	return 0;
}

int multi_config(dev_link_t *link)
{
	client_handle_t handle = link->handle;
	btuartcs_device *btuart = link->priv;
	tuple_t tuple;
	u_char buf[256];
	cisparse_t parse;
	cistpl_cftable_entry_t *cf = &parse.cftable_entry;
	int i, base2 = 0;

	DBFENTER;

	tuple.TupleData = (cisdata_t *)buf;
	tuple.TupleOffset = 0; tuple.TupleDataMax = 255;
	tuple.Attributes = 0;
	tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
	
	/* First, look for a generic full-sized window */
	link->io.NumPorts1 = btuart->multi * 8;
	i = first_tuple(handle, &tuple, &parse);
	while (i != CS_NO_MORE_ITEMS) {
		/* The quad port cards have bad CIS's, so just look for a
		   window larger than 8 ports and assume it will be right */
		DBPRT(DBINIT, "nwin: %d, len: %d\n", cf->io.nwin, cf->io.win[0].len);
		if ((i == CS_SUCCESS) && (cf->io.nwin == 1) &&
		    (cf->io.win[0].len > 8)) {
			link->conf.ConfigIndex = cf->index;
			link->io.BasePort1 = cf->io.win[0].base;
			link->io.IOAddrLines = cf->io.flags & CISTPL_IO_LINES_MASK;
			i = CardServices(RequestIO, link->handle, &link->io);
			base2 = link->io.BasePort1 + 8;
			if (i == CS_SUCCESS) break;
		}
		i = next_tuple(handle, &tuple, &parse);
	}

	/* If that didn't work, look for two windows */
	if (i != CS_SUCCESS) {
		link->io.NumPorts1 = link->io.NumPorts2 = 8;
		btuart->multi = 2;
		i = first_tuple(handle, &tuple, &parse);
		while (i != CS_NO_MORE_ITEMS) {
			if ((i == CS_SUCCESS) && (cf->io.nwin == 2)) {
				link->conf.ConfigIndex = cf->index;
				link->io.BasePort1 = cf->io.win[0].base;
				link->io.BasePort2 = cf->io.win[1].base;
				link->io.IOAddrLines = cf->io.flags & CISTPL_IO_LINES_MASK;
				i = CardServices(RequestIO, link->handle, &link->io);
				base2 = link->io.BasePort2;
				if (i == CS_SUCCESS) break;
			}
			i = next_tuple(handle, &tuple, &parse);
		}
	}
    
	if (i != CS_SUCCESS) {
		cs_error(link->handle, RequestIO, i);
		return -1;
	}
    
	i = CardServices(RequestIRQ, link->handle, &link->irq);
	if (i != CS_SUCCESS) {
		cs_error(link->handle, RequestIRQ, i);
		link->irq.AssignedIRQ = 0;
	}
	/* Socket Dual IO: this enables irq's for second port */
	if (btuart->multi && (btuart->manfid == MANFID_SOCKET)) {
		link->conf.Present |= PRESENT_EXT_STATUS;
		link->conf.ExtStatus = ESR_REQ_ATTN_ENA;
	}
	i = CardServices(RequestConfiguration, link->handle, &link->conf);
	if (i != CS_SUCCESS) {
		cs_error(link->handle, RequestConfiguration, i);
		return -1;
	}
    
	setup_btuart(btuart, link->io.BasePort1, link->irq.AssignedIRQ);
	for (i = 0; i < btuart->multi-1; i++)
		setup_btuart(btuart, base2+(8*i), link->irq.AssignedIRQ);
    
	return 0;
}

/*======================================================================

    btuart_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    serial device available to the system.

======================================================================*/

#define CS_CHECK(fn, args...) \
while ((last_ret=CardServices(last_fn=(fn), args))!=0) goto cs_failed

void btuart_config(dev_link_t *link)
{
	client_handle_t handle = link->handle;
	btuartcs_device *btuart = link->priv;
	tuple_t tuple;
	u_short buf[128];
	cisparse_t parse;
	cistpl_cftable_entry_t *cf = &parse.cftable_entry;
	int i, j, last_ret, last_fn;

	DBFENTER;
    
	tuple.TupleData = (cisdata_t *)buf;
	tuple.TupleOffset = 0;
	tuple.TupleDataMax = 255;
	tuple.Attributes = 0;
	/* Get configuration register information */
	tuple.DesiredTuple = CISTPL_CONFIG;
	last_ret = first_tuple(handle, &tuple, &parse);
	if (last_ret != CS_SUCCESS) {
		last_fn = ParseTuple;
		goto cs_failed;
	}
	link->conf.ConfigBase = parse.config.base;
	link->conf.Present = parse.config.rmask[0];
    
	/* Configure card */
	link->state |= DEV_CONFIG;

	/* Is this a compliant multifunction card? */
	tuple.DesiredTuple = CISTPL_LONGLINK_MFC;
	tuple.Attributes = TUPLE_RETURN_COMMON | TUPLE_RETURN_LINK;
	btuart->multi = (first_tuple(handle, &tuple, &parse) == CS_SUCCESS);

	/* Is this a multiport card? */
	tuple.DesiredTuple = CISTPL_MANFID;
	if (first_tuple(handle, &tuple, &parse) == CS_SUCCESS) {
		btuart->manfid = le16_to_cpu(buf[0]);
		btuart->prodid = le16_to_cpu(buf[1]);
		for (i = 0; i < MULTI_COUNT; i++)
			if ((btuart->manfid == multi_id[i].manfid) &&
			    (btuart->prodid == multi_id[i].prodid))
				break;
		if (i < MULTI_COUNT)
			btuart->multi = multi_id[i].multi;
	}

	/* Another check for dual-btuart cards: look for either serial or
	   multifunction cards that ask for appropriate IO port ranges */
	tuple.DesiredTuple = CISTPL_FUNCID;
	if ((btuart->multi == 0) &&
	    ((first_tuple(handle, &tuple, &parse) != CS_SUCCESS) ||
	     (parse.funcid.func == CISTPL_FUNCID_MULTI) ||
	     (parse.funcid.func == CISTPL_FUNCID_SERIAL))) {
		tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
		if ((first_tuple(handle, &tuple, &parse) == CS_SUCCESS) &&
		    (((cf->io.nwin == 1) && (cf->io.win[0].len == 16)) ||
		     ((cf->io.nwin == 2) && (cf->io.win[0].len == 8) &&
		      (cf->io.win[1].len == 8))))
			btuart->multi = 2;
	}
    
	if (btuart->multi > 1)
		multi_config(link);
	else
		simple_config(link);
    
	DBPRT(DBINIT, "Found %d devices\n", btuart->ndev);

	if (btuart->ndev == 0)
		goto failed;
    
	if (btuart->manfid == MANFID_IBM) {
		conf_reg_t reg = { 0, CS_READ, 0x800, 0 };
		CS_CHECK(AccessConfigurationRegister, link->handle, &reg);
		reg.Action = CS_WRITE;
		reg.Value = reg.Value | 1;
		CS_CHECK(AccessConfigurationRegister, link->handle, &reg);
	}

	link->dev = &btuart->node[0];

	/* register uart device */
	for (i = 0; i < btuart->ndev;i++) {
		last_ret = affix_uart_attach(&btuart->hw[i]);
		if (last_ret) {
			BTERRORF("Can't register network device\n");
			/* detach previously attached */
			for (j = i-1; j >= 0; j--)
				affix_uart_detach(btuart->hw[i].kdev);
			goto failed;
		}
	}

	link->state &= ~DEV_CONFIG_PENDING;
	DBFEXIT;
	return;

cs_failed:
	cs_error(link->handle, last_fn, last_ret);
failed:
	btuart_release((u_long)link);

}

/*======================================================================

    After a card is removed, btuart_release() will unregister the net
    device, and release the PCMCIA configuration.
    
======================================================================*/

void btuart_release(u_long arg)
{
	dev_link_t *link = (dev_link_t *)arg;
	btuartcs_device *btuart = link->priv;
	int i;
    
	DBFENTER;

	DBPRT(DBINIT, "irq?: %d, int?: %d\n", in_irq(), in_interrupt());

	for (i = 0; i < btuart->ndev; i++) {
		unregister_serial(btuart->line[i]);
	}
	link->dev = NULL;

	if (!btuart->slave) {
		CardServices(ReleaseConfiguration, link->handle);
		CardServices(ReleaseIO, link->handle, &link->io);
		CardServices(ReleaseIRQ, link->handle, &link->irq);
	}
    
	link->state &= ~DEV_CONFIG;

}

/*======================================================================

    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.  A CARD_REMOVAL event
    also sets some flags to discourage the serial drivers from
    talking to the ports.
    
======================================================================*/

int btuart_event(event_t event, int priority, event_callback_args_t *args)
{
	dev_link_t	*link = args->client_data;
	btuartcs_device 	*btuart = link->priv;
	int	i;
    
	DBFENTER;
    
	DBPRT(DBINIT, "irq?: %d, int?: %d\n", in_irq(), in_interrupt());

	switch (event) {
	case CS_EVENT_CARD_REMOVAL:
		DBPRT(DBINIT, "CS_EVENT_CARD_REMOVAL\n");
		link->state &= ~DEV_PRESENT;
		if (link->state & DEV_CONFIG) {
			for (i = 0; i < btuart->ndev;i++) 
				affix_uart_suspend(btuart->hw[i].kdev);
			link->state |= DEV_RELEASE_PENDING;
			mod_timer(&link->release, jiffies + HZ/20);
		}
		break;
	case CS_EVENT_CARD_INSERTION:
		DBPRT(DBINIT, "CS_EVENT_CARD_INSERTTION\n");
		link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
		btuart_config(link);
		break;
	case CS_EVENT_PM_SUSPEND:
		DBPRT(DBINIT, "CS_EVENT_PM_SUSPEND\n");
		link->state |= DEV_SUSPEND;
		/* Fall through... */
	case CS_EVENT_RESET_PHYSICAL:
		DBPRT(DBINIT, "CS_EVENT_RESET_PHYSICAL\n");
		if( (link->state & DEV_CONFIG) && !btuart->slave) {
			if (link->open)
				for (i = 0; i < btuart->ndev;i++) 
					affix_uart_suspend(btuart->hw[i].kdev);
			CardServices(ReleaseConfiguration, link->handle);
			DBPRT(DBINIT, "CS_EVENT_RESET_PHYSICAL\n");
		}
		break;
	case CS_EVENT_PM_RESUME:
		DBPRT(DBINIT, "CS_EVENT_PM_RESUME\n");
		link->state &= ~DEV_SUSPEND;
		/* Fall through... */
	case CS_EVENT_CARD_RESET:
		DBPRT(DBINIT, "CS_EVENT_CARD_RESET\n");
		if (DEV_OK(link) && !btuart->slave) {
			CardServices(RequestConfiguration, link->handle, &link->conf);
			if (link->open) 
				for (i = 0; i < btuart->ndev;i++) 
					affix_uart_resume(btuart->hw[i].kdev);
			DBPRT(DBINIT, "CS_EVENT_CARD_RESET\n");
		}
		break;
	}

	DBFEXIT;
	return 0;
}

/*====================================================================*/

int __init init_btuart_cs(void)
{
	int	        err;
	servinfo_t 	serv;

	DBFENTER;

	CardServices(GetCardServicesInfo, &serv);
	if (serv.Revision != CS_RELEASE_CODE) {
		printk(KERN_NOTICE "affix_uart_cs: Card Services release "
		       "does not match!\n");
		return -1;
	}

	err = register_pccard_driver(&dev_info, &btuart_attach, &btuart_detach);
	if (err)
		return err;

	DBFEXIT;
	return 0;
}

void __exit exit_btuart_cs(void)
{
	DBFENTER;

	unregister_pccard_driver(&dev_info);
	while (dev_list != NULL)
		btuart_detach(dev_list);
	DBFEXIT;
}

/*  If we are resident in kernel we want to call init_btuart_cs manually. */
#ifdef MODULE
module_init(init_btuart_cs);
module_exit(exit_btuart_cs);

MODULE_AUTHOR("Dmitry Kasatkin <dmitry.kasatkin@nokia.com>");
MODULE_DESCRIPTION("PCMCIA UART driver for Bluetooth Cards");
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 15)
MODULE_LICENSE("GPL");
#endif

#endif // MODULE

EXPORT_NO_SYMBOLS;

