/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Imre Deak <ext-imre.deak@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: btusb.c,v 1.1.1.1 2002/07/11 12:19:05 kds Exp $

   Physical protocol layer for USB Bluetooth devices

   Fixes:	Imre Deak <ext-imre.deak@nokia.com>

   		Dmitry Kasatkin		: Audio Support, New Request Management
*/		

#define __NO_VERSION__

#include <linux/config.h>

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/fcntl.h>

#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/rtnetlink.h>

#include <linux/usb.h>

#define FILEBIT DBUSB		// for sake of selective debug messages

#include <affix/btdebug.h>
#include <affix/hci.h>
#include <affix/btdebug.h>

unsigned long long	isoc_count;


#define VENDOR_ID_DIGIANSWER			0x8fd
#define DEVICE_ID_DIGIANSWER_USB_BT_DONGLE	1
#define VENDOR_ID_SILICONWAVE			0xc10
#define DEVICE_ID_SILICONWAVE_USB_BT_MODULE	0
#define VENDOR_ID_CSR				0x0a12
#define DEVICE_ID_CSR				0x01
#define VENDOR_3COM				0x506
#define DEVICE_3COM_USB_BT_MODULE		0xa0

#define NUM_OF_USB_BT_DEV (sizeof(usb_bt_list) / sizeof(usb_bt_list[0]))	

/* Class, SubClass, and Protocol codes that describe a Bluetooth device */
#define WIRELESS_CLASS_CODE			0xe0
#define RF_SUBCLASS_CODE			0x01
#define BLUETOOTH_PROGRAMMING_PROTOCOL_CODE	0x01

/* HCI packet types */
#define HCI_CMD_PKT	0x01
#define HCI_ACL_PKT	0x02
#define HCI_SCO_PKT	0x03
#define HCI_EVENT_PKT	0x04

/* Message sizes */
#define EVENT_HDR_SIZE	2
#define ACL_HDR_SIZE	4
#define SCO_HDR_SIZE	3

/*
#define BTUSB_BULK_IN_EP_ADDR		0x82
#define BTUSB_BULK_OUT_EP_ADDR		0x02
#define BTUSB_INT_IN_EP_ADDR		0x81
#define BTUSB_ISOC_IN_EP_ADDR		0x83
#define BTUSB_ISOC_OUT_EP_ADDR		0x03
*/

#define BTUSB_INT_IN_BUFFER_SIZE	16
#define BTUSB_BULK_IN_BUFFER_SIZE	2048	//64

#define BTUSB_ISOC_FRAMES		5
#define BTUSB_ISOC_SIZE			64

#define BTUSB_ISOC_1_MAX_BUFFER_SIZE	9
#define BTUSB_ISOC_2_MAX_BUFFER_SIZE	17
#define BTUSB_ISOC_3_MAX_BUFFER_SIZE	25
#define BTUSB_ISOC_4_MAX_BUFFER_SIZE	33
#define BTUSB_ISOC_5_MAX_BUFFER_SIZE	49

#define BTUSB_INT_IN_INTERVAL		1

/*  Maximum number of consecutive failed USB reads before giving up. */
#define MAX_READ_FAIL_COUNT		100

#define BTUSB_URBALL		0x03
#define BTUSB_URB0PENDING	0x01
#define BTUSB_URB1PENDING	0x02
#define BTUSB_URBLASTPENDING	BTUSB_URB1PENDING	/* last urb */
#define BTUSB_RUNNING		0x10
#define BTUSB_AUDIO_IN		0x20
#define BTUSB_AUDIO_OUT		0x40

#define BTUSB_URBPOOL		2

typedef struct _usb_bt_params
{
	int	uart_leading_byte;	// if there is a uart-style leading byte for packets
	unsigned char control_req_type;
	int	vendor_id;
	int	device_id;
} usb_bt_params_t;

usb_bt_params_t usb_bt_list[] =
{
	{
		uart_leading_byte : 1,
		control_req_type : 0x40,
		vendor_id : VENDOR_ID_DIGIANSWER,
		device_id : DEVICE_ID_DIGIANSWER_USB_BT_DONGLE
	},
        {       uart_leading_byte : 0,		// default last descriptor
                control_req_type : 0x20,
                vendor_id : 0,
                device_id : 0 
        }
};

/* request data structure */
typedef struct btusb_req
{
	struct btusb		*btusb;
	struct urb		*urb[BTUSB_URBPOOL];
	unsigned char		*buffer[BTUSB_URBPOOL];
	int			flags;
	int			num_packets;	/* for iso */
	int			packet_size;	/* for pipe */
	unsigned int		pipe;
	struct sk_buff		*skb;
	int			act_data_len;
	int			pkt_len;

	struct sk_buff_head	queue;
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	devrequest		dr;		/* for setup packets */
#else
	struct usb_ctrlrequest	dr;		/* for setup packets */
#endif
#ifdef CONFIG_AFFIX_PROF
	struct timeval		start_stamp, end_stamp;
	long int		delay;
	int			bytes;
	long int		count;

#endif
} btusb_req_t;


/* driver data structure */
typedef struct btusb
{
	int			signature;
	struct usb_device	*usb_dev;
	struct net_device	*net_dev;
	struct net_device_stats	net_stats;
	int			flowmask;
	usb_bt_params_t		*params;
	int			flags;
	int			isoc;
	
	spinlock_t		btusb_lock;
	int			spinlock_cnt;
	
	btusb_req_t		ctrl_out_req;
	btusb_req_t		int_in_req;
	btusb_req_t		bulk_in_req;
	btusb_req_t		bulk_out_req;
	btusb_req_t		isoc_in_req;
	btusb_req_t		isoc_out_req;

	int			read_fail_cnt;

} btusb_t;

static void btusb_out_complete(struct urb *urb);
static void btusb_int_in_complete(struct urb *urb);
static void btusb_bulk_in_complete(struct urb *urb);
static void btusb_isoc_in_complete(struct urb *urb);
static void btusb_isoc_out_complete(struct urb *urb);

static void *btusb_probe(struct usb_device *dev, unsigned int ifnum,
			 	 const struct usb_device_id *id);
static void btusb_disconnect(struct usb_device *dev, void *ptr);

static struct usb_device_id btusb_ids [] =
{
	{
		USB_DEVICE_INFO(WIRELESS_CLASS_CODE,
				RF_SUBCLASS_CODE,
				BLUETOOTH_PROGRAMMING_PROTOCOL_CODE)
	},
	{ }
};

MODULE_DEVICE_TABLE (usb, btusb_ids);

static struct usb_driver btusb_driver =
{
	name:		"affix_usb",
	probe:		btusb_probe,
	disconnect:	btusb_disconnect,
	id_table:	btusb_ids
};

static void btusb_lock(btusb_t *btusb, unsigned int *old_flags)
{
	spin_lock_irqsave(&btusb->btusb_lock, (*old_flags));
}

static void btusb_unlock(btusb_t *btusb, unsigned int old_flags)
{
	spin_unlock_irqrestore(&btusb->btusb_lock, old_flags);
}


static int btusb_submit_req(btusb_req_t *req, int urb)
{
	int	result = 0;

	DBFENTER;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	result = usb_submit_urb(req->urb[urb]);
#else
	result = usb_submit_urb(req->urb[urb], GFP_ATOMIC);
#endif
	if (result) {
		BTERRORF("usb_submit_urb [%d] failed: %d\n", urb, result);
		req->flags &= ~(BTUSB_RUNNING & (1 << urb));
		return result;
	}
	req->flags |= (1 << urb);

	DBFEXIT;
	return result;
}

static void btusb_release_req(btusb_req_t *req)
{
	int j, mask;

	DBFENTER;
	for (j = 0, mask = 1; j < BTUSB_URBPOOL; j++, mask<<=1) {
		if (req->urb[j] == NULL)
			continue;
		if (req->flags & mask) {
			usb_unlink_urb(req->urb[j]);
			req->flags &= ~mask;
		}
		if (req->buffer[j])
			kfree(req->buffer[j]);
		if (req->skb) {
			dev_kfree_skb_any(req->skb);
			req->skb = NULL;
		}
		usb_free_urb(req->urb[j]);
		req->urb[j] = NULL;
	}
	DBFEXIT;
}

static inline void  __usb_fill_isoc_urb(struct urb *urb, struct usb_device *dev,
		unsigned int pipe, void *buf, int num_packets, int packet_size,
		usb_complete_t complete, void *context) 
{
	int k, offset;

	spin_lock_init(&urb->lock);
	urb->dev = dev;
	urb->pipe = pipe;
	urb->transfer_buffer = buf;
	urb->number_of_packets = num_packets;
	urb->transfer_buffer_length = num_packets*packet_size;
	urb->actual_length = 0;
	urb->complete = complete;
	urb->context = context;
	urb->transfer_flags = USB_ISO_ASAP;
	for (k = 0, offset = 0; k < num_packets; k++, offset += packet_size) {
		urb->iso_frame_desc[k].offset = offset;
		urb->iso_frame_desc[k].length = packet_size;
		urb->iso_frame_desc[k].actual_length = 0;
	}
}


static int btusb_setup_isoc_in(btusb_req_t *req, btusb_t *btusb)
{
	int	j, retval;
	int	num_packets, packet_size, mtu;

	req->btusb = btusb;
	mtu = usb_maxpacket(btusb->usb_dev, req->pipe, usb_pipeout(req->pipe));
	packet_size = req->packet_size = BTUSB_ISOC_SIZE;	// FIXME: use mtu??
	num_packets = req->num_packets = BTUSB_ISOC_FRAMES;
	
	DBPRT(DBXMIT, "mtu: %d, packet_size: %d\n", mtu, packet_size);

	for (j = 0; j < BTUSB_URBPOOL; j++) {
		retval = -ENOMEM;
		req->urb[j] = usb_alloc_urb(num_packets);
		if (!req->urb[j])
			goto err;

		req->buffer[j] = kmalloc(num_packets*packet_size, GFP_KERNEL);
		if (!req->buffer[j])
			goto err;
			
		__usb_fill_isoc_urb(req->urb[j], btusb->usb_dev,
				req->pipe, req->buffer[j],
				num_packets, packet_size,
				btusb_isoc_in_complete, req);
	}
	return 0;
 err:
	for (j = 0; j < BTUSB_URBPOOL; j++) {
		if (req->urb[j]) {
			if (req->buffer[j])
				kfree(req->buffer[j]);
			usb_free_urb(req->urb[j]);
			req->urb[j] = NULL;
		}
	}
	return retval;
}

static int btusb_setup_isoc_out(btusb_req_t *req, btusb_t *btusb)
{
	int	j, retval;
	int	num_packets, packet_size, mtu;

	req->btusb = btusb;
	mtu = usb_maxpacket(btusb->usb_dev, req->pipe, usb_pipein(req->pipe));
	packet_size = req->packet_size = mtu;
	num_packets = req->num_packets = BTUSB_ISOC_FRAMES;
	
	DBPRT(DBXMIT, "mtu: %d, packet_size: %d\n", mtu, packet_size);

	for (j = 0; j < BTUSB_URBPOOL; j++) {
		retval = -ENOMEM;
		req->urb[j] = usb_alloc_urb(num_packets);
		if (!req->urb[j])
			goto err;

		__usb_fill_isoc_urb(req->urb[j], btusb->usb_dev,
				req->pipe, req->buffer[j],
				num_packets, packet_size,
				btusb_isoc_out_complete, req);
	}
	return 0;
 err:
	for (j = 0; j < BTUSB_URBPOOL; j++) {
		if (req->urb[j]) {
			usb_free_urb(req->urb[j]);
			req->urb[j] = NULL;
		}
	}
	return retval;
}



static int btusb_setup_int_in(btusb_req_t *req, btusb_t *btusb)
{
	int 	j, retval;

	req->btusb = btusb;
	for (j = 0; j < BTUSB_URBPOOL; j++) {
		retval = -ENOMEM;
		req->urb[j] = usb_alloc_urb(0);
		if (!req->urb[j])
			goto err;

		req->buffer[j] = kmalloc(BTUSB_INT_IN_BUFFER_SIZE, GFP_KERNEL);
		if (!req->buffer[j])
			goto err;
			
		FILL_INT_URB(req->urb[j], btusb->usb_dev, req->pipe,
				req->buffer[j], BTUSB_INT_IN_BUFFER_SIZE,
				btusb_int_in_complete, req, BTUSB_INT_IN_INTERVAL);
	}
	return 0;
 err:
	for (j = 0; j < BTUSB_URBPOOL; j++) {
		if (req->urb[j]) {
			if (req->buffer[j])
				kfree(req->buffer[j]);
			usb_free_urb(req->urb[j]);
			req->urb[j] = NULL;
		}
	}
	return retval;
}

static int btusb_setup_bulk_in(btusb_req_t *req, btusb_t *btusb)
{
	int 	j, retval;

	req->btusb = btusb;
	for (j = 0; j < BTUSB_URBPOOL; j++) {
		retval = -ENOMEM;
		req->urb[j] = usb_alloc_urb(0);
		if (!req->urb[j])
			goto err;

		req->buffer[j] = kmalloc(BTUSB_BULK_IN_BUFFER_SIZE, GFP_KERNEL);
		if (!req->buffer[j])
			goto err;

		FILL_BULK_URB(req->urb[j], btusb->usb_dev, req->pipe,
				req->buffer[j], BTUSB_BULK_IN_BUFFER_SIZE,
				btusb_bulk_in_complete, req);
	}
	return 0;

err:
	for (j = 0; j < BTUSB_URBPOOL; j++) {
		if (req->urb[j]) {
			if (req->buffer[j])
				kfree(req->buffer[j]);
			usb_free_urb(req->urb[j]);
			req->urb[j] = NULL;
		}
	}
	return retval;
}

static int btusb_setup_bulk_out(btusb_req_t *req, btusb_t *btusb)
{
	int 	j, retval;

	req->btusb = btusb;
	for (j = 0; j < BTUSB_URBPOOL; j++) {
		retval = -ENOMEM;
		req->urb[j] = usb_alloc_urb(0);
		if (!req->urb[j])
			goto err;
	}
	return 0;

err:
	for (j = 0; j < BTUSB_URBPOOL; j++) {
		if (req->urb[j]) {
			usb_free_urb(req->urb[j]);
			req->urb[j] = NULL;
		}
	}
	return retval;
}


static int btusb_setup_ctrl_out(btusb_req_t *req, btusb_t *btusb) 
{
	int 	j, retval;

	req->btusb = btusb;
	for (j = 0; j < BTUSB_URBPOOL; j++) {
		retval = -ENOMEM;
		req->urb[j] = usb_alloc_urb(0);
		if (!req->urb[j])
			goto err;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
		req->dr.requesttype = btusb->params->control_req_type;
		req->dr.request = 0;
		req->dr.value = 0;
		req->dr.index = 0;
#else
		req->dr.bRequestType = btusb->params->control_req_type;
		req->dr.bRequest = 0;
		req->dr.wValue = 0;
		req->dr.wIndex = 0;
#endif
	}

	return 0;
err:
	for (j = 0; j < BTUSB_URBPOOL; j++) {
		if (req->urb[j]) {
			usb_free_urb(req->urb[j]);
			req->urb[j] = NULL;
		}
	}
	return retval;
}

static inline void btusb_fill_control_urb(btusb_req_t *req, struct sk_buff *skb)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	req->dr.length = cpu_to_le16(skb->len);
#else
	req->dr.wLength = cpu_to_le16(skb->len);
#endif
	FILL_CONTROL_URB(req->urb[0], req->btusb->usb_dev, req->pipe,
			(unsigned char*)&req->dr,
			skb->data, skb->len, btusb_out_complete, req);
	req->skb = skb;
}

static inline void btusb_fill_bulk_urb(btusb_req_t *req, struct sk_buff *skb)
{
	FILL_BULK_URB(req->urb[0], req->btusb->usb_dev, req->pipe,
			skb->data, skb->len,
			btusb_out_complete, req);
	req->urb[0]->transfer_flags |= USB_QUEUE_BULK;
	req->skb = skb;
}

static inline void btusb_fill_isoc_urb(btusb_req_t *req, struct sk_buff *skb, int id)
{
	int	num_packets = req->num_packets, packet_size = req->packet_size;
	int 	k, len = skb->len, offset = 0;
	struct	urb	*urb = req->urb[id];

	urb->dev = req->btusb->usb_dev;
	//urb->transfer_flags = 0;	// no USB_ISOC_ASAP
	//urb->start_frame = 1;
	urb->transfer_buffer = skb->data;
	urb->actual_length = 0;
	for (k = 0; k < num_packets && len; k++) {
		urb->iso_frame_desc[k].offset = offset;
		urb->iso_frame_desc[k].actual_length = 0;
		urb->iso_frame_desc[k].length = (len>packet_size)?packet_size:len;
		len -= urb->iso_frame_desc[k].length;
		offset += urb->iso_frame_desc[k].length;
	}
	urb->number_of_packets = k;
	urb->transfer_buffer_length = skb->len - len;
}


static int btusb_audio_mode(btusb_t *btusb, int audio_mode, unsigned int *old_flags)
{
	int		result;

	DBFENTER;
	isoc_count = 0;

	if (!btusb->isoc)
		return -ENODEV;

	if ((audio_mode & AFFIX_AUDIO_IN) && !(btusb->flags & BTUSB_AUDIO_IN)) {
		result = btusb_setup_isoc_in(&btusb->isoc_in_req, btusb);
		if (result) {
			BTERRORF("setup isoc pipes failed with status %d\n", result);
			return result;
		}

		result = btusb_submit_req(&btusb->isoc_in_req, 0);
		if (result) {
			BTERRORF("usb_submit urb for isoc in pipe failed with status %d\n", result);
			return result;
		}
		result = btusb_submit_req(&btusb->isoc_in_req, 1);
		if (result) {
			BTERRORF("usb_submit urb for isoc in pipe failed with status %d\n", result);
			return result;
		}
		btusb->flags |= BTUSB_AUDIO_IN;
	} else {
		if (!(audio_mode & AFFIX_AUDIO_IN) && (btusb->flags & BTUSB_AUDIO_IN)) {
			btusb->flags &= ~BTUSB_AUDIO_IN;
			btusb_unlock(btusb, *old_flags);
			btusb_release_req(&btusb->isoc_in_req);
			btusb_lock(btusb, old_flags);
			BTINFO("Received isoc: %llu\n", isoc_count);
			isoc_count = 0;
		}
	}

	if ((audio_mode & AFFIX_AUDIO_OUT) && !(btusb->flags & BTUSB_AUDIO_OUT)) {
		// setup outgoing request
		result = btusb_setup_isoc_out(&btusb->isoc_out_req, btusb);
		if (result) {
			BTERRORF("setup isoc pipes failed with status %d\n", result);
			return result;
		}
		btusb->flags |= BTUSB_AUDIO_OUT;
	} else {
		if (!(audio_mode & AFFIX_AUDIO_OUT) && (btusb->flags & BTUSB_AUDIO_OUT)) {
			btusb->flags &= ~BTUSB_AUDIO_OUT;
			btusb_unlock(btusb, *old_flags);
			btusb_release_req(&btusb->isoc_out_req);
			btusb_lock(btusb, old_flags);
		}
	}
	
	DBFEXIT;
	return 0;
}


/*----------------------------------------------------------------------------------------------
				Network interface functions.
  ----------------------------------------------------------------------------------------------
*/

static int btusb_net_open(struct net_device *net_dev)
{
	btusb_t	*btusb = net_dev->priv;
	int	result;

	DBFENTER;

	if (btusb == NULL) {
		BTERRORF("btusb is NULL\n");
	}

	if (btusb->flags & BTUSB_RUNNING) {
		BTERRORF("device already open\n");
		DBFEXIT;
		return -EINVAL;
	}

	btusb->flowmask = HCI_CMD_ON | HCI_ACL_ON | HCI_SCO_ON;
	
	result = btusb_setup_int_in(&btusb->int_in_req, btusb);
	if (result) {
		BTERRORF("int setup failed: %d\n", result);
		goto err1;
	}
	result = btusb_submit_req(&btusb->int_in_req, 0);
	if (result) {
		BTERRORF("usb_submit_urb for interrupt in pipe failed: %d\n", result);
		goto err2;
	}

	result = btusb_setup_bulk_in(&btusb->bulk_in_req, btusb);
	if (result) {
		BTERRORF("bulk setup failed: %d\n", result);
		goto err2;
	}
	result = btusb_submit_req(&btusb->bulk_in_req, 0);
	if (result) {
		BTERRORF("usb_submit urb for bulk in pipe failed: %d\n", result);
		goto err3;
	}

	result = btusb_setup_bulk_out(&btusb->bulk_out_req, btusb);
	if (result) {
		BTERRORF("bulk setup failed: %d\n", result);
		goto err3;
	}

	result = btusb_setup_ctrl_out(&btusb->ctrl_out_req, btusb);
	if (result) {
		BTERRORF("ctrl setup failed: %d\n", result);
		goto err4;
	}

	btusb->flags |= BTUSB_RUNNING;
	hci_start_queue(net_dev, btusb->flowmask);

	DBFEXIT;
	return 0;
err4:
	btusb_release_req(&btusb->bulk_out_req);
err3:
	btusb_release_req(&btusb->bulk_in_req);
err2:
	btusb_release_req(&btusb->int_in_req);
err1:
	return result;
}

static int btusb_net_stop(struct net_device *net_dev)
{
	btusb_t *btusb = net_dev->priv;

	DBFENTER;

	hci_stop_queue(net_dev, HCI_CMD_ON | HCI_ACL_ON | HCI_SCO_ON);

	if (btusb == NULL) {
		BTERRORF("btusb is NULL\n");
		DBFEXIT;
		return -1;
	}

	if (!(btusb->flags & BTUSB_RUNNING)){
		BTERRORF("device not opened\n");
		DBFEXIT;
		return -1;
	}
	
	btusb->flags &= ~(BTUSB_RUNNING | BTUSB_AUDIO_IN | BTUSB_AUDIO_OUT);

	/* unlinking urb means calling the completion routine of any pending transaction
	   it's not a problem since there we will check btusb->active.
	 */
	btusb_release_req(&btusb->int_in_req);
	btusb_release_req(&btusb->ctrl_out_req);
	btusb_release_req(&btusb->bulk_in_req);
	btusb_release_req(&btusb->bulk_out_req);
	btusb_release_req(&btusb->isoc_in_req);
	btusb_release_req(&btusb->isoc_out_req);

	DBFEXIT;

	return 0;
}

#ifdef CONFIG_AFFIX_PROF
void btusb_calc_delay(int type, btusb_req_t *req, struct sk_buff *skb)
{
	switch (type) {
		case 1: {
				static unsigned long	time = 0;
				suseconds_t		usec;

				req->count++;
				if (req->skb)
					req->bytes += req->skb->len;
				get_fast_time(&req->end_stamp);
				usec = sub_timeval(&req->end_stamp, &req->start_stamp);
				req->delay += usec;
				if (jiffies - time > HZ ) {
					time = jiffies;
					BTINFOF("complete count: %lu\n", req->count);
					req->count = 0;

					BTINFO("completed %d bytes for %lu usecs\n", (req->skb)?req->skb->len:0, usec);
					BTINFO("%d bytes for sec:\n", req->bytes);
					req->bytes = 0;

					BTINFO("Total completion time: %lu\n", req->delay);
					req->delay = 0;

				}

			}
			break;

		case 2: {

				static long int		time = 0;
				suseconds_t		usec;
				static suseconds_t	usec_count = 0;

				get_fast_time(&req->start_stamp);
				usec = sub_timeval(&req->start_stamp, &req->end_stamp);
				usec_count += usec;
				if (jiffies - time > 100) {
					time = jiffies;
					BTINFO("recover time: %lu\n", usec);
					BTINFO("Total recovery time: %lu\n", usec_count);
					usec_count = 0;
				}
			}
			break;

		case 3:
			{
				//debug
				static	unsigned long	time = 0, count = 0;
				count++;
				if (jiffies - time > HZ ) {
					time = jiffies;
					BTINFOF("complete count: %lu\n", count);
					count = 0;
				}
			}
			break;

		case 4: {
				/* isoc in complete */
				static	unsigned long	time = 0, count = 0;
				
				count += req->skb->len;
				if (jiffies - time > HZ ) {
					time = jiffies;
					BTINFOF("isoc_in %lu bytes\n", count);
					count = 0;
				}

			}
			break;

		case 5: {
				/* isoc out complete */
				static	unsigned long	time = 0, count = 0;

				count += req->skb->len;
				if (jiffies - time > HZ ) {
					time = jiffies;
					BTINFOF("isoc_out %lu bytes\n", count);
					count = 0;
				}
			}
			break;

	}
}
#endif

int btusb_xmit_start(btusb_req_t *req)
{
	int	err = 0;

	DBFENTER;
	
	if (req->flags & BTUSB_URB0PENDING)
		return 0;
	req->skb = skb_dequeue(&req->queue);
	if (req->skb) {
		if (usb_pipebulk(req->pipe))
			btusb_fill_bulk_urb(req, req->skb);
		else
			btusb_fill_control_urb(req, req->skb);
			
		if ((err = btusb_submit_req(req, 0))) {
			BTERRORF("submit urb failed\n");
			skb_queue_head(&req->queue, req->skb);
			req->skb = NULL;
			goto exit;
		}
	} else {
		hci_wake_queue(req->btusb->net_dev, req->btusb->flowmask);
	}
	
exit:
	DBFEXIT;
	return err;
}

int btusb_xmit_start_isoc(btusb_req_t *req, int urb)
{
	int	err = 0;

	DBFENTER;
	
	/* find free URB */
	for (urb = 0; urb < BTUSB_URBPOOL; urb++ ) {
		if (req->flags & (1 << urb))
			continue;
		if (!req->skb) {
			req->skb = skb_dequeue(&req->queue);
			if (!req->skb) {
				hci_wake_queue(req->btusb->net_dev, req->btusb->flowmask);
				break;
			}
#ifdef CONFIG_AFFIX_PROF			
			btusb_calc_delay(5, req, NULL);
#endif
		}
		btusb_fill_isoc_urb(req, req->skb, urb);
		err = btusb_submit_req(req, urb);
		if (err) {
			BTERRORF("nonzero read status received: %d\n", req->urb[urb]->status);
			req->btusb->net_stats.tx_errors++;
			req->btusb->net_stats.tx_aborted_errors++;
			skb_queue_head(&req->queue, req->skb);
			req->skb = NULL;
			goto exit;
		}
		req->btusb->net_stats.tx_packets++;
		req->btusb->net_stats.tx_bytes += req->urb[urb]->transfer_buffer_length;
		skb_pull(req->skb, req->urb[urb]->transfer_buffer_length);
		if (!req->skb->len) {
			dev_kfree_skb_any(req->skb);
			req->skb = NULL;
		}
	}

exit:
	DBFEXIT;
	return err;
}

static int btusb_net_xmit_start(struct sk_buff *skb, struct net_device *net_dev)
{
	btusb_t		*btusb = net_dev->priv;
	btusb_req_t	*req = NULL;
	hci_skb_cb	*cb;
	int		res = 0;
	unsigned int	old_flags;

	btusb_lock(btusb, &old_flags);	
	DBFENTER;

	if (btusb == NULL) {
		res = -ENODEV;
		goto exit;
	}

	if (!hci_running(net_dev)) {
		BTERRORF("%s: xmit called when interface is down\n", net_dev->name);
		dev_kfree_skb(skb);
		goto exit;
	}

	if (!(btusb->flags & BTUSB_RUNNING)){
		BTERRORF("device not opened\n");
		res = -EINVAL;
		goto exit;
	}
	
	if (skb == NULL) {
		//  ready to transmit
		goto exit;
	}

	DBPRTPKT(DBXMIT, skb->data, skb->len);

	cb = (hci_skb_cb *)skb->cb;

	if (btusb->params->uart_leading_byte) {
		unsigned char	*tlp;
		tlp = (unsigned char *)skb_push(skb, 1);
		*tlp = cb->type;
	}

	switch (cb->type) {
		case HCI_CMD_PKT:
			DBPRT(DBXMIT, "send HCI CMD pkt. len=%d\n", skb->len);
			req = &btusb->ctrl_out_req;
			skb_queue_tail(&req->queue, skb);
			break;
		
		case HCI_ACL_PKT:
			DBPRT(DBXMIT, "send HCI ACL pkt. len=%d\n", skb->len);
			req = &btusb->bulk_out_req;
			skb_queue_tail(&req->queue, skb);
#if CONFIG_AFFIX_PROF
			btusb_calc_delay(2, req, NULL);
#endif
			break;

		case HCI_SCO_PKT:
			if (!(btusb->flags & BTUSB_AUDIO_OUT)) {
				BTERRORF("Audio OUT disabled\n");
				res = -EINVAL;
				goto exit;
			}
			DBPRT(DBXMIT, "send HCI SCO pkt. len=%d\n", skb->len);
			req = &btusb->isoc_out_req;
			skb_queue_tail(&req->queue, skb);
			btusb_xmit_start_isoc(req, 0);
			goto exit;
			break;

		default:
			BTERRORF("invalid HCI packet type\n");
			dev_kfree_skb_any(skb);
			goto exit;
	}
	btusb_xmit_start(req);
	net_dev->trans_start = jiffies;
exit:
	DBFEXIT;
	btusb_unlock(btusb, old_flags);
	return res;
} 

static struct net_device_stats *btusb_net_get_stats(struct net_device *net_dev)
{
	btusb_t 	*btusb = net_dev->priv;

	DBFENTER;	
	DBFEXIT;

	return &btusb->net_stats;
}

static int btusb_net_ioctl(struct net_device *dev, void *arg, int cmd)
{
        btusb_t		*btusb = dev->priv;
        int		res = 0;
	hci_struct	*hci;
	unsigned int    old_flags;

	btusb_lock(btusb, &old_flags);
	DBFENTER;

	if (btusb == NULL || btusb->signature != 0xDEADBEAF) {
		BTERRORF("invalid btusb\n");
		res = -ENODEV;
		goto exit;
	} 
	switch (cmd) {
		case SIOCHCI_SET_AUDIO:
			hci = arg;
			res = btusb_audio_mode(btusb, hci->audio_mode, &old_flags);
			break;
			
		default:
			res = -ENOIOCTLCMD;
			goto exit;
	}
exit: 
 	DBFEXIT;
 	btusb_unlock(btusb, old_flags);
        return res;
}

static inline void btusb_receive_skb(btusb_t *btusb, btusb_req_t *req, unsigned int pkt_type)
{
	struct sk_buff *skb = req->skb;
	hci_skb_cb *cb;

	if (pkt_type > HCI_EVENT_PKT) {
		BTERRORF("invalid TLP type: %#.2x\n", pkt_type);
		dev_kfree_skb_any(skb);
		req->skb = NULL;
		return;
	}
	cb = (hci_skb_cb *)skb->cb;
	cb->type = pkt_type;
	skb->mac.raw = (void *)skb->data;
	btusb->net_stats.rx_packets++;
	btusb->net_stats.rx_bytes += skb->len;
	skb->dev = btusb->net_dev;
	skb->protocol = __constant_htons(ETH_P_BLUETOOTH);
	skb->pkt_type = PACKET_HOST;
	DBPRT(DBRECV, "new HCI pkt received. type=%#.2x  len=%d\n", pkt_type, skb->len);
	hci_rx(skb);
	req->skb = NULL;
}

/*----------------------------------------------------------------------------------------------
				USB callback functions
  ----------------------------------------------------------------------------------------------
*/
static inline btusb_t *get_btusb(btusb_req_t *req, const char *function)
{
	if (req == NULL || req->btusb == NULL || req->btusb->signature != 0xDEADBEAF) {
		BTERRORF("%s - invalid btusb\n", function);
		return NULL;
	}
	return req->btusb;
}


static void btusb_out_complete(struct urb *urb)
{
	btusb_req_t	*req = urb->context;
	btusb_t		*btusb = get_btusb(req, __FUNCTION__);
	unsigned int    old_flags;

	btusb_lock(btusb, &old_flags);
	DBFENTER;
	
	if (btusb == NULL)
		goto exit;
	
	if (urb->status == 0) {
		btusb->net_stats.tx_packets++;
		if (req->skb)
			btusb->net_stats.tx_bytes += req->skb->len;
	} else {
		BTERRORF("nonzero completion status: %d\n", urb->status);
		btusb->net_stats.tx_errors++;
		btusb->net_stats.tx_aborted_errors++;
	}

#ifdef CONFIG_AFFIX_PROF
	btusb_calc_delay(1, req, NULL);
#endif

	if (req->skb) {
		dev_kfree_skb_any(req->skb);
		req->skb = NULL;
	}

	/* clear state */
	req->flags &= ~BTUSB_URB0PENDING;
	btusb_xmit_start(req);
exit:
	DBFEXIT;
	btusb_unlock(btusb, old_flags);
}

static void btusb_isoc_out_complete(struct urb *urb)
{
	btusb_req_t	*req = (btusb_req_t *)urb->context;
	btusb_t		*btusb = get_btusb(req, __FUNCTION__);
	int	 	len = urb->actual_length, id;
	unsigned int	old_flags;

	if (len == 0)
		return;

	btusb_lock(btusb, &old_flags);
	DBFENTER;

	if (btusb == NULL)
		goto exit;

	DBPRT(DBXMIT, "wrote: %d, status: %d\n", len, urb->status);
	
#ifdef CONFIG_AFFIX_PROF
	btusb_calc_delay(3, req, NULL);
#endif

	for (id = 0; id < BTUSB_URBPOOL; id++) {
		if (urb == req->urb[id])
			break;
	}
	if (id >= BTUSB_URBPOOL) {
		BTERRORF("Unknown URB\n");
		goto exit;
	}

	if (!(btusb->flags&BTUSB_RUNNING && btusb->flags&BTUSB_AUDIO_OUT)) {
		BTERRORF("network interface is down\n");
		goto exit;
	}

	if (urb->status) {
		BTERRORF("nonzero read status received: %d\n", urb->status);
		btusb->net_stats.tx_errors++;
		btusb->net_stats.tx_aborted_errors++;
	} else {
		btusb->net_stats.tx_packets++;
		if (req->skb)
			btusb->net_stats.tx_bytes += req->skb->len;
	}

	if (!len) {
		BTERRORF("zero length written\n");
	}

	req->flags &= ~(1 << id);
	btusb_xmit_start_isoc(req, 1 << id);

exit:
	DBFEXIT;
	btusb_unlock(btusb, old_flags);
}


static void btusb_int_in_complete(struct urb *urb)
{
	btusb_req_t	*req = (btusb_req_t *)urb->context;
	btusb_t		*btusb = get_btusb(req, __FUNCTION__);
	int 		len = urb->actual_length;
	void		*dst;
	void		*src;
	unsigned int	old_flags;

	btusb_lock(btusb, &old_flags);
	DBFENTER;

	if (btusb == NULL)
		goto exit;

	if (!(btusb->flags & BTUSB_RUNNING)){
		BTERRORF("network interface is down\n");
		goto exit;
	}

	if (urb->status) {
		BTERRORF("nonzero int status received: %d\n", urb->status);
		goto exit;
	}

	if (!len) {
		BTERRORF("zero length interrupt pkt\n");
		goto exit;
	}

	DBPRT(DBRECV, "interrupt pkt. len=%d\n", len);
	DBPRTPKT(DBRECV, urb->transfer_buffer, len);
	
	if (req->skb == NULL) {	/* new packet */
		len -= btusb->params->uart_leading_byte;
		if (len < EVENT_HDR_SIZE) {
			BTERRORF("too small int packet\n");
			goto exit;
		}
		req->pkt_len = ((unsigned char *)urb->transfer_buffer)
					[1 + btusb->params->uart_leading_byte]
				 + EVENT_HDR_SIZE;
		req->skb = dev_alloc_skb(req->pkt_len);
		if (req->skb == NULL) {
			BTERRORF("dev_alloc_skb failed\n");
			goto exit;
		}
		dst = skb_put(req->skb, req->pkt_len);
		src = (char *)urb->transfer_buffer + btusb->params->uart_leading_byte;
		req->act_data_len = 0;
	} else {	
		dst = &req->skb->data[req->act_data_len];
		src = urb->transfer_buffer;
	}
	
	if (req->act_data_len + len > req->pkt_len) {
		BTERRORF("USB pkt too big\n");
		dev_kfree_skb_any(req->skb);
		req->skb = NULL;
		goto exit;
	}
	memcpy(dst, src, len);
	req->act_data_len += len;
	if (req->act_data_len == req->pkt_len)
		btusb_receive_skb(btusb, req, HCI_EVENT_PKT);	/*  pass skb to network layer above us */

exit:
	DBFEXIT;
	btusb_unlock(btusb, old_flags);
}

static void btusb_bulk_in_complete(struct urb *urb)
{
	btusb_req_t	*req = (btusb_req_t *)urb->context;
	btusb_t		*btusb = get_btusb(req, __FUNCTION__);
	unsigned char	*dst, *src;
	int	 	len = urb->actual_length, id;
	unsigned int	old_flags;

	btusb_lock(btusb, &old_flags);
	DBFENTER;

	if (btusb == NULL)
		goto exit;

	for (id = 0; id < BTUSB_URBPOOL; id++) {
		if (urb == req->urb[id])
			break;
	}
	if (id >= BTUSB_URBPOOL) {
		BTERRORF("Unknown URB\n");
		goto exit;
	}

	if (!(btusb->flags & BTUSB_RUNNING)){
		BTERRORF("network interface is down\n");
		goto exit;
	}

	if (urb->status) {
		BTERRORF("nonzero read status received: %d\n", urb->status);
		if (++btusb->read_fail_cnt == MAX_READ_FAIL_COUNT) {
			BTERRORF("Maximum read failure count reached. Give up.\n");
			goto exit;
		}
		goto resubmit;
	}
	btusb->read_fail_cnt = 0;

	if (!len) {
		BTERRORF("zero length read\n");
		goto resubmit;
	}

	DBPRT(DBRECV, "bulk in pkt. len=%d\n", len);
	DBPRTPKT(DBRECV, urb->transfer_buffer, len);

	src = (unsigned char *)urb->transfer_buffer;
	while (len) {	
		int	copy_bytes;
		
		if (req->skb == NULL) {	/* new packet */
			if (btusb->params->uart_leading_byte) {
				len--;
				src++;
			}
			if (len < ACL_HDR_SIZE) {
				BTERRORF("too small read packet\n");
				btusb->net_stats.rx_dropped++;
				goto resubmit;
			}
			req->pkt_len = le16_to_cpu(*(unsigned short *)(&(src[2]))) + ACL_HDR_SIZE;
			req->skb = dev_alloc_skb(req->pkt_len);
			if (req->skb == NULL) {
				BTERRORF("dev_alloc_skb failed\n");
				btusb->net_stats.rx_dropped++;
				goto resubmit;
			}
			dst = skb_put(req->skb, req->pkt_len);
			req->act_data_len = 0;
		} else {	
			dst = &req->skb->data[req->act_data_len];
		}
		copy_bytes = btmin(req->pkt_len - req->act_data_len, len);
		memcpy(dst, src, copy_bytes);
		req->act_data_len += copy_bytes;
		if (req->act_data_len == req->pkt_len)
			btusb_receive_skb(btusb, req, HCI_ACL_PKT);	/*  pass skb to network layer above us */
		src += copy_bytes;
		len -= copy_bytes;
	}			
resubmit:
	urb->dev = btusb->usb_dev;
	if (btusb_submit_req(req, id)) {
		BTERRORF("resubmit read failed\n");
	}
exit:
	DBFEXIT;
	btusb_unlock(btusb, old_flags);
}


static void btusb_isoc_in_complete(struct urb *urb)
{
	btusb_req_t	*req = (btusb_req_t *)urb->context;
	btusb_t		*btusb = get_btusb(req, __FUNCTION__);
	unsigned char	*dst, *src;
	int	 	len = urb->actual_length, id, i;
	unsigned int	old_flags;

#ifdef CONFIG_AFFIX_PROF
	{
		static long int	time = 0;

		if (jiffies - time > HZ) {
			time = jiffies;
			if (len == 0)
				BTINFO("zero packet\n");
		}
	}
#endif

	//if (len == 0)
	//	return;

	btusb_lock(btusb, &old_flags);
	DBFENTER;

	if (btusb == NULL)
		goto exit;

	for (id = 0; id < BTUSB_URBPOOL; id++) {
		if (urb == req->urb[id])
			break;
	}
	if (id >= BTUSB_URBPOOL) {
		BTERRORF("Unknown URB\n");
		goto exit;
	}

	if (!(btusb->flags&BTUSB_RUNNING && btusb->flags&BTUSB_AUDIO_IN)) {
		BTERRORF("network interface is down\n");
		goto exit;
	}

	if (urb->status) {
		BTERRORF("nonzero read status received: %d\n", urb->status);
		if (++btusb->read_fail_cnt == MAX_READ_FAIL_COUNT) {
			BTERRORF("Maximum read failure count reached. Give up.\n");
			goto exit;
		}
		goto resubmit;
	}
	btusb->read_fail_cnt = 0;

	if (!len) {
		//BTERRORF("zero length read\n");
		goto resubmit;
	}
	
	DBPRT(DBRECV, "isoc in pkt. total len=%d\n", len);

	for (i = 0; i < urb->number_of_packets; i++) {
		if (urb->iso_frame_desc[i].status) {
			BTINFOF("frame %u status %d\n", i, urb->iso_frame_desc[i].status);
			continue;
		}
		src = (unsigned char *)urb->transfer_buffer + urb->iso_frame_desc[i].offset;
		len = urb->iso_frame_desc[i].actual_length;

		isoc_count += len;

		DBPRT(DBRECV, "isoc in pkt. len=%d\n", len);
		DBPRTPKT(DBRECV, urb->transfer_buffer, len);

		while (len) {	
			int	copy_bytes;

			if (req->skb == NULL) {	/* new packet */
				if (btusb->params->uart_leading_byte) {
					len--;
					src++;
				}
				if (len < SCO_HDR_SIZE) {
					BTERRORF("too small read packet\n");
					btusb->net_stats.rx_dropped++;
					goto resubmit;
				}
				req->pkt_len = *(unsigned char *)(&(src[2])) + SCO_HDR_SIZE;
				req->skb = dev_alloc_skb(req->pkt_len);
				if (req->skb == NULL) {
					BTERRORF("dev_alloc_skb failed\n");
					btusb->net_stats.rx_dropped++;
					goto resubmit;
				}
				dst = skb_put(req->skb, req->pkt_len);
				req->act_data_len = 0;
			} else {	
				dst = &req->skb->data[req->act_data_len];
			}
			copy_bytes = btmin(req->pkt_len - req->act_data_len, len);
			memcpy(dst, src, copy_bytes);
			req->act_data_len += copy_bytes;
			if (req->act_data_len == req->pkt_len) {
#ifdef CONFIG_AFFIX_PROF
				btusb_calc_delay(4, req, NULL);
#endif
				btusb_receive_skb(btusb, req, HCI_SCO_PKT);	/*  pass skb to network layer above us */
			}
			src += copy_bytes;
			len -= copy_bytes;
		}
	}
resubmit:
	urb->dev = btusb->usb_dev;
	if (btusb_submit_req(req, id)) {
		BTERRORF("resubmit read failed\n");
	}
exit:
	DBFEXIT;
	btusb_unlock(btusb, old_flags);
}

/*----------------------------------------------------------------------------------------------
				USB init and shutdown functions
  ----------------------------------------------------------------------------------------------
*/
static void *btusb_probe(struct usb_device *usb_dev, unsigned int ifnum,
			 	  const struct usb_device_id *id)
{
	btusb_t		*btusb = NULL;
	int		btnum;
	int		i, s, e, ep_found, vendor_id, device_id;
	struct usb_interface_descriptor	*iface;
	
	static struct usb_device	*dev = NULL;

	DBFENTER;
	
	DBPRT(DBINIT, "dev: %p, usb_dev: %p\n", dev, usb_dev);
		
	for (btnum = 0; btnum < NUM_OF_USB_BT_DEV; btnum++) {
		vendor_id = usb_bt_list[btnum].vendor_id;
		device_id = usb_bt_list[btnum].device_id;
		if( (vendor_id == usb_dev->descriptor.idVendor &&
		    device_id == usb_dev->descriptor.idProduct) || 
				(vendor_id == 0 && device_id == 0) )
			break;
	}
	if (btnum == NUM_OF_USB_BT_DEV)	{
		BTINFOF("bluetooth device not supported\n");
		goto failed;
	}

	if (ifnum != 0) {
		DBPRT(DBINIT, "ignore interface %d\n", ifnum);
		goto failed;
	}
	
#define EP(x) (iface->endpoint[x])

#ifdef CONFIG_AFFIX_DEBUG
	DBPRT(DBCTRL, "Number of Interfaces: %d\n", usb_dev->actconfig->bNumInterfaces);
	for (i = 0; i < usb_dev->actconfig->bNumInterfaces; i++) {
		DBPRT(DBCTRL, "   Interface: %d\n", i);
		DBPRT(DBCTRL, "   Number of altsettings: %d\n", usb_dev->actconfig->interface[i].num_altsetting);
		for (s = 0; s < usb_dev->actconfig->interface[i].num_altsetting; s++) {
			iface = &usb_dev->actconfig->interface[i].altsetting[s];
			DBPRT(DBCTRL, "      Altsetting: %d\n", s);
			DBPRT(DBCTRL, "      Number of endpoints: %d\n", iface->bNumEndpoints);
			for (e = 0; e < iface->bNumEndpoints; e++) {
				DBPRT(DBCTRL, "        EndpointAddress: %#x, Attributes: %#x, Interval: %d\n",
				EP(e).bEndpointAddress, EP(e).bmAttributes, EP(e).bInterval); 

			}
		}
	}
#endif
	btusb = kmalloc(sizeof(btusb_t), GFP_KERNEL);
	if (btusb == NULL) {
		BTERRORF("failed to reserve memory for dev\n");
		goto failed;
	}
	memset(btusb, 0, sizeof(btusb_t));

	/* initialize some data */
	skb_queue_head_init(&btusb->bulk_out_req.queue);
	skb_queue_head_init(&btusb->isoc_out_req.queue);
	skb_queue_head_init(&btusb->ctrl_out_req.queue);

	iface = &usb_dev->actconfig->interface[0].altsetting[0];
	ep_found = 0;
	DBPRT(DBINIT, "Checking Main interface\n");
	for (i = 0; i < iface->bNumEndpoints; i++) {
		switch (EP(i).bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) {
			case USB_ENDPOINT_XFER_BULK:
				if ((EP(i).bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN) {
					btusb->bulk_in_req.pipe = usb_rcvbulkpipe(usb_dev, EP(i).bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
					ep_found |= 0x01;
				} else {
					btusb->bulk_out_req.pipe = usb_sndbulkpipe(usb_dev, EP(i).bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
					ep_found |= 0x02;
				}
				break;
			case USB_ENDPOINT_XFER_INT:
				if ((EP(i).bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN) {
					btusb->int_in_req.pipe = usb_rcvintpipe(usb_dev, EP(i).bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
					ep_found |= 0x04;
				}
				break;
			default:
		}
	}
	if (ep_found != 0x07) {
		BTERRORF("invalid endpoint address / attribute\n");
		goto failed;
	}
	btusb->ctrl_out_req.pipe = usb_sndctrlpipe(usb_dev, 0);
	//DBPRT(DBINIT, "bulk_in: %#x, bulk_out: %#x, int_in: %#x\n",
	//		btusb->bulk_in_req.ep, btusb->bulk_out_req.ep, btusb->int_in_req.ep);

	/* now check if there is ISOC interface */
	if (usb_dev->actconfig->bNumInterfaces > 1) {
		DBPRT(DBINIT, "Checking ISOC\n");
		iface = &usb_dev->actconfig->interface[1].altsetting[0];
		for (i = 0; i < iface->bNumEndpoints; i++) {
			switch (EP(i).bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) {
				case USB_ENDPOINT_XFER_ISOC:
					if ((EP(i).bEndpointAddress & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN) {
						btusb->isoc_in_req.pipe = usb_rcvisocpipe(usb_dev, EP(i).bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
						ep_found |= 0x10;
					} else {
						btusb->isoc_out_req.pipe = usb_sndisocpipe(usb_dev, EP(i).bEndpointAddress & USB_ENDPOINT_NUMBER_MASK);
						ep_found |= 0x20;
					}
					break;
				default:
			}
		}
	}

#undef EP

	btusb->signature = 0xDEADBEAF;
	btusb->usb_dev = usb_dev;
	spin_lock_init(&btusb->btusb_lock);
	btusb->params = &usb_bt_list[btnum];
	if ((btusb->net_dev = affix_alloc_btdev()) == NULL)
		goto failed;
	btusb->net_dev->priv = btusb;
	btusb->net_dev->open = btusb_net_open;
	btusb->net_dev->stop = btusb_net_stop;
	btusb->net_dev->hard_start_xmit = btusb_net_xmit_start;
	btusb->net_dev->get_stats = btusb_net_get_stats;
	btusb->net_dev->do_ioctl = (void *)btusb_net_ioctl;
	btusb->net_dev->hard_header_len = btusb->params->uart_leading_byte;

	if ((ep_found & 0xF0) == 0x30) {
		DBPRT(DBINIT, "Found ISOC interface\n");
		usb_driver_claim_interface(&btusb_driver, &usb_dev->actconfig->interface[1],btusb);
		usb_set_interface(usb_dev, 1, 2);
		btusb->isoc = 1;
		DBPRT(DBINIT, "ISOC interface configured\n");
	}

	SET_MODULE_OWNER(btusb->net_dev);

	if (register_netdev(btusb->net_dev) != 0) {
		BTERRORF("register_netdev failed\n");
		goto failed2;
	}

	BTINFOF("affix_usb bound to device\n");

	DBFEXIT;
	return btusb;

failed2:
	kfree(btusb->net_dev);
failed:
	if (btusb) 
		kfree(btusb);
	return NULL;
}

static void btusb_disconnect(struct usb_device *usb_dev, void *ptr)
{
	btusb_t *btusb = (btusb_t *)ptr;

	DBFENTER;

	if (btusb == NULL || btusb->signature != 0xDEADBEAF) {
		BTERRORF("invalid btusb\n");
		DBFEXIT;
		return;
	}
	if (btusb->isoc)
		usb_driver_release_interface(&btusb_driver, &usb_dev->actconfig->interface[1]);

	rtnl_lock();
	dev_close(btusb->net_dev);
	unregister_netdevice(btusb->net_dev);
	rtnl_unlock();

	btusb_release_req(&btusb->int_in_req);
	btusb_release_req(&btusb->ctrl_out_req);
	btusb_release_req(&btusb->bulk_in_req);
	btusb_release_req(&btusb->bulk_out_req);
	btusb_release_req(&btusb->isoc_in_req);
	btusb_release_req(&btusb->isoc_out_req);

	skb_queue_purge(&btusb->bulk_out_req.queue);
	skb_queue_purge(&btusb->isoc_out_req.queue);
	skb_queue_purge(&btusb->ctrl_out_req.queue);

	kfree(btusb);

	BTINFOF("BTUSB driver disconnected from device\n");
	DBFEXIT;
}

/*----------------------------------------------------------------------------------------------
				Module init and shutdown functions
  ----------------------------------------------------------------------------------------------
*/
int __init btusb_init(void)
{
	int result;

	DBFENTER;
	result = usb_register(&btusb_driver);
	if (result < 0) {
		BTERRORF("usb_register failed for the USB bluetooth driver. Error number %d\n", result);
		DBFEXIT;
		return -1;
	}

	BTINFOF("affix_usb registered\n");

	DBFEXIT;
	return 0;
}


void __exit btusb_exit(void)
{
	DBFENTER;

	usb_deregister(&btusb_driver);
	BTINFOF("affix_usb deregistered\n");

	DBFEXIT;
}

/* ------------------------------------------------------------------------------------------ */

MODULE_AUTHOR("Imre Deak/Dmitry Kasatkin");
MODULE_DESCRIPTION("Affix USB driver for Bluetooth devices");
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 15)
MODULE_LICENSE("GPL");
#endif

/*  If we are resident in kernel we will call module_init manually. */
#ifdef MODULE
module_init(btusb_init);
module_exit(btusb_exit);
#endif

EXPORT_NO_SYMBOLS;


