#ifndef _AFFIX_BLUEZ_H
#define _AFFIX_BLUEZ_H

#define FILEBIT	DBDRV

#include <linux/netdevice.h>
#include <linux/rtnetlink.h>

#include <affix/btdebug.h>
#include <affix/hci.h>
#include <affix/bluetooth.h>

/* HCI device states */
enum hci_dev_states_t
{
	HCI_INIT = 0,
	HCI_UP,
	HCI_RUNNING,
};

/* HCI device types */
#define HCI_UART 	0
#define HCI_USB		1
#define HCI_VHCI	2
#define HCI_PCCARD	3

#define BLUEZ_SKB_RESERVE	8

#define HCI_MAX_FRAME_SIZE	2048

/* HCI Packet types */
#define HCI_COMMAND_PKT		0x01
#define HCI_ACLDATA_PKT 	0x02
#define HCI_SCODATA_PKT 	0x03
#define HCI_EVENT_PKT		0x04

/* --------  HCI Packet structures  -------- */
#define HCI_TYPE_LEN		1
#define HCI_COMMAND_HDR_SIZE 	3
#define HCI_EVENT_HDR_SIZE 	2
#define HCI_ACL_HDR_SIZE 	4
#define HCI_SCO_HDR_SIZE 	3

typedef struct {
	__u16 	opcode;		/* OCF & OGF */
	__u8 	plen;
} __attribute__ ((packed))	hci_command_hdr;

typedef struct {
	__u8 	evt;
	__u8 	plen;
} __attribute__ ((packed))	hci_event_hdr;

typedef struct {
	__u16 	handle;		/* Handle & Flags(PB, BC) */
	__u16 	dlen;
} __attribute__ ((packed))	hci_acl_hdr;

typedef struct {
	__u16 	handle;
	__u8 	dlen;
} __attribute__ ((packed))	hci_sco_hdr;


/* ----- SKB helpers ----- */
struct bluez_skb_cb {
	int    incomming;
};
#define bluez_cb(skb)	((struct bluez_skb_cb *)(skb->cb)) 


static inline struct sk_buff *bluez_skb_alloc(unsigned int len, int how)
{
	struct sk_buff *skb;

	if ((skb = alloc_skb(len + BLUEZ_SKB_RESERVE, how))) {
		skb_reserve(skb, BLUEZ_SKB_RESERVE);
		bluez_cb(skb)->incomming  = 0;
	}
	return skb;
}



/* ----- HCI Devices ----- */

struct hci_dev_stats {
	__u32 err_rx;
	__u32 err_tx;
	__u32 cmd_tx;
	__u32 evt_rx;
	__u32 acl_tx;
	__u32 acl_rx;
	__u32 sco_tx;
	__u32 sco_rx;
	__u32 byte_rx;
	__u32 byte_tx;
};

struct hci_dev {
	char		name[8];
	void		*driver_data;
	__u32	 	flags;
	__u8	 	type;
	struct hci_dev_stats 	stat;
	
	struct sk_buff_head 	txq;
	struct tasklet_struct 	tx_task;

	int (*open)(struct hci_dev *hdev);
	int (*close)(struct hci_dev *hdev);
	int (*flush)(struct hci_dev *hdev);
	int (*send)(struct sk_buff *skb);

	void (*destruct)(struct hci_dev *hdev);
	int (*ioctl)(struct hci_dev *hdev, unsigned int cmd, unsigned long arg);

	/* Affix */
	struct net_device	*dev;
	struct net_device_stats	net_stats;
	int			flowmask;
};

static inline int hci_dev_init(struct net_device *dev)
{
	struct hci_dev	*hdev = dev->priv;

	DBFENTER;
	if (test_and_set_bit(HCI_INIT, &(hdev->flags)))
		return 0;
	DBFEXIT;
	return 0;
}

static inline void hci_dev_uninit(struct net_device *dev)
{
	struct hci_dev	*hdev = dev->priv;

	DBFENTER;
	if (!test_and_clear_bit(HCI_INIT, &(hdev->flags)))
		return;
	DBFEXIT;
}

static inline int hci_dev_open(struct net_device *dev)
{
	int		err;
	struct hci_dev	*hdev = dev->priv;

	DBFENTER;
	if (test_and_set_bit(HCI_UP, &(hdev->flags)))
		return 0;

	hdev->flowmask = HCI_CMD_ON | HCI_ACL_ON | HCI_SCO_ON;

	DBPRT(DBINIT, "opening device ...\n");
	err = hdev->open(hdev);
	DBPRT(DBINIT, "device opened: %d\n", err);
	if (err)
		return err;
	hci_start_queue(hdev->dev, hdev->flowmask);

	DBFEXIT;
	return 0;
}

static inline int hci_dev_stop(struct net_device *dev)
{
	struct hci_dev	*hdev = dev->priv;
	int		err;

	if (!test_and_clear_bit(HCI_UP, &(hdev->flags)))
		return 0;

	hci_stop_queue(hdev->dev, hdev->flowmask);
	tasklet_kill(&hdev->tx_task);
	skb_queue_purge(&hdev->txq);
	err = hdev->close(hdev);

	return err;
}

static inline int hci_dev_ioctl(struct net_device *dev, void *arg, int cmd)
{
	DBFENTER;
	switch (cmd) {
	default:
		return -ENOIOCTLCMD;
	}
	DBFEXIT;
	return 0;
}

static inline struct net_device_stats *hci_dev_get_stats(struct net_device *dev)
{
	struct hci_dev	*hdev = dev->priv;

	return &hdev->net_stats;
}

static inline int hci_dev_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct hci_dev	*hdev = dev->priv;
	hci_skb_cb	*cb = (hci_skb_cb*)skb->cb;

	DBFENTER;
	if (!hci_running(dev)) {
		dev_kfree_skb(skb);
		return 0;
	}

	if (hci_queue_stopped(dev)) {
		return -EBUSY;
	}

	if (skb == NULL) {
		//dev_tint(dev);
		return 0;
	}

	dev->trans_start = jiffies;
	skb->dev = (struct net_device*)hdev;
	skb->pkt_type = cb->type;

	hci_stop_queue(dev, hdev->flowmask);
	skb_queue_tail(&hdev->txq, skb);
	tasklet_schedule(&hdev->tx_task);

	DBFEXIT;
        return 0;
}

static inline void hci_tx_task(unsigned long arg)
{
	struct hci_dev *hdev = (struct hci_dev *) arg;
	struct sk_buff *skb;
	int		err;

	DBFENTER;
	while ((skb = skb_dequeue(&hdev->txq))) {
		err = hdev->send(skb);
		DBPRT(DBXMIT, "packet sent to driver: %d\n", err);
		if (err)
			break;
	}
	hci_wake_queue(hdev->dev, hdev->flowmask);
	DBFEXIT;
}

/* Register HCI device */
static inline int hci_register_dev(struct hci_dev *hdev)
{
	struct net_device	*dev;
	int			err;

	DBFENTER;

	dev = affix_alloc_btdev();
	if (dev == NULL)
		return -ENOMEM;

	/* bluez stuff */
	tasklet_init(&hdev->tx_task, hci_tx_task, (unsigned long) hdev);
	skb_queue_head_init(&hdev->txq);
	memset(&hdev->stat, 0, sizeof(struct hci_dev_stats));
	hdev->flags = 0;
	hdev->dev = dev;

	/* Affix stuff */
	dev->priv = hdev;	/* set private pointer */
	dev->init = hci_dev_init;
	dev->uninit = hci_dev_uninit;
	dev->open = hci_dev_open;
	dev->stop = hci_dev_stop;
	dev->do_ioctl = (void*)hci_dev_ioctl;
	dev->get_stats = hci_dev_get_stats;
	dev->hard_start_xmit = hci_dev_xmit;
	dev->hard_header_len = BLUEZ_SKB_RESERVE;

	SET_MODULE_OWNER(dev);

	rtnl_lock();
	err = register_netdevice(dev);
	rtnl_unlock();

	DBFEXIT;

	return (err)? err : dev->ifindex;
}

/* Unregister HCI device */
static inline int hci_unregister_dev(struct hci_dev *hdev)
{
	DBFENTER;
	rtnl_lock();
	dev_close(hdev->dev);
	unregister_netdevice(hdev->dev);
	rtnl_unlock();
	hdev->dev = NULL;
	DBFEXIT;
	return 0;
}

/* Receive frame from HCI drivers */
static inline int hci_recv_frame(struct sk_buff *skb)
{
	struct hci_dev *hdev = (struct hci_dev *) skb->dev;
	hci_skb_cb	*cb = (hci_skb_cb*)skb->cb;

	DBFENTER;

	if (!hdev || !(hdev->flags & (HCI_UP | HCI_INIT))) {
		kfree_skb(skb);
		return -1;
	}

	cb->type = skb->pkt_type;
	skb->dev = hdev->dev;
	skb->mac.raw = (void*)skb->data;
	skb->protocol = __constant_htons(ETH_P_BLUETOOTH);
	skb->pkt_type = PACKET_HOST;
	
	hci_rx(skb);	/* send to upper protocol layer */

	DBFEXIT;
	return 0;
}

#endif

