/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: hci.h,v 1.16 2002/09/04 08:57:23 kds Exp $

   Host Controller Interface

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
                Imre Deak <ext-imre.deak@nokia.com>
*/


#ifndef _HCI_KERNEL_H
#define _HCI_KERNEL_H

#include <linux/wait.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/if.h>
#include <linux/init.h>
#include <net/sock.h>

#include <asm/uaccess.h>

#include <affix/types.h>
#include <affix/hci_types.h>

#include <affix/btlist.h>
#include <affix/bluetooth.h>
#include <affix/btcore.h>

#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 17)
#define get_fast_time(a)	do_gettimeofday(a)
#endif

/******* MACRO DEFINITION ******************/

#define ENTERSTATE(ch, st)	(ch->state = st)
#define SETSTATE(ch, st)	(xchg(&ch->state, st))
#define ENTERSUBSTATE(ch, st)	(ch->substate = st)
#define STATE(ch)		(ch->state)
#define SUBSTATE(ch)		(ch->substate)

#define LV(v)	((v)!=0)

#define	HCI_CMD_ON		0x0001
#define	HCI_ACL_ON		0x0002
#define	HCI_SCO_ON		0x0004


/**************   variable declaration    ****************************/

extern struct proc_dir_entry	*proc_affix;

extern btlist_head_t		hcicons;
extern btlist_head_t		listeners;
extern btlist_head_t		hcidevs;

/*
 * important types
 */
typedef struct hci_struct	hci_struct;
typedef struct hci_listener_t	hci_listener_t;
typedef struct hci_con		hci_con;

/*
 * remote device record
 */
typedef struct {
	struct list_head	q;	/* List Support */

	atomic_t	refcnt;		/* number of users */
	int		state;		/* device state */
#define BTDEV_AUTHPENDING	0x0001

	BD_ADDR		bda;		/* Device Address - id */
	int		flags;		/* valid fields */
#define	NBT_INQUIRY	0x01
#define NBT_PIN		0x02
#define NBT_KEY		0x04
	
	/* like INQUIRY_ITEM */
	__u8		PS_Repetition_Mode;
	__u8		PS_Period_Mode;
	__u8		PS_Mode;
	__u32		Class_of_Device;
	__u16		Clock_Offset;
	
	/* Security Specific */
	__u8		Key_Type;
	__u8		Link_Key[16];
	__u8		PIN_Code_Length;
	__u8		PIN_Code[16];

	int		paired;
	int		trusted;
} btdev_struct;


/* Listener definition */

struct hci_listener_t {
	struct list_head	q;
	
	__u64			event_mask;	/* events to listen */
	void			(*handler)(hci_listener_t *evl, __u8 event, struct sk_buff *skb);
	hci_struct		*hci;	/* listen this device */
	void			*data;
	
	wait_queue_head_t	sleep;	/* waiting processes */
	struct sk_buff_head	queue;	/* for incomming events */

	__u16			opcode;

	int			err;
};

hci_listener_t *hci_alloc_listener(void);

static inline void hci_free_listener(hci_listener_t *evl)
{
	skb_queue_purge(&evl->queue);
	kfree(evl);
}

static inline void hci_add_listener(hci_listener_t *evl)
{
	if (!btl_linked(evl))
		btl_add_tail(&listeners, evl);
}

static inline void hci_remove_listener(hci_listener_t *evl)
{
	btl_unlink(&listeners, evl);
}

static inline void __hci_remove_listener(hci_listener_t *evl)
{
	__btl_unlink(&listeners, evl);
}

static inline int listener_error(hci_listener_t *evl)
{
	int err=xchg(&evl->err,0);
	return err;
}


/*
 * Control structure for HCI interface
 */

enum hci_state {
	HCI_START = 0
};


#define	DEV_HCI(dev)	((hci_struct*)(dev+1))

struct hci_struct {
	struct list_head	q;	/* for queueing */

	int			state;	/* device state */
	__u16			manfid;
	struct net_device	*dev;
	int			hdrlen;
	char			name[IFNAMSIZ];
	int			hd;		/* HCI descriptor */
	BD_ADDR			bda;

	/* default device attributes */
	__u16			pkt_type;	/* packet type at connection */
	__u8			allow_switch;	/* connection */
	__u8			role;		/* role at connection */
	int			sec_mode;	/* security mode */

	/* scheduler stuff */
	unsigned long		flowmask;
	unsigned long		devflowmask;
	struct sk_buff*		(*select_packet)(hci_struct *hci, __u8 type, hci_con **con);
	__u16			acl_plen;
	__u8			sco_plen;
	__u8			def_cmd_num;
	__u16			def_acl_num;
	__u16			def_sco_num;
	__u8			cmd_num;
	__u16			acl_num;
	__u16			sco_num;
	/* new rx stuff */
	struct tasklet_struct 	rx_task;
	struct sk_buff_head	rx_queue;

	/* command processing */
	struct sk_buff_head	cmd_queue;
	btlist_head_t		hci_locks;	/* locked commands */
	INQUIRY_ITEM		target_dev;	/* connecting bda */

	/* audio */
	__u16			audio_setting;
	int			audio_mode;

#ifdef DTL1_PKT_SCHED_FIX
	int			fixit;
	struct sk_buff		*delayed_acl_skb;
#endif
#ifdef CONFIG_AFFIX_PROF
	struct timeval		acl_start;
	suseconds_t		acl_delay;
	int			acl_trials;
#endif
};

extern struct notifier_block *affix_chain;

#define BTDEV_AUTH_COMPLETE	0x0100

static inline void hci_start(hci_struct *hci)
{
	set_bit(HCI_START, &hci->state);
	notifier_call_chain(&affix_chain, NETDEV_UP, hci);
}
static inline void hci_stop(hci_struct *hci)
{
	clear_bit(HCI_START, &hci->state);
	notifier_call_chain(&affix_chain, NETDEV_DOWN, hci);
}

static inline int hci_active(hci_struct *hci)
{
	return test_bit(HCI_START, &hci->state);
}

struct hci_con {
	struct list_head	q;		/* for queueing */

	con_state		state;		/* object state */
	atomic_t		refcnt;		/* number of users */
	struct timer_list	timer;
	int			id;

	int		retry;			/* retry counter */

	hci_struct	*hci;			/* device */
	int		hdrlen;

	BD_ADDR		bda;			/* remote side address */
	CHANDLE		chandle;
	__u8		link_type;
	__u8		encryp_mode;
	__u16		pkt_type;

	struct sk_buff		*rx_skb;	/* for reassembling */

	MTU			mtu;
	unsigned long		prior;		/* 0 - highest */
	struct sk_buff_head	tx_queue;

	/* security block */
	btdev_struct	*btdev;

};

/*
 * skb->cb control block data structure
 */
typedef struct {
	int		type;
	hci_con		*con;
	int		pb;
} hci_skb_cb;



/**************   functions declaration    ****************************/

/* Connection management */
hci_con *hcc_lookup_chandle(hci_struct *hci, CHANDLE chandle);
hci_con *hcc_lookup_acl(hci_struct *hci, BD_ADDR *bda);
hci_con *hcc_lookup_id(int id);
hci_con *hcc_alloc(void);
hci_con *hcc_create(hci_struct *hci, BD_ADDR *bda);
void hcc_destroy(hci_con *con);
void hcc_bind(hci_con *con, hci_struct *hci);

static inline void hcc_hold(hci_con *con)
{
	atomic_inc(&con->refcnt);
}

void hcc_put(hci_con *con);

/*  Disconnect an unused HCI connection after this time.
 *  A new connection request will cancel the disconnection
 */
#define HCC_DEFER_DISC_TIMEOUT	(20 * HZ)
#define HCC_CONN_TIMEOUT	(50 * HZ)
#define HCC_DISC_TIMEOUT	(5 * HZ)
#define HCC_AUTH_TIMEOUT	HCC_CONN_TIMEOUT

static inline void hcc_start_timer(hci_con *con, unsigned long timeout)
{
	if (!del_timer(&con->timer))
		hcc_hold(con);
	mod_timer(&con->timer, jiffies + timeout);
}

/*  called when someone wants to use the HCI connection before the timer above
 *  would expire.
 */
static inline void hcc_stop_timer(hci_con *con)
{
	if (del_timer(&con->timer))
		hcc_put(con);
}

/*
 * HCI management 
 */
hci_struct *hci_select(void);
hci_struct *hci_create(struct net_device *dev);
hci_struct *__hci_lookup_bda(BD_ADDR *bda);
hci_struct *hci_lookup_bda(BD_ADDR *bda);
hci_struct *__hci_lookup(int hd);
hci_struct *hci_lookup(int hd);

static inline void hci_hold(hci_struct *hci)
{
	dev_hold(hci->dev);
}

static inline void hci_put(hci_struct *hci)
{
	dev_put(hci->dev);
}

/* L2CAP support layer */
struct sk_buff *lp_reassemble_packet(hci_con *con, __u8 first, struct sk_buff *skb);
int lp_receive_data(hci_con *con, __u8 first, __u8 bc, struct sk_buff *skb);
int lp_connect_ind(hci_con *con);
int lp_connect_cfm(hci_con *con, __u8 status);
int lp_disconnect_ind(hci_con *con);
int lp_connect_req(hci_struct *hci, BD_ADDR *bda, hci_con **con);
int lp_add_sco(hci_struct *hci, BD_ADDR *bda, hci_con **ret);
int lp_connect_rsp(hci_con *con, __u8 status);
int lp_disconnect_req(hci_con *con, __u8 reason);
int lp_auth_req(hci_con *con);

/* transmitting part*/
int hci_send_data(hci_con *con, struct sk_buff *skb);
int hci_send_voice(hci_con *con, struct sk_buff *skb);
int hci_send_cmd(hci_struct *hci, __u16 opcode, __u8 len, void *p);
int hci_send(hci_con *con, struct sk_buff *skb);


/* receving part */
int hci_receive_event(hci_struct *hci, struct sk_buff *skb);
int hci_receive_acl(hci_struct *hci, struct sk_buff *skb);
int hci_receive_data(hci_struct *hci, struct sk_buff *skb);
void hci_rx_task(unsigned long arg);


/* network device susbsystem */
struct net_device *affix_alloc_btdev(void);
void hci_sched_init(struct net_device *dev);
void bda2str(char *str, BD_ADDR *p);

/*event listeners */
int hci_wait_event(hci_listener_t *evl, int in, signed long timeout);

static inline int test_bit64(int n, __u64 *mask)
{
	__u64	a = (1 << n);

	return (*mask & a) != 0;
}
	
/* inline functions */

static inline int hci_running(struct net_device *dev)
{
	return netif_running(dev);
}

static inline int hci_rx(struct sk_buff *skb)
{
#ifdef CONFIG_AFFIX_OLDRX 
	return netif_rx(skb);
#else

	hci_struct	*hci = DEV_HCI(skb->dev);
	skb_queue_tail(&hci->rx_queue, skb);
	tasklet_hi_schedule(&hci->rx_task);
	return 0;
#endif
}

static inline void hci_wake_queue(struct net_device *dev, unsigned long mask)
{
	hci_struct	*hci = DEV_HCI(dev);
	unsigned long	oldmask = hci->devflowmask;
	
	hci->devflowmask |= mask;
	//DBPRT(DBCTRL, "oldmask: %lx, devflowmask: %lx\n", oldmask, hci->devflowmask);
	if (hci->devflowmask ^ oldmask) {
		if (!oldmask)
			netif_wake_queue(hci->dev);
		else
			netif_schedule(hci->dev);
	}
}

static inline void hci_stop_queue(struct net_device *dev, unsigned long mask)
{
	hci_struct	*hci = DEV_HCI(dev);
	
	if (hci->devflowmask & mask) {
		/* it's running */
		hci->devflowmask &= ~mask;
		//DBPRT(DBCTRL, "devflowmask: %lx\n", hci->devflowmask);
		if (!hci->devflowmask)
			netif_stop_queue(hci->dev);
	}
}

static inline void hci_start_queue(struct net_device *dev, unsigned long mask)
{
	hci_struct	*hci = DEV_HCI(dev);
	
	hci->devflowmask |= mask;
	netif_start_queue(dev);
}

static inline int hci_queue_stopped(struct net_device *dev)
{
	return netif_queue_stopped(dev);
}

static inline void hci_device_attach(struct net_device *dev)
{
	netif_device_attach(dev);
}

static inline void hci_device_detach(struct net_device *dev)
{
	netif_device_detach(dev);
}

static inline int hci_device_present(struct net_device *dev)
{
	return netif_device_present(dev);
}

static inline int hcc_linkup(hci_con *con)
{
	return ( STATE(con) == CON_OPEN);
}

static inline struct sk_buff *hci_alloc_cmd_skb(hci_struct *hci, int size)
{
	struct sk_buff	*skb;


	skb = alloc_skb(hci->hdrlen + size, GFP_ATOMIC);
	if (skb == NULL)
		return NULL;
	skb_reserve(skb, hci->hdrlen);
	skb_put(skb, size);

	return skb;
}

static inline void bdacpy(void *rev_bda, void *bda)
{
	int	i;

	for (i = 0; i < 6; i++)
		((__u8*)rev_bda)[i] = ((__u8*)bda)[5-i];
}



/* Command lock object */
typedef struct {
	struct list_head	q;		/* for queueing */

	atomic_t		refcnt;		/* holders */
	struct semaphore	sema;		/* sleepers */

	hci_struct		*hci;
	__u16			opcode;		/* pending command */
} hci_lock_t;

static inline void hci_lock_hold(hci_lock_t *lock)
{
	atomic_inc(&lock->refcnt);
}

static inline void hci_lock_free(hci_lock_t *lock)
{
	kfree(lock);
}

hci_lock_t *hci_lock_create(hci_struct *hci, __u16 opcode);
void hci_lock_destroy(hci_lock_t *lock);

static inline void hci_lock_put(hci_lock_t *lock)
{
	if (atomic_dec_and_test(&lock->refcnt))
		hci_lock_destroy(lock);
}

static inline hci_lock_t *hci_lock(hci_struct *hci, __u16 opcode)
{
	hci_lock_t	*lock;

	lock = hci_lock_create(hci, opcode);
	if (lock)
		down(&lock->sema);

	return lock;
}

static inline void hci_unlock(hci_lock_t *lock)
{
	up(&lock->sema);
	hci_lock_put(lock);
}

int hci_msg_ready(struct sk_buff *skb);

static inline int is_kernel(void)
{
	return segment_eq(get_fs(),KERNEL_DS);
}

int __init affix_init(void);

typedef struct {
	struct module	*owner;

	kdev_t	kdev;
	int	port, irq;
	int	type;
	int	manfid, prodid;
	int	proto;
	int	flags;
	int	speed;
	int	*count;
} affix_uart_t;

struct affix_uart_operations {
	struct module	*owner;

	int (*attach)(affix_uart_t *serial);
	int (*detach)(kdev_t kdev);
	void (*suspend)(kdev_t kdev);
	void (*resume)(kdev_t kdev);
};

extern struct affix_uart_operations	affix_uart_ops;
int affix_set_uart(struct affix_uart_operations *ops);
int affix_open_uart(struct open_uart *line);
int affix_close_uart(struct open_uart *line);

int affix_register_notifier(struct notifier_block *nb);
int affix_unregister_notifier(struct notifier_block *nb);

static inline suseconds_t sub_timeval(struct timeval *t1, struct timeval *t2)
{
	time_t	dsec = (t1->tv_sec-t2->tv_sec);
	
	if (dsec != 0)
		return (1000000*dsec + t1->tv_usec - t2->tv_usec);
	else
		return (t1->tv_usec - t2->tv_usec);
}

static inline int affix_sock_rx(struct sock *sk, struct sk_buff *skb)
{
#ifdef CONFIG_AFFIX_OLDRX 
	return sock_queue_rcv_skb(sk, skb);
#else
#if 0
	/* we cannot drop ???? packet FIXME: */
	if (atomic_read(&sk->rmem_alloc) + skb->truesize >= (unsigned)sk->rcvbuf)
                return -ENOMEM;
#endif
	skb->dev = NULL;
	skb_set_owner_r(skb, sk);
	skb_queue_tail(&sk->receive_queue, skb);
	if (!sk->dead)
		sk->data_ready(sk,skb->len);
	return 0;
#endif
}

#endif
