/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: hci_cmds.h,v 1.16 2002/09/03 07:34:36 kds Exp $

   Host Controller Interface

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/		


#ifndef _HCI_COMMANDS_H
#define _HCI_COMMANDS_H

/*
 * HCI_USE_STOP_EVENT, stop_event - remove listener
 * HCI_AUTO_RELEASE - unlock command interface
 * HCI_W4_EXTRA_EVENT, extra_event - returns after receiving that event
 * event_mask - events we listening for
 * 
 */

int hci_open(char *name)
{
	int		fd, err;
	struct hci_open	cmd;

	fd = btsys_socket(PF_AFFIX, SOCK_RAW, BTPROTO_HCI);
	if (fd == -1) {
		return -1;
	}

	if (name == NULL)
		cmd.cmd = HCI_OPEN_ANY;
	else {
		cmd.cmd = HCI_OPEN_NAME;
		strncpy(cmd.name, name, IFNAMSIZ);
	}

	err = btsys_ioctl(fd, BTIOC_OPEN_HCI, &cmd);
	if (err) {
		btsys_close(fd);
		return err;
	}

	return fd;
}

int hci_open_id(int hd)
{
	int		fd, err;
	struct hci_open	cmd;

	fd = btsys_socket(PF_AFFIX, SOCK_RAW, BTPROTO_HCI);
	if (fd == -1) {
		return -1;
	}

	cmd.cmd = HCI_OPEN_ID;
	cmd.hd = hd;

	err = btsys_ioctl(fd, BTIOC_OPEN_HCI, &cmd);
	if (err) {
		btsys_close(fd);
		return err;
	}

	return fd;
}

int hci_open_mgr(void)
{
	int		fd, err;
	struct hci_open	cmd;

	fd = btsys_socket(PF_AFFIX, SOCK_RAW, BTPROTO_HCI);
	if (fd == -1) {
		return -1;
	}

	cmd.cmd = HCI_OPEN_MGR;

	err = btsys_ioctl(fd, BTIOC_OPEN_HCI, &cmd);
	if (err) {
		btsys_close(fd);
		return err;
	}

	return fd;
}

/*
 * Link Control
 */

#if !defined(__KERNEL__) || defined(CONFIG_AFFIX_PAN)
static inline int ItemLookup(INQUIRY_ITEM *Items, __u8 Num, BD_ADDR *bda)
{
	int	i;
	for (i = 0; i < Num; i++)
		if (memcmp(&Items[i].bda, bda, 6) == 0)
			return i;
	return -1;
}


int HCI_Inquiry(int fd, __u8 Inquiry_Length, __u8 Max_Num_Responses, INQUIRY_ITEM *Items, __u8 *Num_Responses)
{
	int				err, pos;
	unsigned char			buf[EVENT_MAX_SIZE];
	struct Inquiry			cmd;
	struct Inquiry_Result_Event	*ir = (void*)buf;
	struct Inquiry_Complete_Event	*ic = (void*)buf;
	struct hci_cmd_arg	cp;
	struct hci_event_arg		ep;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_INQUIRY);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.LAP = __htob24(GIAC);				// use only GIAC for now
	cmd.Inquiry_Length = Inquiry_Length;
	cmd.Max_Num_Responses = Max_Num_Responses;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK | INQUIRY_RESULT_MASK | INQUIRY_COMPLETE_MASK;
	cp.stop_event = HCI_E_INQUIRY_COMPLETE;
	cp.event = (void*)buf;	/* not used */

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cp.status)			/* Command Status */
		return cp.status;

	// now need to receive Inquiry Results until Inquiry Complete
	ep.event = (void*)buf;

	*Num_Responses = 0;
	do {
		err = hci_recv_event(fd, &ep);
		if (err < 0) {
			return err;
		}
		if (ep.event->EventCode == HCI_E_INQUIRY_RESULT) {
			int	i;
			for (i = 0; i < ir->Num_Responses; i++) {
				pos = ItemLookup(Items, *Num_Responses, &ir->Results[i].bda);
				if (pos == -1) {
					// new device
					pos = (*Num_Responses)++;
					Items[pos].bda = ir->Results[i].bda;
				}

				Items[pos].PS_Repetition_Mode = ir->Results[i].PS_Repetition_Mode;
				Items[pos].PS_Period_Mode = ir->Results[i].PS_Period_Mode;
				Items[pos].PS_Mode = ir->Results[i].PS_Mode;
				Items[pos].Class_of_Device = __btoh24(ir->Results[i].Class_of_Device);
				Items[pos].Clock_Offset = __btoh16(ir->Results[i].Clock_Offset);
			}


		}
	} while (ep.event->EventCode != HCI_E_INQUIRY_COMPLETE);

	return ic->Status;
}
#endif

int __HCI_CreateConnection(int fd, INQUIRY_ITEM *dev, __u16 Packet_Type, __u8 Allow_Role_Switch)
{
	int				err;
	struct Create_Connection	cmd;
	struct hci_cmd_arg	cp;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_CREATE_CONNECTION);
	cmd.hdr.Length = CMD_LEN(cmd);

	cmd.bda = dev->bda;
	cmd.Packet_Type = __htob16(Packet_Type);
	cmd.PS_Repetition_Mode = dev->PS_Repetition_Mode;
	cmd.PS_Mode = dev->PS_Mode;
	cmd.Clock_Offset = __htob16(dev->Clock_Offset);
	cmd.Allow_Role_Switch = Allow_Role_Switch;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK;
	cp.stop_event = HCI_E_COMMAND_STATUS;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return cp.status;
}

int __HCI_Disconnect(int fd, CHANDLE Connection_Handle, __u8 Reason)
{
	int					err;
	struct Disconnect			cmd;
	struct hci_cmd_arg		cp;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_DISCONNECT);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Connection_Handle = __htob16(Connection_Handle);
	cmd.Reason = Reason;
	
	cp.cmd = (void*)&cmd;
#if 1
	/* may be bug in Nokia BT card */
	cp.flags = HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = 0;
	cp.stop_event = HCI_E_COMMAND_STATUS;
#else
	cp.flags = HCI_W4_STATUS | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK;
	cp.stop_event = HCI_E_COMMAND_STATUS;
#endif

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return cp.status;
}

int __HCI_AddSCOConnection(int fd, CHANDLE Connection_Handle, __u16 Packet_Type)
{
	int				err;
	struct Add_SCO_Connection	cmd;
	struct hci_cmd_arg	cp;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_ADD_SCO_CONNECTION);
	cmd.hdr.Length = CMD_LEN(cmd);

	cmd.Connection_Handle = __htob16(Connection_Handle);
	cmd.Packet_Type = __htob16(Packet_Type);

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK;
	cp.stop_event = HCI_E_COMMAND_STATUS;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return cp.status;
}

int __HCI_AcceptConnectionRequest(int fd, BD_ADDR *bda, __u8 Role)
{
	int					err;
	struct Accept_Connection_Request	cmd;
	struct hci_cmd_arg		cp;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_ACCEPT_CONNECTION_REQUEST);
	cmd.hdr.Length = CMD_LEN(cmd);

	cmd.bda = *bda;
	cmd.Role = Role;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK;
	cp.stop_event = HCI_E_COMMAND_STATUS;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return cp.status;
}

int __HCI_RejectConnectionRequest(int fd, BD_ADDR *bda, __u8 Reason)
{
	int					err = 0;
	struct Reject_Connection_Request	cmd;
	struct hci_cmd_arg		cp;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_REJECT_CONNECTION_REQUEST);
	cmd.hdr.Length = CMD_LEN(cmd);

	cmd.bda = *bda;
	cmd.Reason = Reason;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK;
	cp.stop_event = HCI_E_COMMAND_STATUS;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return cp.status;
}

int __HCI_AuthenticationRequested(int fd, CHANDLE Connection_Handle)
{
	int					err;
	struct Authentication_Requested		cmd;
	struct hci_cmd_arg			cp;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_AUTHENTICATION_REQUESTED);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Connection_Handle = __htob16(Connection_Handle);
	
	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK;
	cp.stop_event = HCI_E_COMMAND_STATUS;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return cp.status;
}


/*
 * XXX: Multiple Command Complete
*/
int HCI_LinkKeyRequestReply(int fd, BD_ADDR *bda, __u8 *Link_Key)
{
	int					err;
	struct Link_Key_Request_Reply		cmd;
	struct hci_cmd_arg		cp;
	struct Link_Key_Request_Reply_Event	cce;

	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_LINK_KEY_REQUEST_REPLY);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.bda = *bda;
	memcpy(cmd.Link_Key, Link_Key, 16);

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	return cce.Status;
}

/*
  XXX: Multiple Command Complete
*/
int HCI_LinkKeyRequestNegativeReply(int fd, BD_ADDR *bda)
{
	int					err;
	struct Link_Key_Request_Negative_Reply	cmd;
	struct hci_cmd_arg		cp;
	struct Link_Key_Request_Reply_Event	cce;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_LINK_KEY_REQUEST_NEGATIVE_REPLY);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.bda = *bda;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	return cce.Status;
}

/*
  XXX: Multiple Command Complete
*/
int HCI_PINCodeRequestReply(int fd, BD_ADDR *bda, __u8 PIN_Code_Length, __u8 *PIN_Code)
{
	int					err;
	struct PIN_Code_Request_Reply		cmd;
	struct hci_cmd_arg		cp;
	struct PIN_Code_Request_Reply_Event	cce;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_PIN_CODE_REQUEST_REPLY);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.bda = *bda;
	cmd.PIN_Code_Length = PIN_Code_Length;
	memcpy(cmd.PIN_Code, PIN_Code, 16);

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	return cce.Status;
}

/*
  XXX:Multiple Command Complete
*/
int HCI_PINCodeRequestNegativeReply(int fd, BD_ADDR *bda)
{
	int					err;
	struct PIN_Code_Request_Negative_Reply	cmd;
	struct hci_cmd_arg		cp;
	struct PIN_Code_Request_Reply_Event	cce;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_PIN_CODE_REQUEST_NEGATIVE_REPLY);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.bda = *bda;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	return cce.Status;
}

#if !defined(__KERNEL__) || defined(CONFIG_AFFIX_PAN)
int HCI_ChangeConnectionPacketType(int fd, CHANDLE Connection_Handle, __u16 Packet_Type)
{
	int						err;
	unsigned char					buf[EVENT_MAX_SIZE];
	struct Change_Connection_Packet_Type		cmd;
	struct Connection_Packet_Type_Changed_Event	*cce = (void*)buf;
	struct hci_cmd_arg				cp;
	struct hci_event_arg				ep;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_CHANGE_CONNECTION_PACKET_TYPE);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Connection_Handle = __htob16(Connection_Handle);
	cmd.Packet_Type = __htob16(Packet_Type);
	
	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK | CONNECTION_PACKET_TYPE_CHANGED_MASK;
	cp.event = (void*)buf;	/* not used */

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cp.status)
		return cp.status;

	ep.event = (void*)buf;

	for (;;) {
		err = hci_recv_event(fd, &ep);
		if (err < 0) {
			return err;
		}

		if (__btoh16(cce->Connection_Handle) == Connection_Handle) {
			hci_listen_event(fd, 0);	/* remove listener */
			return cce->Status;
		}
	}

	return 0;	
}
#endif

/*
 * Link Policy
 */

#if !defined(__KERNEL__) || defined(CONFIG_AFFIX_PAN)
int HCI_SwitchRole(int fd, BD_ADDR *bda, __u8 Role)
{
	int				err;
	unsigned char			buf[EVENT_MAX_SIZE];
	struct Role_Switch		cmd;
	struct Role_Change_Event	*cce = (void*)buf;
	struct hci_cmd_arg		cp;
	struct hci_event_arg		ep;

	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_SWITCH_ROLE);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.bda = *bda;
	cmd.Role = Role;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK | ROLE_CHANGE_MASK;
	cp.event = (void*)buf;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cp.status)
		return cp.status;

	ep.event = (void*)buf;

	for (;;) {
		err = hci_recv_event(fd, &ep);
		if (err < 0) {
			return err;
		}

		if (memcmp(&cce->bda, bda, 6) == 0) {
			hci_listen_event(fd, 0);	/* remove listener */
			return cce->Status;
		}
	}

	return 0;	
}
#endif


/*
 * Baseband Commands
 */

#if !defined(__KERNEL__) || defined(CONFIG_AFFIX_PAN)
int HCI_ReadScanEnable(int fd, __u8 *Scan_Enable)
{
	int				err = 0;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Read_Scan_Enable_Event	rsee;


	/* setup command parameters */
	cmd.OpCode = __htob16(HCI_C_READ_SCAN_ENABLE);
	cmd.Length = 0;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&rsee;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	*Scan_Enable = rsee.Scan_Enable;

	return rsee.Status;
}
#endif

int HCI_WriteScanEnable(int fd, __u8 Scan_Enable)
{
	int				err;
	struct Write_Scan_Enable	cmd;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_WRITE_SCAN_ENABLE);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Scan_Enable = Scan_Enable;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs ;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return ccs.Status;
}

int HCI_WriteClassOfDevice(int fd, __u32 Class_of_Device)
{
	int				err = 0;
	struct Write_Class_of_Device	cmd;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;

	cmd.hdr.OpCode = __htob16(HCI_C_WRITE_CLASS_OF_DEVICE);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Class_of_Device = __htob24(Class_of_Device);	/* XXX: check it */

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return ccs.Status;
}

int HCI_ChangeLocalName(int fd, char *Name)
{
	int				err;
	struct Change_Local_Name	cmd;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;

	cmd.hdr.OpCode = __htob16(HCI_C_CHANGE_LOCAL_NAME);
	cmd.hdr.Length = CMD_LEN(cmd);

	strncpy((char*)cmd.Name, Name, 248);

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs ;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return ccs.Status;
}

/*
  XXX: multiple Command Completes
*/
#if !defined(__KERNEL__) || defined(CONFIG_AFFIX_PAN)
int HCI_WriteLinkSupervisionTimeout(int fd, CHANDLE Connection_Handle, __u16 Link_Supervision_Timeout)
{
	int						err = 0;
	struct Write_Link_Supervision_Timeout		cmd;
	struct hci_cmd_arg			cp;
	struct Write_Link_Supervision_Timeout_Event	cce;

	cmd.hdr.OpCode = __htob16(HCI_C_WRITE_LINK_SUPERVISION_TIMEOUT);
	cmd.hdr.Length = CMD_LEN(cmd);

	cmd.Connection_Handle = __htob16(Connection_Handle);
	cmd.Link_Supervision_Timeout = __htob16(Link_Supervision_Timeout);

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return cce.Status;
}
#endif

/*
 * Informational
 */

int HCI_ReadBufferSize(int fd, __u16 *HC_ACL_Data_Packet_Length, __u8 *HC_SCO_Data_Packet_Length,
		       __u16 *Total_Num_ACL_Data_Packets, __u16 *Total_Num_SCO_Data_Packets)
{
	int				err = 0;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Read_Buffer_Size_Event	cce;

	cmd.OpCode = __htob16(HCI_C_READ_BUFFER_SIZE);
	cmd.Length = 0;
	
	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cce.Status)
		return cce.Status;

	*HC_ACL_Data_Packet_Length = __btoh16(cce.HC_ACL_Data_Packet_Length);
	*HC_SCO_Data_Packet_Length = cce.HC_SCO_Data_Packet_Length;
	*Total_Num_ACL_Data_Packets = __btoh16(cce.Total_Num_ACL_Data_Packets);
	*Total_Num_SCO_Data_Packets = __btoh16(cce.Total_Num_SCO_Data_Packets);

	return 0;
}

int HCI_ReadBDAddr(int fd, BD_ADDR *bda)
{
	int				err = 0;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg		cp;
	struct Read_BD_ADDR_Event	rbda;

	cmd.OpCode = __htob16(HCI_C_READ_BD_ADDR);
	cmd.Length = 0;
	
	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&rbda;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (rbda.Status)
		return rbda.Status;

	*bda = rbda.bda;

	return 0;
}

int HCI_ReadLocalVersionInformation(int fd, __u8 *HCI_Version, __u16 *HCI_Revision,
				    __u8 *LMP_Version, __u16 *Manufacture_Name, __u16 *LMP_Subversion)
{
	int						err = 0;
	HCI_Command_Packet_Header			cmd;
	struct hci_cmd_arg			cp;
	struct Read_Local_Version_Information_Event	cce;

	cmd.OpCode = __htob16(HCI_C_READ_LOCAL_VERSION_INFORMATION);
	cmd.Length = 0;
	
	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cce.Status)
		return cce.Status;

	*HCI_Version = cce.HCI_Version;
	*HCI_Revision = __btoh16(cce.HCI_Revision);
	*LMP_Version = cce.LMP_Version;
	*Manufacture_Name = __btoh16(cce.Manufacture_Name);
	*LMP_Subversion = __btoh16(cce.LMP_Subversion);

	return 0;
}




#ifdef __KERNEL__

/*
 * only kernel space commands
 */

#else

/*
 * only user space commands
 */

/* Link Control Commands */


int HCI_InquiryCancel(int fd)
{
	int				err;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;


	/* setup command parameters */
	cmd.OpCode = __htob16(HCI_C_INQUIRY_CANCEL);
	cmd.Length = 0;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return ccs.Status;	// return if an error
}

int HCI_PeriodicInquiryMode(int fd, __u16 Max_Period_Length, __u16 Min_Period_Length, __u16 Inquiry_Length, __u8 Max_Num_Responses)
{
	int				err;
	struct Periodic_Inquiry_Mode	cmd;
	struct hci_cmd_arg		cp;
	struct Command_Complete_Status	ccs;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_PERIODIC_INQUIRY_MODE);
	cmd.hdr.Length = CMD_LEN(cmd);

	cmd.Max_Period_Length = __htob16(Max_Period_Length);
	cmd.Min_Period_Length = __htob16(Min_Period_Length);
	cmd.LAP = __htob24(GIAC);
	cmd.Inquiry_Length = Inquiry_Length;
	cmd.Max_Num_Responses = Max_Num_Responses;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return ccs.Status;	// return if an error
}

int HCI_ExitPeriodicInquiryMode(int fd)
{
	int				err;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;


	/* setup command parameters */
	cmd.OpCode = __htob16(HCI_C_EXIT_PERIODIC_INQUIRY_MODE);
	cmd.Length = 0;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return ccs.Status;	// return if an error
}


int HCI_CreateConnection(int fd, INQUIRY_ITEM *dev, __u16 Packet_Type, __u8 Allow_Role_Switch, CHANDLE *Connection_Handle, __u8 *Link_Type, __u8 *Encryption_Mode)
{
	int					err = 0;
	unsigned char				buf[EVENT_MAX_SIZE];
	struct Create_Connection		cmd;
	struct Connection_Complete_Event	*cce = (void*)buf;
	struct hci_cmd_arg		cp;
	struct hci_event_arg			ep;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_CREATE_CONNECTION);
	cmd.hdr.Length = CMD_LEN(cmd);

	cmd.bda = dev->bda;
	cmd.Packet_Type = __htob16(Packet_Type);
	cmd.PS_Repetition_Mode = dev->PS_Repetition_Mode;
	cmd.PS_Mode = dev->PS_Mode;
	cmd.Clock_Offset = __htob16(dev->Clock_Offset);
	cmd.Allow_Role_Switch = Allow_Role_Switch;


	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK | CONNECTION_COMPLETE_MASK;
	cp.event = (void*)buf;	/* not used */

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cp.status)
		return cp.status;

	ep.event = (void*)buf;

	for (;;) {
		err = hci_recv_event(fd, &ep);
		if (err < 0) {
			return err;
		}

		if (memcmp(&cce->bda, &dev->bda, 6) == 0) {
			hci_listen_event(fd, 0);	/* remove listener */
			if (cce->Status)
				return cce->Status;
			*Connection_Handle = __btoh16(cce->Connection_Handle);
			*Link_Type = cce->Link_Type;
			*Encryption_Mode = cce->Encryption_Mode;
			break;
		}
	}

	return 0;	
}

int HCI_Disconnect(int fd, CHANDLE Connection_Handle, __u8 Reason)
{
	int					err;
	unsigned char				buf[EVENT_MAX_SIZE];
	struct Disconnect			cmd;
	struct Disconnection_Complete_Event	*dce = (void*)buf;
	struct hci_cmd_arg		cp;
	struct hci_event_arg			ep;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_DISCONNECT);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Connection_Handle = __htob16(Connection_Handle);
	cmd.Reason = Reason;
	
	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK | DISCONNECTION_COMPLETE_MASK;
	cp.event = (void*)buf;	/* not used */

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cp.status)
		return cp.status;

	ep.event = (void*)buf;

	for (;;) {
		err = hci_recv_event(fd, &ep);
		if (err < 0) {
			return err;
		}

		if (__btoh16(dce->Connection_Handle) == Connection_Handle) {
			hci_listen_event(fd, 0);	/* remove listener */
			return dce->Status;
		}
	}

	return 0;	
}


int HCI_AcceptConnectionRequest(int fd, BD_ADDR *bda, __u8 Role, CHANDLE *Connection_Handle, __u8 *Link_Type, __u8 *Encryption_Mode)
{
	int					err = 0;
	unsigned char				buf[EVENT_MAX_SIZE];
	struct Accept_Connection_Request	cmd;
	struct Connection_Complete_Event	*cce = (void*)buf;
	struct hci_cmd_arg		cp;
	struct hci_event_arg			ep;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_ACCEPT_CONNECTION_REQUEST);
	cmd.hdr.Length = CMD_LEN(cmd);

	cmd.bda = *bda;
	cmd.Role = Role;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK | CONNECTION_COMPLETE_MASK;
	cp.event = (void*)buf;	/* not used */

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cp.status)
		return cp.status;

	ep.event = (void*)buf;

	for (;;) {
		err = hci_recv_event(fd, &ep);
		if (err < 0) {
			return err;
		}

		if (memcmp(&cce->bda, bda, 6) == 0) {
			hci_listen_event(fd, 0);	/* remove listener */
			if (cce->Status)
				return cce->Status;
			*Connection_Handle = __btoh16(cce->Connection_Handle);
			*Link_Type = cce->Link_Type;
			*Encryption_Mode = cce->Encryption_Mode;
			break;
		}
	}

	return 0;	
}

int HCI_RejectConnectionRequest(int fd, BD_ADDR *bda, __u8 Reason)
{
	int					err = 0;
	unsigned char				buf[EVENT_MAX_SIZE];
	struct Reject_Connection_Request	cmd;
	struct Connection_Complete_Event	*cce = (void*)buf;
	struct hci_cmd_arg		cp;
	struct hci_event_arg			ep;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_REJECT_CONNECTION_REQUEST);
	cmd.hdr.Length = CMD_LEN(cmd);

	cmd.bda = *bda;
	cmd.Reason = Reason;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK | CONNECTION_COMPLETE_MASK;
	cp.event = (void*)buf;	/* not used */

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cp.status)
		return cp.status;

	ep.event = (void*)buf;

	for (;;) {
		err = hci_recv_event(fd, &ep);
		if (err < 0) {
			return err;
		}

		if (memcmp(&cce->bda, bda, 6) == 0) {
			hci_listen_event(fd, 0);	/* remove listener */
			return cce->Status;
		}
	}

	return 0;	
}

int HCI_AuthenticationRequested(int fd, CHANDLE Connection_Handle)
{
	int					err;
	unsigned char				buf[EVENT_MAX_SIZE];
	struct Authentication_Requested		cmd;
	struct Authentication_Complete_Event	*cce = (void*)buf;
	struct hci_cmd_arg		cp;
	struct hci_event_arg			ep;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_AUTHENTICATION_REQUESTED);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Connection_Handle = __htob16(Connection_Handle);
	
	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK | AUTHENTICATION_COMPLETE_MASK;
	cp.stop_event = HCI_E_AUTHENTICATION_COMPLETE;
	cp.event = (void*)buf;	/* not used */

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cp.status)
		return cp.status;

	ep.event = (void*)buf;

	for (;;) {
		err = hci_recv_event(fd, &ep);
		if (err < 0) {
			return err;
		}

		if (__btoh16(cce->Connection_Handle) == Connection_Handle) {
			hci_listen_event(fd, 0);	/* remove listener */
			return cce->Status;
		}
	}

	return 0;	
}

int HCI_SetConnectionEncryption(int fd, CHANDLE Connection_Handle, __u8 Encryption_Enable)
{
	int					err;
	unsigned char				buf[EVENT_MAX_SIZE];
	struct Set_Connection_Encryption 	cmd;
	struct Encryption_Change_Event		*cce = (void*)buf;
	struct hci_cmd_arg		cp;
	struct hci_event_arg			ep;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_SET_CONNECTION_ENCRYPTION);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Connection_Handle = __htob16(Connection_Handle);
	cmd.Encryption_Enable = Encryption_Enable;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK | ENCRYPTION_CHANGE_MASK;
	cp.stop_event = HCI_E_ENCRYPTION_CHANGE;
	cp.event = (void*)buf;	/* not used */

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cp.status)
		return cp.status;

	ep.event = (void*)buf;

	for (;;) {
		err = hci_recv_event(fd, &ep);
		if (err < 0) {
			return err;
		}

		if (__btoh16(cce->Connection_Handle) == Connection_Handle) {
			hci_listen_event(fd, 0);	/* remove listener */
			return cce->Status;
		}
	}

	return 0;	
}

/*
  May be do not receive Link_Key_Notification at all...
*/

int HCI_ChangeConnectionLinkKey(int fd, CHANDLE Connection_Handle,
				BD_ADDR *bda, __u8 *Link_Key, __u8 *Key_Type)
{
	int					err;
	unsigned char				buf[EVENT_MAX_SIZE];
	struct Change_Connection_Link_Key 	cmd;
	struct Change_Connection_Link_Key_Complete_Event		*cce = (void*)buf;
	struct hci_cmd_arg		cp;
	struct hci_event_arg			ep;
	struct Link_Key_Notification_Event	*lkne = (void*)buf;

	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_CHANGE_CONNECTION_LINK_KEY);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Connection_Handle = __htob16(Connection_Handle);

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK  | CHANGE_CONNECTION_LINK_KEY_COMPLETE_MASK;
	if (bda)
		cp.event_mask |= LINK_KEY_NOTIFICATION_MASK;
	cp.event = (void*)buf;	/* not used */

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cp.status)
		return cp.status;

	ep.event = (void*)buf;

	for (;;) {
		err = hci_recv_event(fd, &ep);
		if (err < 0) {
			return err;
		}

		if (bda) {
			if (ep.event->EventCode == HCI_E_LINK_KEY_NOTIFICATION) {
			/* geather results */
				memcpy(Link_Key, lkne->Link_Key, 16);
				*Key_Type = lkne->Key_Type;
			} else if (ep.event->EventCode == HCI_E_CHANGE_CONNECTION_LINK_KEY_COMPLETE) {
				if (__btoh16(cce->Connection_Handle) == Connection_Handle) {
					hci_listen_event(fd, 0);	/* remove listener */
					return cce->Status;
				}
			}
		} else {
			if (__btoh16(cce->Connection_Handle) == Connection_Handle) {
				hci_listen_event(fd, 0);	/* remove listener */
				return cce->Status;
			}

		}
	}

	return 0;	
}

/*
  Hmm... One Connection_Handle for all slaves ???!!! XXX:
*/
int HCI_MasterLinkKey(int fd, __u8 Key_Flag, CHANDLE *Connection_Handle)
{
	int					err;
	unsigned char				buf[EVENT_MAX_SIZE];
	struct Master_Link_Key 			cmd;
	struct Master_Link_Key_Complete_Event	*cce = (void*)buf;
	struct hci_cmd_arg		cp;

	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_MASTER_LINK_KEY);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Key_Flag = Key_Flag;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_USE_STOP_EVENT | HCI_W4_EXTRA_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK | MASTER_LINK_KEY_COMPLETE_MASK;
	cp.stop_event = HCI_E_MASTER_LINK_KEY_COMPLETE;
	cp.extra_event = HCI_E_MASTER_LINK_KEY_COMPLETE;
	cp.event = (void*)cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cp.status)
		return cp.status;
	*Connection_Handle = cce->Connection_Handle;
	return cce->Status;
}


int HCI_RemoteNameRequest(int fd, INQUIRY_ITEM *dev, char *Name)
{
	int					err;
	unsigned char				buf[EVENT_MAX_SIZE];
	struct Remote_Name_Request	 	cmd;
	struct Remote_Name_Request_Complete_Event		*cce = (void*)buf;
	struct hci_cmd_arg		cp;
	struct hci_event_arg			ep;

	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_REMOTE_NAME_REQUEST);
	cmd.hdr.Length = CMD_LEN(cmd);

	cmd.bda = dev->bda;
	cmd.PS_Repetition_Mode = dev->PS_Repetition_Mode;
	cmd.PS_Mode = dev->PS_Mode;
	cmd.Clock_Offset = __htob16(dev->Clock_Offset);

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK | REMOTE_NAME_REQUEST_COMPLETE_MASK;
	cp.event = (void*)buf;	/* not used */

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cp.status)
		return cp.status;

	ep.event = (void*)buf;

	for (;;) {
		err = hci_recv_event(fd, &ep);
		if (err < 0) {
			return err;
		}
		if (memcmp(&cce->bda, &dev->bda, 6) == 0) {
			hci_listen_event(fd, 0);	/* remove listener */
			if (cce->Status)
				return cce->Status;
			strncpy(Name, cce->Name, 248);
			break;
		}
	}

	return 0;	
}


/*
 * Link Policy commands
 */

int HCI_ParkMode(int fd, CHANDLE Connection_Handle, __u16 Beacon_Max_Interval, __u16 Beacon_Min_Interval)
{
	int				err;
	struct Park_Mode		cmd;
	struct hci_cmd_arg		cp;
	unsigned char			buf[EVENT_MAX_SIZE];
	struct Mode_Change_Event	*cce = (void*)buf;
	struct hci_event_arg		ep;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_PARK_MODE);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Connection_Handle = __htob16(Connection_Handle);
	cmd.Beacon_Max_Interval = __htob16(Beacon_Max_Interval);
	cmd.Beacon_Min_Interval = __htob16(Beacon_Max_Interval);

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK | MODE_CHANGE_MASK;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	if (cp.status)
		return cp.status;

	ep.event = (void*)buf;

	for (;;) {
		err = hci_recv_event(fd, &ep);
		if (err < 0) {
			return err;
		}

		if (__btoh16(cce->Connection_Handle) == Connection_Handle) {
			hci_listen_event(fd, 0);	/* remove listener */
			return cce->Status;
		}
	}
}

int HCI_ExitParkMode(int fd, CHANDLE Connection_Handle, __u8 *Current_Mode, __u16 *Interval)
{
	int				err;
	struct Exit_Park_Mode		cmd;
	struct hci_cmd_arg		cp;
	unsigned char			buf[EVENT_MAX_SIZE];
	struct Mode_Change_Event	*cce = (void*)buf;
	struct hci_event_arg		ep;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_EXIT_PARK_MODE);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Connection_Handle = __htob16(Connection_Handle);

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_STATUS | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_STATUS_MASK | MODE_CHANGE_MASK;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	if (cp.status)
		return cp.status;

	ep.event = (void*)buf;

	for (;;) {
		err = hci_recv_event(fd, &ep);
		if (err < 0) {
			return err;
		}

		if (__btoh16(cce->Connection_Handle) == Connection_Handle) {
			hci_listen_event(fd, 0);	/* remove listener */
			*Current_Mode = cce->Current_Mode;
			*Interval = __btoh16(cce->Interval);
			return cce->Status;
		}
	}
}



int HCI_RoleDiscovery(int fd, CHANDLE Connection_Handle, __u8 *Current_Role)
{
	int					err;
	struct Role_Discovery			cmd;
	struct hci_cmd_arg		cp;
	struct Role_Discovery_Event		cce;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_ROLE_DISCOVERY);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Connection_Handle = __htob16(Connection_Handle);

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	*Current_Role = cce.Current_Role;

	return cce.Status;
}


int HCI_ReadLinkPolicy(int fd, CHANDLE Connection_Handle, __u8 *Link_Policy_Settings)
{
	int					err;
	struct Read_Link_Policy			cmd;
	struct hci_cmd_arg		cp;
	struct Read_Link_Policy_Event		cce;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_READ_LINK_POLICY_SETTINGS);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Connection_Handle = __htob16(Connection_Handle);

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	*Link_Policy_Settings = cce.Link_Policy_Settings;

	return cce.Status;
}

int HCI_WriteLinkPolicy(int fd, CHANDLE Connection_Handle, __u8 Link_Policy_Settings)
{
	int					err;
	struct Write_Link_Policy		cmd;
	struct hci_cmd_arg		cp;
	struct Write_Link_Policy_Event		cce;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_WRITE_LINK_POLICY_SETTINGS);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Connection_Handle = __htob16(Connection_Handle);
	cmd.Link_Policy_Settings = Link_Policy_Settings;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	return cce.Status;
}


/* HC & BB commands */

int HCI_SetEventMask(int fd, __u64 Event_Mask)
{
	int				err;
	struct Set_Event_Mask		cmd;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_SET_EVENT_MASK);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Event_Mask = __htob64(Event_Mask);

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	return ccs.Status;
}

int HCI_Reset(int fd)
{
	int				err;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;


	/* setup command parameters */
	cmd.OpCode = __htob16(HCI_C_RESET);
	cmd.Length = 0;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	return ccs.Status;
}

int HCI_SetEventFilter(int fd, __u8 Filter_Type, __u8 Filter_Condition_Type, __u8 *Condition, __u8 Condition_Length)
{
	int				err;
	unsigned char			buf[sizeof(struct Set_Event_Filter) + 7];/*XXX max*/
	struct Set_Event_Filter		*cmd = (void*)buf;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;


	/* setup command parameters */
	cmd->hdr.OpCode = __htob16(HCI_C_SET_EVENT_FILTER);
	cmd->hdr.Length = CMD_LEN(*cmd) + Condition_Length;
	cmd->Filter_Type = Filter_Type;
	cmd->Filter_Condition_Type = Filter_Condition_Type;
	memcpy(cmd->Condition, Condition, Condition_Length);


	cp.cmd = (void*)cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	return ccs.Status;
}

int HCI_ReadPINType(int fd, __u8 *PIN_Type)
{
	int				err = 0;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Read_PIN_Type_Event	cce;


	/* setup command parameters */
	cmd.OpCode = __htob16(HCI_C_READ_PIN_TYPE);
	cmd.Length = 0;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	*PIN_Type = cce.PIN_Type;

	return cce.Status;
}

int HCI_WritePINType(int fd, __u8 PIN_Type)
{
	int				err = 0;
	struct Write_PIN_Type		cmd;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_WRITE_PIN_TYPE);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.PIN_Type = PIN_Type;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return ccs.Status;
}

int HCI_CreateNewUnitKey(int fd)
{
	int				err = 0;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;


	/* setup command parameters */
	cmd.OpCode = __htob16(HCI_C_CREATE_NEW_UNIT_KEY);
	cmd.Length = 0;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return ccs.Status;
}

int HCI_ReadStoredLinkKey(int fd, BD_ADDR *bda, __u8 Read_All_Flag, __u16 *Max_Num_Keys,
			  __u16 *Num_Keys, struct Link_Key *Link_Keys)
{
	int				err = 0;
	unsigned char			buf[EVENT_MAX_SIZE];
	struct Read_Stored_Link_Key	cmd;
	struct hci_cmd_arg	cp;
	struct Read_Stored_Link_Key_Event	*cce = (void*)buf;
	struct Return_Link_Keys_Event	*rlke  = (void*)buf;
	struct hci_event_arg		ep;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_READ_STORED_LINK_KEY);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.bda = *bda;
	cmd.Read_All_Flag = Read_All_Flag;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK | RETURN_LINK_KEYS_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	ep.event = (void*)buf;

	do {
		err = hci_recv_event(fd, &ep);
		if (err < 0) {
			return err;
		}

		if (ep.event->EventCode == HCI_E_RETURN_LINK_KEYS) {
			*Num_Keys = rlke->Num_Keys;
			memcpy(Link_Keys, rlke->Link_Keys, *Num_Keys*sizeof(struct Link_Key));
			Link_Keys += *Num_Keys;
		}

	} while (ep.event->EventCode != HCI_E_COMMAND_COMPLETE);

	*Max_Num_Keys = __btoh16(cce->Max_Num_Keys);
	*Num_Keys = __btoh16(cce->Num_Keys_Read);
	
	/*
	 * XXX:	may be compare Num_Keys_Read and total read keys
	*/

	return cce->Status;	
}

int HCI_WriteStoredLinkKey(int fd, __u8 Num_Keys_To_Write, struct Link_Key *Link_Keys,
			   __u8 *Num_Keys_Written)
{
	int					err = 0;
	unsigned char				buf[CMD_MAX_SIZE];
	struct Write_Stored_Link_Key		*cmd = (void*)buf;
	struct hci_cmd_arg		cp;
	struct Write_Stored_Link_Key_Event	ccs;
	int					klen;


	/* setup command parameters */
	klen = Num_Keys_To_Write*sizeof(struct Link_Key);

	cmd->hdr.OpCode = __htob16(HCI_C_WRITE_STORED_LINK_KEY);
	cmd->hdr.Length = CMD_LEN(*cmd) + klen;
	cmd->Num_Keys_To_Write = Num_Keys_To_Write;
	memcpy(cmd->Link_Keys, Link_Keys, klen);

	cp.cmd = (void*)cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	*Num_Keys_Written = ccs.Num_Keys_Written;
	
	return ccs.Status;
}

int HCI_DeleteStoredLinkKey(int fd, BD_ADDR *bda, __u8 Delete_All_Flag, __u16 *Num_Keys_Deleted)
{
	int				err = 0;
	struct Delete_Stored_Link_Key	cmd;
	struct hci_cmd_arg	cp;
	struct Delete_Stored_Link_Key_Event	cce;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_DELETE_STORED_LINK_KEY);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.bda = *bda;
	cmd.Delete_All_Flag = Delete_All_Flag;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	*Num_Keys_Deleted = __btoh16(cce.Num_Keys_Deleted);

	return cce.Status;
}


int HCI_ReadLocalName(int fd, char *Name)
{
	int				err = 0;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Read_Local_Name_Event	rlne;

	cmd.OpCode = __htob16(HCI_C_READ_LOCAL_NAME);
	cmd.Length = 0;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&rlne ;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (rlne.Status)
		return rlne.Status;

	strncpy(Name, (char*)rlne.Name, 248);

	return 0;
}

int HCI_ReadPageTimeout(int fd, __u16 *Page_Timeout)
{
	int				err = 0;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Read_Page_Timeout_Event	rpte;


	/* setup command parameters */
	cmd.OpCode = __htob16(HCI_C_READ_PAGE_TIMEOUT);
	cmd.Length = 0;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&rpte;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	*Page_Timeout = rpte.Page_Timeout;

	return rpte.Status;
}

int HCI_WritePageTimeout(int fd, __u16 Page_Timeout)
{
	int				err;
	struct Write_Page_Timeout	cmd;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_WRITE_PAGE_TIMEOUT);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Page_Timeout = Page_Timeout;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs ;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return ccs.Status;
}




int HCI_ReadAuthenticationEnable(int fd, __u8 *Authentication_Enable)
{
	int				err = 0;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Read_Authentication_Enable_Event	cce;

	cmd.OpCode = __htob16(HCI_C_READ_AUTHENTICATION_ENABLE);
	cmd.Length = 0;
	
	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cce.Status)
		return cce.Status;

	*Authentication_Enable = cce.Authentication_Enable;

	return 0;
}

int HCI_WriteAuthenticationEnable(int fd, __u8 Authentication_Enable)
{
	int				err = 0;
	struct Write_Authentication_Enable	cmd;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;

	cmd.hdr.OpCode = __htob16(HCI_C_WRITE_AUTHENTICATION_ENABLE);
	cmd.hdr.Length = CMD_LEN(cmd);

	cmd.Authentication_Enable = Authentication_Enable;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return ccs.Status;
}

int HCI_ReadEncryptionMode(int fd, __u8 *Encryption_Mode)
{
	int				err = 0;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Read_Encryption_Mode_Event	cce;

	cmd.OpCode = __htob16(HCI_C_READ_ENCRYPTION_MODE);
	cmd.Length = 0;
	
	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cce.Status)
		return cce.Status;

	*Encryption_Mode = cce.Encryption_Mode;

	return 0;
}

int HCI_WriteEncryptionMode(int fd, __u8 Encryption_Mode)
{
	int				err = 0;
	struct Write_Encryption_Mode	cmd;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;

	cmd.hdr.OpCode = __htob16(HCI_C_WRITE_ENCRYPTION_MODE);
	cmd.hdr.Length = CMD_LEN(cmd);

	cmd.Encryption_Mode = Encryption_Mode;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return ccs.Status;
}

int HCI_ReadClassOfDevice(int fd, __u32 *Class_of_Device)
{
	int				err = 0;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Read_Class_of_Device_Event	cce;

	cmd.OpCode = __htob16(HCI_C_READ_CLASS_OF_DEVICE);
	cmd.Length = 0;
	
	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cce.Status)
		return cce.Status;

	*Class_of_Device = __btoh24(cce.Class_of_Device);	/* XXX: check it */

	return 0;
}

int HCI_ReadVoiceSetting(int fd, __u16 *Voice_Setting)
{
	int				err;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Read_Voice_Setting_Event	ccs;


	/* setup command parameters */
	cmd.OpCode = __htob16(HCI_C_READ_VOICE_SETTING);
	cmd.Length = 0;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	*Voice_Setting = __btoh16(ccs.Voice_Setting);
			
	return ccs.Status;
}

int HCI_WriteVoiceSetting(int fd, __u16 Voice_Setting)
{
	int				err;
	struct Write_Voice_Setting	cmd;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_WRITE_VOICE_SETTING);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Voice_Setting = __htob16(Voice_Setting);

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	return ccs.Status;
}

int HCI_ReadSCOFlowControl(int fd, __u8 *Flow_Control)
{
	int				err;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Read_SCO_Flow_Control_Event	ccs;


	/* setup command parameters */
	cmd.OpCode = __htob16(HCI_C_READ_SCO_FLOW_CONTROL_ENABLE);
	cmd.Length = 0;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	*Flow_Control = ccs.Flow_Control;
			
	return ccs.Status;
}

int HCI_WriteSCOFlowControl(int fd, __u8 Flow_Control)
{
	int				err;
	struct Write_SCO_Flow_Control	cmd;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_WRITE_SCO_FLOW_CONTROL_ENABLE);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Flow_Control = Flow_Control;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	return ccs.Status;
}


/*
 * XXX: multiple command completes
*/
int HCI_ReadLinkSupervisionTimeout(int fd, CHANDLE Connection_Handle, __u16 *Link_Supervision_Timeout)
{
	int					err = 0;
	struct Read_Link_Supervision_Timeout	cmd;
	struct hci_cmd_arg		cp;
	struct Read_Link_Supervision_Timeout_Event	cce;

	cmd.hdr.OpCode = __htob16(HCI_C_READ_LINK_SUPERVISION_TIMEOUT);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.Connection_Handle = __htob16(Connection_Handle);
	
	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cce.Status)
		return cce.Status;

	*Link_Supervision_Timeout = __btoh16(cce.Link_Supervision_Timeout);

	return 0;
}

int HCI_ReadNumberOfSupportedIAC(int fd, __u8 *Num_Supported_IAC)
{
	int				err = 0;
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Read_Number_Of_Supported_IAC_Event	cce;

	cmd.OpCode = __htob16(HCI_C_READ_NUMBER_OF_SUPPORTED_IAC);
	cmd.Length = 0;
	
	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cce.Status)
		return cce.Status;

	*Num_Supported_IAC = cce.Num_Supported_IAC;

	return 0;
}

int HCI_ReadCurrentIACLAP(int fd, __u8 *Num_Current_IAC, __u32 *IAC_LAP)
{
	int				err = 0, i;
	unsigned char			buf[EVENT_MAX_SIZE];
	HCI_Command_Packet_Header	cmd;
	struct hci_cmd_arg	cp;
	struct Read_Current_IAC_LAP_Event	*cce = (void*)buf;

	cmd.OpCode = __htob16(HCI_C_READ_CURRENT_IAC_LAP);
	cmd.Length = 0;
	
	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cce->Status)
		return cce->Status;

	*Num_Current_IAC = cce->Num_Current_IAC;
	for (i=0; i<*Num_Current_IAC; i++)
		IAC_LAP[i] = __btoh24(cce->IAC_LAP[i].v);

	return 0;
}

int HCI_WriteCurrentIACLAP(int fd, __u8 Num_Current_IAC, __u32 *IAC_LAP)
{
	int				err = 0, i;
	unsigned char			buf[CMD_MAX_SIZE];
	struct Write_Current_IAC_LAP	*cmd = (void*)buf;
	struct hci_cmd_arg	cp;
	struct Command_Complete_Status	ccs;

	cmd->hdr.OpCode = __htob16(HCI_C_WRITE_CURRENT_IAC_LAP);
	cmd->hdr.Length = CMD_LEN(*cmd) + Num_Current_IAC*BD_LAP_SIZE;
	cmd->Num_Current_IAC = Num_Current_IAC;
	for (i=0; i<Num_Current_IAC; i++)
		cmd->IAC_LAP[i].v = __htob24(IAC_LAP[i]);
	
	cp.cmd = (void*)cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	return ccs.Status;
}


/* Informational Commands */
int HCI_ReadLocalSupportedFeatures(int fd, __u64 *LMP_Features)
{
	int						err = 0;
	HCI_Command_Packet_Header			cmd;
	struct hci_cmd_arg			cp;
	struct Read_Local_Supported_Features_Event	cce;

	cmd.OpCode = __htob16(HCI_C_READ_LOCAL_SUPPORTED_FEATURES);
	cmd.Length = 0;
	
	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cce.Status)
		return cce.Status;

	*LMP_Features = __btoh64(cce.LMP_Features);

	return 0;
}

int HCI_ReadCountryCode(int fd, int *Country_Code)
{
	int					err = 0;
	HCI_Command_Packet_Header		cmd;
	struct hci_cmd_arg		cp;
	struct Read_Country_Code_Event		cce;

	cmd.OpCode = __htob16(HCI_C_READ_COUNTRY_CODE);
	cmd.Length = 0;
	
	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&cce;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;

	if (cce.Status)
		return cce.Status;

	*Country_Code = cce.Country_Code;

	return 0;
}


// Ericsson specific

int HCI_EricssonWritePCMSettings(int fd, __u8 Settings)
{
	int					err;
	struct Ericsson_Write_PCM_Settings	cmd;
	struct hci_cmd_arg		cp;
	struct Command_Complete_Status		ccs;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_ERICSSON_WRITE_PCM_SETTINGS);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.PCM_Settings = Settings;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	return ccs.Status;
}


int HCI_EricssonSetSCODataPath(int fd, __u8 Path)
{
	int					err;
	struct Ericsson_Set_SCO_Data_Path	cmd;
	struct hci_cmd_arg		cp;
	struct Command_Complete_Status		ccs;


	/* setup command parameters */
	cmd.hdr.OpCode = __htob16(HCI_C_ERICSSON_SET_SCO_DATA_PATH);
	cmd.hdr.Length = CMD_LEN(cmd);
	cmd.SCO_Data_Path = Path;

	cp.cmd = (void*)&cmd;
	cp.flags = HCI_W4_COMPLETE | HCI_USE_STOP_EVENT | HCI_AUTO_RELEASE;
	cp.event_mask = COMMAND_COMPLETE_MASK;
	cp.stop_event = HCI_E_COMMAND_COMPLETE;
	cp.event = (void*)&ccs;

	err = hci_exec_cmd(fd, &cp);
	if (err)
		return err;
	
	return ccs.Status;
}

#endif

#endif

