/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Muller Ulrich <ulrich.muller@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/*
  $Id: bt_driver.c,v 1.7 2002/08/30 15:16:29 kds Exp $
 
  bt_driver.c - access to Bluetooth HCI/L2CAP functionality
*/

/**********************************************************************************/

#include <linux/config.h>
#include <linux/version.h>

#include <linux/wait.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/if.h>
#include <linux/types.h>
#include <linux/errno.h>


/* Affix includes */
#define FILEBIT	DBPAN

#include <affix/types.h>
#include <affix/bluetooth.h>
#include <affix/btcore.h>
#include <affix/btdebug.h>
#include <affix/hci.h>
#include <affix/l2cap.h>
#include <linux/smp_lock.h>

#include "pan.h"
#include "bt_driver.h"
#include "debug.h"


static int errno;
/************************************************************************************************/
/* constants */

/* class of device field according Bluetooth Assigned Numbers (network byte order)*/
static _BD_CLASS CLASS_OF_DEVICE_COMPUTER =			{ 0x00, 0x01, 0x02 };
static _BD_CLASS CLASS_OF_DEVICE_LAN_ACCESS_POINT_LOAD[8] = {	{ 0x00, 0x03, 0x02 }, { 0x20, 0x03, 0x02},
								{ 0x40, 0x03, 0x02 }, { 0x60, 0x03, 0x02},
								{ 0x80, 0x03, 0x02 }, { 0xA0, 0x03, 0x02},
								{ 0xC0, 0x03, 0x02 }, { 0xE0, 0x03, 0x02} };
/* a LAN access point with 0-7 connections */

/***************************************************************************************/
/* HCI section */

/* The NOBLIN HCI manager must be called from a special thread context. As the main
   module should not switch the context, a thread is opened for every function call. */

#define thread_name "hci btd"

DECLARE_MUTEX(exit_sema);	/* semaphore to signal termination of thread */
static void *param[6];		/* pointers to parameters for hci calls in thread */

int _hci_open_thread(void *p) {
	lock_kernel();
	daemonize();
	strcpy(current->comm, thread_name);
	unlock_kernel();

	*((int*) param[0]) = hci_open((char*) param[1]);	/* listen HCI events */

	up(&exit_sema);
	return 0;
}

_HCI_handle* _hci_open(char *name) {
	_HCI_handle *fd = (_HCI_handle*) kmalloc( sizeof(_HCI_handle), GFP_ATOMIC);
	if(!fd) /* out of memory */
		return NULL;

	down(&exit_sema); /* mark thread running */

	param[0] = (void*) fd;
	param[1] = (void*) name;

	if (kernel_thread(_hci_open_thread, NULL, 0) < 0) {
		DBG_F(DBG_MISC, "error creating thread\n");
		up(&exit_sema); /* not running */
		kfree(fd);
		return NULL;
	}
	down(&exit_sema); /* wait until not running */
	up(&exit_sema); /* unlock semaphore */
	return fd;
}


int _HCI_Inquiry_thread(void *p) {
	lock_kernel();
	daemonize();
	strcpy(current->comm, thread_name);
	unlock_kernel();

	//result = HCI_Inquiry(*handle, Inquiry_Length, Max_Num_Responses, Items, Num_Responses);
	*((int*) param[0]) = HCI_Inquiry( *((int*) param[1]), *((__u8*) param[2]), *((__u8*) param[3]),
		(INQUIRY_ITEM*) param[4], (__u8*) param[5]);

	up(&exit_sema);
	return 0;
}

int _HCI_Inquiry(_HCI_handle* handle, __u8 Inquiry_Length, __u8 Max_Num_Responses, _HCI_Inquiry_Item *inq, __u8 *Num_Responses) {
	int result, i;
	INQUIRY_ITEM *Items = (INQUIRY_ITEM*) kmalloc (Max_Num_Responses * sizeof(INQUIRY_ITEM), GFP_ATOMIC);
	if(!Items)
		return -ENOMEM;

	down(&exit_sema); /* mark thread running */

	param[0] = (void*) &result;
	param[1] = (void*) handle;
	param[2] = (void*) &Inquiry_Length;
	param[3] = (void*) &Max_Num_Responses;
	param[4] = (void*) Items;
	param[5] = (void*) Num_Responses;

	if (kernel_thread(_HCI_Inquiry_thread, NULL, 0) < 0) {
		DBG_F(DBG_MISC, "error creating thread\n");
		result = -EACCES;
	} else /* thread will up semaphore on exit */
		down(&exit_sema); /* wait until not running */

	up(&exit_sema); /* unlock semaphore */

	if(result != 0)
		*Num_Responses = 0;

	for (i = 0; i < *Num_Responses; i++) {
		memcpy(&inq[i].bda, &Items[i].bda, sizeof(_BD_ADDR));
		memcpy(&inq[i].bdc, &Items[i].bda + 1, sizeof(_BD_CLASS));//XXX: checkit
		//memcpy(&inq[i].bdc, &Items[i].Class_of_Device, sizeof(_BD_CLASS));
	}
	kfree(Items);
	return result;
}


int _HCI_WriteLinkSupervisionTimeout_thread(void *p) {
	lock_kernel();
	daemonize();
	strcpy(current->comm, thread_name);
	unlock_kernel();

	//result = HCI_WriteLinkSupervisionTimeout(*fd, Connection_Handle, Link_Supervision_Timeout);
	*((int*) param[0]) = HCI_WriteLinkSupervisionTimeout( *((int*) param[1]), *((__u32*) param[2]), *((__u16*) param[3]));

	up(&exit_sema);
	return 0;
}

int _HCI_WriteLinkSupervisionTimeout(_HCI_handle* fd, __u32 Connection_Handle, __u16 Link_Supervision_Timeout) {
	int result;

	down(&exit_sema); /* mark thread running */

	param[0] = (void*) &result;
	param[1] = (void*) fd;
	param[2] = (void*) &Connection_Handle;
	param[3] = (void*) &Link_Supervision_Timeout;

	if (kernel_thread(_HCI_WriteLinkSupervisionTimeout_thread, NULL, 0) < 0) {
		DBG_F(DBG_MISC, "error creating thread\n");
		result = -EACCES;
	} else /* thread will up semaphore on exit */
		down(&exit_sema); /* wait until not running */

	up(&exit_sema); /* unlock semaphore */
	return result;
}


int _HCI_WriteClassOfDevice_thread(void *p) {
	lock_kernel();
	daemonize();
	strcpy(current->comm, thread_name);
	unlock_kernel();

	//result = HCI_WriteClassOfDevice(*fd, *((BD_CLASS*) Class_of_Device));
	*((int*) param[0]) = HCI_WriteClassOfDevice( *((int*) param[1]), *((__u32*) param[2]));

	up(&exit_sema);
	return 0;
}

int _HCI_WriteClassOfDevice(_HCI_handle* fd, _BD_CLASS *Class_of_Device) {
	int result;

	down(&exit_sema); /* mark thread running */

	param[0] = (void*) &result;
	param[1] = (void*) fd;
	param[2] = (void*) Class_of_Device;

	if (kernel_thread(_HCI_WriteClassOfDevice_thread, NULL, 0) < 0) {
		DBG_F(DBG_MISC, "error creating thread\n");
		result = -EACCES;
	} else /* thread will up semaphore on exit */
		down(&exit_sema); /* wait until not running */

	up(&exit_sema); /* unlock semaphore */
	return result;
}


int _HCI_ReadScanEnable_thread(void *p) {
	lock_kernel();
	daemonize();
	strcpy(current->comm, thread_name);
	unlock_kernel();

	//result = HCI_ReadScanEnable(*fd, Scan_Enable);
	*((int*) param[0]) = HCI_ReadScanEnable( *((int*) param[1]), (__u8*) param[2]);

	up(&exit_sema);
	return 0;
}

int _HCI_ReadScanEnable(_HCI_handle* fd, __u8 *Scan_Enable) {
	int result;

	down(&exit_sema); /* mark thread running */

	param[0] = (void*) &result;
	param[1] = (void*) fd;
	param[2] = (void*) Scan_Enable;

	if (kernel_thread(_HCI_ReadScanEnable_thread, NULL, 0) < 0) {
		DBG_F(DBG_MISC, "error creating thread\n");
		result = -EACCES;
	} else /* thread will up semaphore on exit */
		down(&exit_sema); /* wait until not running */

	up(&exit_sema); /* unlock semaphore */
	return result;
}


int _HCI_WriteScanEnable_thread(void *p) {
	lock_kernel();
	daemonize();
	strcpy(current->comm, thread_name);
	unlock_kernel();

	//result = HCI_WriteScanEnable(*fd, Scan_Enable);
	*((int*) param[0]) = HCI_WriteScanEnable( *((int*) param[1]), *((__u8*) param[2]));

	up(&exit_sema);
	return 0;
}

int _HCI_WriteScanEnable(_HCI_handle* fd, __u8 Scan_Enable) {
	int result;

	down(&exit_sema); /* mark thread running */

	param[0] = (void*) &result;
	param[1] = (void*) fd;
	param[2] = (void*) &Scan_Enable;

	if (kernel_thread(_HCI_WriteScanEnable_thread, NULL, 0) < 0) {
		DBG_F(DBG_MISC, "error creating thread\n");
		result = -EACCES;
	} else /* thread will up semaphore on exit */
		down(&exit_sema); /* wait until not running */

	up(&exit_sema); /* unlock semaphore */
	return result;
}


int _HCI_ReadBDAddr_thread(void *p) {
	lock_kernel();
	daemonize();
	strcpy(current->comm, thread_name);
	unlock_kernel();

	//result = HCI_ReadBDAddr(*fd, (BD_ADDR*) bda);
	*((int*) param[0]) = HCI_ReadBDAddr( *((int*) param[1]), (BD_ADDR*) param[2]);

	up(&exit_sema);
	return 0;
}

int _HCI_ReadBDAddr(_HCI_handle* fd, _BD_ADDR *bda) {
	int result;

	down(&exit_sema); /* mark thread running */

	param[0] = (void*) &result;
	param[1] = (void*) fd;
	param[2] = (void*) bda;

	if (kernel_thread(_HCI_ReadBDAddr_thread, NULL, 0) < 0) {
		DBG_F(DBG_MISC, "error creating thread\n");
		result = -EACCES;
	} else /* thread will up semaphore on exit */
		down(&exit_sema); /* wait until not running */

	up(&exit_sema); /* unlock semaphore */
	return result;
}


int _HCI_SwitchRole_thread(void *p) {
	lock_kernel();
	daemonize();
	strcpy(current->comm, thread_name);
	unlock_kernel();

	//result = HCI_SwitchRole(*fd, (BD_ADDR*) bda, Role);
	*((int*) param[0]) = HCI_SwitchRole( *((int*) param[1]), (BD_ADDR*) param[2], *((__u8*) param[3]));

	up(&exit_sema);
	return 0;
}

int _HCI_SwitchRole(_HCI_handle* fd, _BD_ADDR *bda, __u8 Role) {
	int result;

	down(&exit_sema); /* mark thread running */

	param[0] = (void*) &result;
	param[1] = (void*) fd;
	param[2] = (void*) bda;
	param[3] = (void*) &Role;

	if (kernel_thread(_HCI_SwitchRole_thread, NULL, 0) < 0) {
		DBG_F(DBG_MISC, "error creating thread\n");
		result = -EACCES;
	} else /* thread will up semaphore on exit */
		down(&exit_sema); /* wait until not running */

	up(&exit_sema); /* unlock semaphore */
	return result;
}

int _HCI_ReadNumBroadcastRetransmissions(_HCI_handle* handle, __u8 *num) {
	return -EFAULT;
}

int _HCI_WriteNumBroadcastRetransmissions(_HCI_handle* handle, __u8 num) {
	return -EFAULT;
}


int _HCI_ChangeConnectionPacketType_thread(void *p) {
	lock_kernel();
	daemonize();
	strcpy(current->comm, thread_name);
	unlock_kernel();

	//result = HCI_ChangeConnectionPacketType(*fd, CHANDLE Connection_Handle, __u16 Packet_Type);
	*((int*) param[0]) = HCI_ChangeConnectionPacketType( *((int*) param[1]), *((CHANDLE*) param[2]), *((__u16*) param[3]));

	up(&exit_sema);
	return 0;
}

int _HCI_ChangeConnectionPacketType(_HCI_handle* fd, _L2CAP_con *handle, __u16 type) {
	int result;

	down(&exit_sema); /* mark thread running */

	param[0] = (void*) &result;
	param[1] = (void*) fd;
	param[2] = (void*) &handle->ch->con->chandle;
	param[3] = (void*) &type;

	if (kernel_thread(_HCI_ChangeConnectionPacketType_thread, NULL, 0) < 0) {
		DBG_F(DBG_MISC, "error creating thread\n");
		result = -EACCES;
	} else /* thread will up semaphore on exit */
		down(&exit_sema); /* wait until not running */

	up(&exit_sema); /* unlock semaphore */
	return result;
}


int _hci_close_thread(void *p) {
	lock_kernel();
	daemonize();
	strcpy(current->comm, thread_name);
	unlock_kernel();

	//result = hci_close(*fd);
	*((int*) param[0]) = hci_close( *((int*) param[1]));

	up(&exit_sema);
	return 0;
}

int _hci_close(_HCI_handle* fd) {
	int result;

	down(&exit_sema); /* mark thread running */

	param[0] = (void*) &result;
	param[1] = (void*) fd;

	if (kernel_thread(_hci_close_thread, NULL, 0) < 0) {
		DBG_F(DBG_MISC, "error creating thread\n");
		result = -EACCES;
	} else /* thread will up semaphore on exit */
		down(&exit_sema); /* wait until not running */

	up(&exit_sema); /* unlock semaphore */
	return result;
}


/***************************************************************************************/
/* L2CAP section */
/* register and unregister upper protocol to l2cap
   As the drivers use different protocol structures, it is not possible to register
   the same set of functions to them. Instead the functions _l2ca_* are always
   registered. They modify the parameters and call the "real" callback functions.
   To minimize the overhead of a lookup to the correct functions, the "upper layer
   pointer" of the drivers l2cap connection handle is used as a pointer to the
   correct funtions. But as upper layers always use the "upper layer pointer", a new
   connection handle is used.
*/

static _L2CAP_layer *_L2CAP_protocolls = NULL; /* list of all registered upper protocols */
/* the list is only parsed at connection/disconnection and protocoll registration/unregistration,
   and the list is expected to be small */

/* search for registered upper protocol with psm "psm"
   returns pointer to callback functions on success, NULL otherwise */
_L2CAP_layer* _L2CA_GetUpper(_PSM psm) {
	_L2CAP_layer *p = _L2CAP_protocolls;
	while (p && (p->psm != psm))
		p = p->next;
	return p;
}

/* internal connection handle: whenever a new connection is created (ConnectReq/Ind)
   or closed (DisconnectRes/Cfm, negative ConnectRes/Cfm), an internal connection
   handle is created or closed and linked with the drivers handle */

/* creates a new l2cap connection handle:
   allocate memory, register itself to upper layer pointer, lookup upper layer functions */
_L2CAP_con* _L2CAP_con_create(l2cap_ch *ch, int initiator) {
	_L2CAP_con *con = (_L2CAP_con*) kmalloc (sizeof(_L2CAP_con), GFP_ATOMIC);
	if(!con)
		return NULL;

	l2ca_hold(ch);
	con->ch = ch;
	con->priv = NULL;
	con->initiator = initiator;

	memcpy(&con->bda, &ch->bda, sizeof(_BD_ADDR));

	con->state = _L2CAP_CON_UNCONFIGURED;
	con->mtu_out = _PAN_MTU;
	con->layer = _L2CA_GetUpper(ch->psm);
	if(!con->layer) {
		kfree(con);
		return NULL;
	}

	ch->priv = (void*) con;
	return con;
}

/* frees memory of a connection handle */
void _L2CAP_con_free(_L2CAP_con *con) {
	con->ch->priv = NULL;
	kfree(con);
}

/* protocol callbacks for the drivers */

int _l2ca_connect_ind(l2cap_ch *ch) {
	_L2CAP_con *con = _L2CAP_con_create(ch, 0); /* create handle */

	DBG_ENTER(DBG_BTD);

	if(!con) {
		l2ca_connect_rsp(ch, CON_RSP_RESOURCE, 0);
		return -ENOMEM; /* the driver currently ignores return parameters */
	}

	con->layer->ConnectInd(con); /* drop param - unused here */

	DBG_EXIT(DBG_BTD);
	return 0;
}

int _l2ca_connect_cfm(l2cap_ch *ch, __u16 result, __u16 status) {
	_L2CAP_con *con = (_L2CAP_con*) ch->priv;
	DBG_ENTER(DBG_BTD);
	if(con) {
		con->layer->ConnectCfm(con, result, status);
		switch(result) { /* destroy handle on failure */
		case _L2CAP_ConnectRsp_Res_PSMNeg:
		case _L2CAP_ConnectRsp_Res_SecNeg:
		case _L2CAP_ConnectRsp_Res_NoSrc:
			_L2CAP_con_free(con);
			l2ca_put(ch);
		default:
		}
	}
	DBG_EXIT(DBG_BTD);
	return 0;
}

int _l2ca_config_ind(l2cap_ch *ch) {
	_L2CAP_con *con = (_L2CAP_con*) ch->priv;
	DBG_ENTER(DBG_BTD);
	if(con) {
		con->mtu_out = ch->in_flow.mtu; /* hmm, meaning is mtu size of outflow */
		con->layer->ConfigInd(con);
	}
	DBG_EXIT(DBG_BTD);
	return 0;
}

int _l2ca_config_cfm(l2cap_ch *ch, __u16 result) {	
	_L2CAP_con *con = (_L2CAP_con*) ch->priv;
	DBG_ENTER(DBG_BTD);		
	/* this function shouldn't be called anymore at L2CAP connection creation! */
	if(con) {
		con->state |= _L2CAP_CON_CONFIG_OUT;
		con->layer->ConfigCfm(con, result);
	}
	DBG_EXIT(DBG_BTD);
	return 0;
}

int _l2ca_disconnect_ind(l2cap_ch *ch) {
	_L2CAP_con *con = (_L2CAP_con*) ch->priv;
	DBG_ENTER(DBG_BTD);
	if(con)
		con->layer->DisconnectInd(con);
	DBG_EXIT(DBG_BTD);
	return 0;
}


int _l2ca_data_ind(l2cap_ch *ch, struct sk_buff *skb) {
	_L2CAP_con *con = (_L2CAP_con*) ch->priv;
	DBG_ENTER(DBG_BTD);
	if(con)
		con->layer->ReceiveData(con, skb);
	DBG_EXIT(DBG_BTD);
	return 0;
}

/* register upper protocol to internal data structure
   returns 0 on success, fails if psm is already in use */
int _L2CA_register_protocol_internal(_L2CAP_layer *ops) {
	_L2CAP_layer *proto;

	if(_L2CA_GetUpper(ops->psm)) return -EBUSY; /* psm already registered */

	/* make a copy of ops for permanent use */
	proto = kmalloc(sizeof(_L2CAP_layer), GFP_ATOMIC);

	if (!proto) return -ENOMEM;

	memcpy(proto, ops, sizeof(_L2CAP_layer));
	proto->next = _L2CAP_protocolls;
	_L2CAP_protocolls = proto;
	return 0;
}

/* unregister upper protocol from internal data structure
   returns 0 on success, fails if psm is not in use */
int _L2CA_unregister_protocol_internal(_PSM psm) {
	_L2CAP_layer *p1, *p2;;

	if(!_L2CAP_protocolls) return -ENOENT; /* list empty */

	if(_L2CAP_protocolls->psm == psm) { /* delete first */
		p1 = _L2CAP_protocolls;
		_L2CAP_protocolls = _L2CAP_protocolls->next;
		kfree(p1);
		return 0;
	}

	/* check from second to last entry */
	p1 = _L2CAP_protocolls;
	p2 = _L2CAP_protocolls->next;
	while(p2 && (p2->psm != psm) ) {
		p1 = p2;
		p2 = p2->next;
	}

	if(p2) { /* p2 points to entry to be deleted, p1 its precedor */
		p1->next = p2->next;
		kfree(p2);
		return 0;
	}
	return -ENOENT;
}

/* register upper protocol to L2CAP layer
   returns 0 on success */
int _L2CA_register_protocol(_L2CAP_layer *ops) {
	_L2CAP_layer *layer;

	if(!ops) return -EFAULT;

	if(_L2CA_register_protocol_internal(ops))
		return -EBUSY;

	if( (layer = _L2CA_GetUpper(ops->psm)) == NULL) { /* internal test */
		DBG_F(DBG_MISC, "registered protocol unavailable\n");
		return -EFAULT;
	}
	SET_MODULE_OWNER(&layer->pops);
	layer->pops.data_ind        = _l2ca_data_ind;
	layer->pops.connect_ind     = _l2ca_connect_ind;
	layer->pops.connect_cfm     = _l2ca_connect_cfm;
	layer->pops.config_ind      = _l2ca_config_ind;
	layer->pops.config_cfm      = _l2ca_config_cfm;
	layer->pops.disconnect_ind  = _l2ca_disconnect_ind;
	if( l2ca_register_protocol(ops->psm, &layer->pops)) { /* error */
		_L2CA_unregister_protocol(ops->psm);
		return -EBUSY;
	}

	DBG_F(DBG_BTD, "L2CAP PSM 0x%x registered\n", ops->psm);
	return 0;
}

/* unregister upper protocol from L2CAP layer
   returns 0 on success */
int _L2CA_unregister_protocol(_PSM psm) {
	l2ca_unregister_protocol(psm);
	return _L2CA_unregister_protocol_internal(psm);
}

/***************************************************************************************/
/* Upper layers to l2cap */

int _L2CA_ConnectReq(_BD_ADDR *bdaddr, _PSM psm) {
	_L2CAP_layer *layer = _L2CA_GetUpper(psm);
	_L2CAP_con *con;
	l2cap_ch *ch;
	_BD_ADDR rev_bda;

	memcpy(&rev_bda, bdaddr, sizeof(_BD_ADDR));

	if(!layer) {
		DBG_F(DBG_MISC, "cannot find local protocol for psm 0x%x\n", psm);
		return -EFAULT;
	}

	ch = l2ca_create(psm, &layer->pops);
	if (ch == NULL) {
		DBG_F(DBG_MISC, "cannot create NOBLIN connection handle\n");
		return -ENOMEM;
	}

	con = _L2CAP_con_create(ch, 1);
	if(!con) {
		DBG_F(DBG_MISC, "cannot create internal connection handle\n");
		l2ca_put(ch);
		return -ENOMEM;
	}
	memcpy(&con->bda, bdaddr, sizeof(_BD_ADDR));
	con->initiator = 1;

	/* channel configuration */
	l2ca_setmtu(ch, _PAN_MTU);

	return l2ca_connect_req(ch, (BD_ADDR*) &rev_bda);
}

int _L2CA_ConnectRsp(_L2CAP_con *con, __u16 response, __u16 status) {
	l2cap_ch *ch = con->ch;

	l2ca_connect_rsp(ch, response, status);
	switch(response) { /* if connection is rejected, destroy internal connection handle */
	case _L2CAP_ConnectRsp_Res_PSMNeg:
	case _L2CAP_ConnectRsp_Res_SecNeg:
	case _L2CAP_ConnectRsp_Res_NoSrc:
		_L2CAP_con_free(con);
		l2ca_put(ch);
	default:
	}
	return 0;
}

/* meaning is: incoming mtu size, outgoing flush_timeout and QoS
   link_timeout is not set by L2CAP, but by HCI
   a value of 0 means default value */
int _L2CA_ConfigReq(_L2CAP_con *con, __u16 mtu, __u16 flush_timeout, _L2CAP_QoS *QoS, __u16 link_timeout) {
	l2cap_ch *ch = con->ch;

	ch->out_flow.flags = 0;
	if(mtu) {
		set_bit(CFG_OPT_MTU, &ch->out_flow.flags);
		ch->out_flow.mtu       = mtu;
	}
	if(flush_timeout) {
		set_bit(CFG_OPT_FLUSHTO, &ch->out_flow.flags);
		ch->out_flow.flush_to  = flush_timeout;
	}
	if(QoS) {
		set_bit(CFG_OPT_QOS, &ch->out_flow.flags);
		ch->out_flow.qos.flags           = QoS->flags;
		ch->out_flow.qos.service_type    = QoS->service_type;
		ch->out_flow.qos.token_rate      = QoS->token_rate;
		ch->out_flow.qos.token_size      = QoS->token_size;
		ch->out_flow.qos.bandwidth       = QoS->bandwidth;
		ch->out_flow.qos.latency         = QoS->latency;
		ch->out_flow.qos.delay_variation = QoS->delay_variation;
	}

	ch->out_flow.link_to   = link_timeout;
        return l2ca_config_req(ch);
}

/* on sucess, only a mtu has to be specified (<=req.size)
   on failure, all specified (!=0) options are rejected/corrected */
int _L2CA_ConfigRsp(_L2CAP_con *con, __u16 mtu, __u16 flush_timeout, _L2CAP_QoS *QoS, __u16 result) {
	l2cap_ch *ch = con->ch;

/* shouldn't be called anymore (new Affix L2CAP API) */

	if(result == _L2CAP_ConfigRsp_Res_Success)
		con->state |= _L2CAP_CON_CONFIG_IN;

	ch->in_flow.flags = 0;
	if(mtu) {
		con->mtu_out = mtu;

		set_bit(CFG_OPT_MTU, &ch->in_flow.flags);
		ch->in_flow.mtu = mtu;
	}
	if(flush_timeout) {
		set_bit(CFG_OPT_FLUSHTO, &ch->in_flow.flags);
		ch->in_flow.flush_to  = flush_timeout;
	}
	if(QoS) {
		set_bit(CFG_OPT_QOS, &ch->in_flow.flags);
		ch->in_flow.qos.flags           = QoS->flags;
		ch->in_flow.qos.service_type    = QoS->service_type;
		ch->in_flow.qos.token_rate      = QoS->token_rate;
		ch->in_flow.qos.token_size      = QoS->token_size;
		ch->in_flow.qos.bandwidth       = QoS->bandwidth;
		ch->in_flow.qos.latency         = QoS->latency;
		ch->in_flow.qos.delay_variation = QoS->delay_variation;
	}

	return l2ca_config_rsp(ch, result);
}

int _L2CA_DisconnectReq(_L2CAP_con *con) {
	l2cap_ch *ch = con->ch;

	/* Affix does not use DisconnectRes/Cfm, so we emulate the signal here */
	con->state |= _L2CAP_CON_DISCONNECT;
	con->layer->DisconnectCfm(con);
	/* destroy connection handles */
	_L2CAP_con_free(con);
	l2ca_put(ch);
	return 0;
}

int _L2CA_DisconnectRsp(_L2CAP_con *con) {
	l2cap_ch *ch = con->ch;
	_L2CAP_con_free(con);
	l2ca_put(ch);
	/* in NOBLIN, there is no DisconnectRsp/Cfm */
	return 0;
}

int _L2CA_DataWrite(_L2CAP_con *con, struct sk_buff *skb) {
	int	err;
	if((con->state & _L2CAP_CON_CONFIGURED) != _L2CAP_CON_CONFIGURED) {
		DBG_F(DBG_MISC, "error: connection unconfigured\n");
		dev_kfree_skb(skb);
		return -EACCES;
	}

	if(skb->len > con->mtu_out) {
		DBG_F(DBG_BTD, "packet too large, %d bytes\n", skb->len);
		dev_kfree_skb(skb);
		return -EFAULT;
	}


	DBG_F(DBG_BTD, "sending %d bytes\n", skb->len);
	err = l2ca_send_data(con->ch, skb);
	if (err)
		kfree_skb(skb);
	return err;
}

int _L2CA_DataWriteBroadcast(_PSM psm, struct sk_buff *skb) {
	dev_kfree_skb(skb);
	return -EFAULT;
}

/************************************************************************************************/
/* convert bluetooth addr to string */

char *BD_ADDR2str(_BD_ADDR *bda) {
	static unsigned char buf[2][18];
	static int num = 0; /* we use function with up to two calls in one printk statement, this requires
			that there are two buffers in memory */

	num = 1 - num; /* switch buf */
	sprintf(buf[num], "%02x:%02x:%02x:%02x:%02x:%02x", (*bda)[5], (*bda)[4], (*bda)[3], (*bda)[2], (*bda)[1], (*bda)[0]);
	return buf[num];
}

/************************************************************************************************/
/* convert bluetooth CoD to string */

char *BD_CLASS2str(_BD_CLASS *bdc) {
	static unsigned char buf[2][9];
	static int num = 0; /* we use function with up to two calls in one printk statement, this requires
			that there are two buffers in memory */

	if(!memcmp(bdc, &CLASS_OF_DEVICE_COMPUTER, 3))
		return "Computer";
	if(!memcmp(bdc, &CLASS_OF_DEVICE_LAN_ACCESS_POINT_LOAD[0], 3))
		return "LAN Access Point/no connections";
	if(!memcmp(bdc, &CLASS_OF_DEVICE_LAN_ACCESS_POINT_LOAD[1], 3))
		return "LAN Access Point/1 connection";
	if(!memcmp(bdc, &CLASS_OF_DEVICE_LAN_ACCESS_POINT_LOAD[2], 3))
		return "LAN Access Point/2 connections";
	if(!memcmp(bdc, &CLASS_OF_DEVICE_LAN_ACCESS_POINT_LOAD[3], 3))
		return "LAN Access Point/3 connections";
	if(!memcmp(bdc, &CLASS_OF_DEVICE_LAN_ACCESS_POINT_LOAD[4], 3))
		return "LAN Access Point/4 connections";
	if(!memcmp(bdc, &CLASS_OF_DEVICE_LAN_ACCESS_POINT_LOAD[5], 3))
		return "LAN Access Point/5 connections";
	if(!memcmp(bdc, &CLASS_OF_DEVICE_LAN_ACCESS_POINT_LOAD[6], 3))
		return "LAN Access Point/6 connections";
	if(!memcmp(bdc, &CLASS_OF_DEVICE_LAN_ACCESS_POINT_LOAD[7], 3))
		return "LAN Access Point/7 connections";

	num = 1 - num; /* switch buf */
	sprintf(buf[num], "%02x:%02x:%02x", (*bdc)[2], (*bdc)[1], (*bdc)[0]);
	return buf[num];
}

/************************************************************************************************/
/* reverses byte order of _BD_ADDR */

void _BD_ADDR_reverse(_BD_ADDR *bda) {
	_BD_ADDR tmp;
	int j;
	memcpy(&tmp, bda, 6);
	for (j = 0; j < 6; j++)
		(*bda)[j] = tmp[5-j];
}

/************************************************************************************************/
/* copy bdaddr to reverse orderd ethernet addr */

void _BD_ADDR2ETH_ADDR(void *eta, _BD_ADDR *bda) {
	int j;
	for (j = 0; j < 6; j++)
		((__u8*) eta)[j] = (*bda)[5-j];
}

/************************************************************************************************/
/* compare _BD_ADDR with reverse ordered ethernet addr */

int comp_BD_ADDR_ETH_ADDR(_BD_ADDR *bda, void *etaddr) {
	__u8* eta = etaddr;
	return !( (*bda)[0] == eta[5] && (*bda)[1] == eta[4] && (*bda)[2] == eta[3] &&
		  (*bda)[3] == eta[2] && (*bda)[4] == eta[1] && (*bda)[5] == eta[0]);
}


