/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Muller Ulrich <ulrich.muller@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/*
  $Id: bt_driver.h,v 1.3 2002/07/25 15:33:21 kds Exp $
 
  bt_driver.h - access to Bluetooth HCI/L2CAP functionality over various drivers

   Although different Bluetooth mini-drivers should provide the same functionality,
   their API and data structures may be different. This file serves as an abstraction
   from the used driver.
   Supported drivers:
   - Nokia Bluetooth Stack for Linux (NOBLIN), 1_1-8 or higher,
     define CONFIG_BLUETOOTH_NOBLIN for use
     define CONFIG_BLUETOOTH_NOBLIN_WORKAROUND to avoid usage of HCI Manager
   The drivers need some changes, see README.NOBLIN and README.OpenBT for details
*/

/***************************************************************************************/
/* Includes */

#ifndef _BT_DRIVER_H
#define _BT_DRIVER_H


#include <linux/skbuff.h>

//#include "config.h"

#include <affix/btdebug.h>
#include <affix/hci.h>
#include <affix/btcore.h>
#include <affix/l2cap.h>

/************************************************************************************************/
/* general types */

#define default_bt_device "bt0"
#define _PAN_MTU 1691

/* hard_header_len: when passing a sk_buff to the driver, it needs some room in front to insert
   a header for lower protocols (l2cap, hci, tlp). NOBLIN needs 12 (L2CAP_HDR_LEN + ACL_HDR_LEN
   + TLP_HDR_LEN) bytes, OpenBT 0, as all packets must be converted to OpenBT own memory buffers.
   Thus, hard_header_len is guaranteed to be available for the header. As mentioned in some sb_buff
   documentations, some cpus have performacs disadvantages if upper layers are not aligned with
   a multiple of 16 byte. */
  #define _L2CAP_hard_header_len 16

typedef __u8 _BD_ADDR[6]; /* 48 bit hardware address */
typedef __u8 _BD_CLASS[3]; /* 24 bit class */

typedef struct _HCI_Inquiry_Item {
	_BD_ADDR bda;
	_BD_CLASS bdc;
} _HCI_Inquiry_Item;

typedef int _HCI_handle; /* NOBLIN-MGR */


typedef __u16 _PSM; /* L2CAP PSM */
typedef struct { /* L2CAP QoS options */
	__u8	flags; /* reserved for future use and must be set to zero */
	__u8	service_type;
#define _L2CAP_QOS_NO_TRAFFIC		0x00
#define _L2CAP_QOS_BEST_EFFORT		0x01
#define _L2CAP_QOS_GUARANTEED		0x02
	__u32	token_rate;
	__u32	token_size;
	__u32	bandwidth;
	__u32	latency;
	__u32	delay_variation;
} _L2CAP_QoS;

typedef struct { /* L2CAP connection handle */
	l2cap_ch		*ch;		/* L2CAP connection handle */
	void*			priv;		/* pointer for upper layer */
	_BD_ADDR		bda;		/* address of remote device */
	int			initiator;      /* has this device established the connection? */
	int			state;		/* is channel already configured? */
#define _L2CAP_CON_UNCONFIGURED	0x00
#define _L2CAP_CON_CONFIG_IN	0x01
#define _L2CAP_CON_CONFIG_OUT	0x02
#define _L2CAP_CON_CONFIGURED	(_L2CAP_CON_CONFIG_IN | _L2CAP_CON_CONFIG_OUT)
#define _L2CAP_CON_DISCONNECT	0x04		/* waiting for disconnectcfm */
	int			mtu_out;	/* mtu the connection partner can accept; learned ConfigInd */
	struct _L2CAP_layer	*layer;		/* internal use */
} _L2CAP_con;

/************************************************************************************************/
/* HCI functions
   see the Bluetooth Specification for a detailed description of the parameters. A return value
   of 0 means success. To avoid name conflicts with the driver functions, all function names
   begin with "_".
*/
_HCI_handle* _hci_open(char *name); /* return value NULL means error */
int _HCI_Inquiry(_HCI_handle* handle, __u8 Inquiry_Length, __u8 Max_Num_Responses, _HCI_Inquiry_Item *inq, __u8 *Num_Responses);
//int _HCI_ReadLinkSupervisionTimeout(_HCI_handle* handle, __u32 Connection_Handle, __u16 *Link_Supervision_Timeout);
int _HCI_WriteLinkSupervisionTimeout(_HCI_handle* handle, __u32 Connection_Handle, __u16 Link_Supervision_Timeout);
int _HCI_WriteClassOfDevice(_HCI_handle* handle, _BD_CLASS *Class_of_Device);
int _HCI_ReadScanEnable(_HCI_handle* handle, __u8 *Scan_Enable);
#define _HCI_Scan_off     0x00
#define _HCI_Scan_Inquiry 0x01
#define _HCI_Scan_Page    0x02 /* can be combined with "or" */
int _HCI_WriteScanEnable(_HCI_handle* handle, __u8 Scan_Enable);
int _HCI_ReadBDAddr(_HCI_handle* handle, _BD_ADDR *bda);
int _HCI_SwitchRole(_HCI_handle* handle, _BD_ADDR *bda, __u8 Role);
#define _HCI_Role_Master	0x00
#define _HCI_Role_Slave		0x01
int _HCI_ReadNumBroadcastRetransmissions(_HCI_handle* handle, __u8 *num);
int _HCI_WriteNumBroadcastRetransmissions(_HCI_handle* handle, __u8 num);
int _HCI_ChangeConnectionPacketType(_HCI_handle* handle, _L2CAP_con* l2handle, __u16 type);
/* note: we need handle as device handle and l2handle as connection handle */
#define _DM1 0x0008
#define _DM3 0x0400
#define _DM5 0x4000
#define _DH1 0x0010
#define _DH3 0x0800
#define _DH5 0x8000
int _hci_close(_HCI_handle* handle);

/************************************************************************************************/
/* L2CAP <-> upper layer */

/* Upper Layer to L2CAP events according to spec D 4.3.1:
	L2CA_ConnectReq
	L2CA_ConnectRsp
	L2CA_ConnectRspNeg
	L2CA_ConfigReq
	L2CA_ConfigRsp
	L2CA_ConfigRspNeg
	L2CA_DisconnectReq
	L2CA_DisconnectRsp
	L2CA_DataRead // polling
	L2CA_DataWrite
   We skip L2CA_DataRead here and implement it as signal from L2CAP to Upper Layer
*/

int _L2CA_ConnectReq(_BD_ADDR *bdaddr, _PSM psm);
int _L2CA_ConnectRsp(_L2CAP_con *con, __u16 response, __u16 status);
#define _L2CAP_ConnectRsp_Res_Success 0x0000
#define _L2CAP_ConnectRsp_Res_Pending 0x0001
#define _L2CAP_ConnectRsp_Res_PSMNeg  0x0002
#define _L2CAP_ConnectRsp_Res_SecNeg  0x0003
#define _L2CAP_ConnectRsp_Res_NoSrc   0x0004
#define _L2CAP_ConnectRsp_Stat_NoInformation 0x0000 /* for pending */
#define _L2CAP_ConnectRsp_Stat_AuthenPending 0x0001
#define _L2CAP_ConnectRsp_Stat_AuthorPending 0x0002
/* meaning is: incoming mtu size, outgoing flush_timeout and QoS
   link_timeout is not set by L2CAP, but by HCI
   a value of 0 means default value */
int _L2CA_ConfigReq(_L2CAP_con *con, __u16 mtu, __u16 flush_timeout, _L2CAP_QoS *QoS, __u16 link_timeout);
/* on sucess, only a mtu has to be specified ( <= req.size )
   on failure, all specified (!=0) options are rejected/corrected */
int _L2CA_ConfigRsp(_L2CAP_con *con, __u16 mtu, __u16 flush_timeout, _L2CAP_QoS *QoS, __u16 result);
#define _L2CAP_ConfigRsp_Res_Success                0x0000
#define _L2CAP_ConfigRsp_Res_Failure_unacceptable   0x0001
#define _L2CAP_ConfigRsp_Res_Failure_rejected       0x0002
#define _L2CAP_ConfigRsp_Res_Failure_unknown        0x0003
int _L2CA_DisconnectReq(_L2CAP_con *con);
int _L2CA_DisconnectRsp(_L2CAP_con *con);
int _L2CA_DataWrite(_L2CAP_con *con, struct sk_buff *skb);
int _L2CA_DataWriteBroadcast(_PSM psm, struct sk_buff *skb);

/* 2. L2CAP to Upper Layer actions according to Spec D 4.3.2
	L2CA_ConnectInd
	L2CA_ConnectCfm
	L2CA_ConnectCfmNeg      (AFFIX: parameter of ConnectCfm)
	L2CA_ConnectPnd         (AFFIX: missing, but signaling ConnectPnd is optional, so we dont need it)
	L2CA_ConfigInd
	L2CA_ConfigCfm
	L2CA_ConfigCfmNeg       (AFFIX: parameter of ConnectCfm)
	L2CA_DisconnectInd
	L2CA_DisconnectCfm      (AFFIX: missing! we assume that we can immediately close the channel after DisconnectReq)
	L2CA_TimeOutInd         (AFFIX: missing: we dont need it)
	L2CA_QuSViolationInd    (AFFIX: missing: we dont need it)
*/

/* Function pointers to upper layers, confirm and indication functions */
typedef struct _L2CAP_layer {
	void (*ConnectInd) (_L2CAP_con *con);
	void (*ConnectCfm) (_L2CAP_con *con, __u16 result, __u16 status);
#define _L2CAP_ConnectCfm_Res_Success 0x0000
#define _L2CAP_ConnectCfm_Res_Pending 0x0001
#define _L2CAP_ConnectCfm_Res_PSMNeg  0x0002
#define _L2CAP_ConnectCfm_Res_SecNeg  0x0003
#define _L2CAP_ConnectCfm_Res_NoSrc   0x0004
#define _L2CAP_ConnectCfm_Stat_NoInformation 0x0000
#define _L2CAP_ConnectCfm_Stat_AuthenPending 0x0001
#define _L2CAP_ConnectCfm_Stat_AuthorPending 0x0002
	void (*ConfigInd) (_L2CAP_con *con);
	void (*ConfigCfm) (_L2CAP_con *con, __u16 result);
#define _L2CAP_ConfigCfm_Res_Success 0x0000
#define _L2CAP_ConfigCfm_Res_Failure 0x0001
#define _L2CAP_ConfigCfm_Res_Reject  0x0002
#define _L2CAP_ConfigCfm_Res_Unknown 0x0003
	void (*DisconnectInd) (_L2CAP_con *con);
	void (*DisconnectCfm) (_L2CAP_con *con);
	void (*ReceiveData) (_L2CAP_con *con, struct sk_buff *skb);

	_PSM psm; /* psm of the registered upper layer */
	struct _L2CAP_layer *next; /* linked list, for internal use */

	l2cap_proto_ops pops; /* required for channel creation */
} _L2CAP_layer;


/* register upper protocol to L2CAP layer
   returns 0 on success */
int _L2CA_register_protocol(_L2CAP_layer *ops);

/* unregister upper protocol from L2CAP layer
   returns 0 on success */
int _L2CA_unregister_protocol(_PSM psm);

/************************************************************************************************/
/* convert bluetooth addr to string */
char *BD_ADDR2str(_BD_ADDR *bda);

/* convert bluetooth CoD to string */
char *BD_CLASS2str(_BD_CLASS *bdc);

/* reverses byte order of _BD_ADDR */
void _BD_ADDR_reverse(_BD_ADDR *bda);

/* copy bdaddr to reverse orderd ethernet addr */
void _BD_ADDR2ETH_ADDR(void *eta, _BD_ADDR *bda);

/* compare _BD_ADDR with reverse ordered ethernet addr */
int comp_BD_ADDR_ETH_ADDR(_BD_ADDR *bda, void *eta);

/************************************************************************************************/
/* constants */

/* class of device field according Bluetooth Assigned Numbers (network byte order)*/
extern _BD_CLASS CLASS_OF_DEVICE_COMPUTER;
extern _BD_CLASS CLASS_OF_DEVICE_LAN_ACCESS_POINT_LOAD[8]; /* a LAN access point with 0-7 connections */

/************************************************************************************************/

#endif

