/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Muller Ulrich <ulrich.muller@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/*
  $Id: pan.c,v 1.7 2002/08/30 14:25:24 kds Exp $
 
  pan.c - PAN kernel module

  Fixes
  	Dmitry Kasatkin		- Integrated to the Affix
				- device registration/unregistration

   This module implements a PAN User, Group Ad-Hoc Network and Network Access Point.
   The role is setup at startup. Functionality:
   PAN User:
   - registers net device "ethx"
   - register BNEP layer as upper layer of L2CAP of Bluetooth device "bt0"
   - create connection to GAN/NAP
   - pass all packets from BNEP to net device and vica versa
   Group Ad-Hoc Network/Network Access Point:
   - registers net device "ethx"
   - register BNEP layer as upper layer of L2CAP of Bluetooth device "bt0"
   - accept connections from PAN Users
   - route packets between net device and all BNEP connections
   A NAP passes all packets to the net device that are not addressed to a member
   of the Piconet; in contract, a GAN dropps these packets.

   The module gets initialized and closed with the init_module() and cleanup_module()
   fuctions at the bottom. It registeres an upper L2CAP layer (pan_* functions) immediatley.
   The interface to the net device are the bt_* functions. btd_* functions are used to
   manage the connections. Connections are initiated by the connect_thread(). The main
   reason to put the connection initiation into a thread is that it runs in normal context,
   whereas so_* and bt_* functions are usually called in restricted context, e.g. interrupts
   disabled. Thus the connection thread can have more functionality, e.g. call HCI functions.

*/

/********************************************************************************/

/* kernel includes: */
#include <linux/config.h>

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/netdevice.h>
#include <linux/proc_fs.h>
#include <linux/if_arp.h>
#include <net/ip.h>
#include <linux/vmalloc.h>
#include <linux/errno.h>
#include <linux/etherdevice.h>
#include <asm/uaccess.h>

/* other includes: */
#define FILEBIT		DBPAN

#include <affix/btdebug.h>

#include "pan.h"
#include "bnep.h"

#define btmin(a,b)	(((a)<(b))?a:b)

/************************************************************************************************/
/* module definition */

static char* role;	/* my role in PAN, e.g. PANU */
static char* device;	/* my network interface, e.g. bt0 */

MODULE_PARM(role, "s");			/* string */
MODULE_PARM_DESC(role, "Role in PAN: panu (default), nap or gan");
MODULE_PARM(device, "s");		/* string */
MODULE_PARM_DESC(device, "Bluetooth device to use, default: " default_bt_device);

/************************************************************************************************/
/* global variables */

struct bt_dev		*btdev0;			/* Bluetooth device */

#ifdef CONFIG_PROC_FS
const char		proc_name[] = "pan";		/* proc file system entry name */
struct proc_dir_entry 	*proc_entry;			/* proc file system entry operations */
#endif

#define SigConnectCfm_Pos 1
#define SigConnectCfm_Neg 2
static int SigConnectCfm = 0;				/* set when pos/neg ConnectCfm is received */
static int SigDisconnectCfm = 0;			/* set when DisconnectCfm is received */
static DECLARE_WAIT_QUEUE_HEAD(SigConnectCfm_wait);	/* waked up when ConnectCfm/Ind is received */
static DECLARE_WAIT_QUEUE_HEAD(SigDisconnectCfm_wait);	/* waked up when DisconnectCfm/Ind is received */

/************************************************************************************************/
/* proc entry */

#ifdef CONFIG_PROC_FS

/* report information when user reads /proc/net/pan */
/* this method is only valid for outputs smaller than a page size */
static int proc_read(char *page, char **start, off_t off, int count, int *eof, void *data) {
#define pprintf(args...)	if(len + 80 < PAGE_SIZE) len += sprintf(page+len, args);
	struct bt_dev *btdev = btdev0; /* currently, only one device supported */
	int i, len = 0;

	/* create ouput */
	/* - role */
	pprintf("role: ");
	if (btdev->my_role == panu_role) {
		pprintf("PAN User\n");
	} else if (btdev->my_role == nap_role) {
		pprintf("Network Access Point\n");
	} else
		pprintf("Group Ad-hoc Network\n");

	/* - local filters */
	if(btdev->pf.count) {
		pprintf("local protocol filter:\n");
		for (i = 0; i < btdev->pf.count; i++)
			pprintf("- 0x%04x - 0x%04x\n", btdev->pf.protocol[i][0], btdev->pf.protocol[i][1]);
	} else {
		pprintf("no local protocol filter\n");
	}
	if(btdev->mf.count) {
		pprintf("local multicast filter:\n");
		for (i = 0; i < btdev->mf.count; i++)
			pprintf("- %s - %s\n", ETH_ADDR2str(&btdev->mf.multicast[i][0]), ETH_ADDR2str(&btdev->mf.multicast[i][1]));
	} else {
		pprintf("no local multicast filter\n");
	}

	/* remote devices */
	if (btdev->connections) {
		struct con_list *cl = btdev->connections;
		pprintf("connections:\n");
		while(cl) {
			struct bnep_layer *bl = cl->priv;
			pprintf("- %s\n", BD_ADDR2str(&cl->con->bda));

			if(bl) {
				if(!bl->setup_complete) {
					pprintf("  connection is not setup by BNEP\n");
				} else {
					/* - remote filters */
					if(bl->pf.count) {
						pprintf("  protocol filter:\n");
						for (i = 0; i < bl->pf.count; i++)
							pprintf("  - 0x%04x - 0x%04x\n", bl->pf.protocol[i][0],
								bl->pf.protocol[i][1]);
					} else {
						pprintf("  no protocol filter\n");
					}
					if(bl->mf.count) {
						pprintf("  multicast filter:\n");
						for (i = 0; i < bl->mf.count; i++)
							pprintf("  - %s - %s\n", ETH_ADDR2str(&bl->mf.multicast[i][0]),
								ETH_ADDR2str(&bl->mf.multicast[i][1]));
					} else {
						pprintf("  no multicast filter\n");
					}
				}
			} else {
				pprintf("  connection is not configured by L2CAP\n");
			}
			cl = cl->next;
		}
	} else {
		pprintf("no connections\n");
	}

//done:
	/* return output */
	if (off >= len) {
		*eof = 1;
		return 0;
	}

	*start = page + off;
	if (count >= len - off) {
		*eof = 1;
	}
	return( btmin(count, len - off) );
}
#endif

/************************************************************************************************/
/* helper for bt_dev */

int __init bt_init(struct net_device *dev); /* forward declaration */

/* create new bt_dev struct for Bluetooth device "name", init hci handle and bdaddr, register netdev
   return NULL on faiure */
struct bt_dev *btd_init(Role my_role, char* name) {
	struct bt_dev *btdev = (struct bt_dev *) vmalloc (sizeof(struct bt_dev));

	if(!btdev) {
		ERR("out of memory\n");
		return NULL;
	}

	memset(btdev, 0, sizeof(struct bt_dev));

	init_MUTEX(&btdev->thread_sem);

	strncpy(btdev->name, name, IFNAMSIZ);

	btdev->name[IFNAMSIZ - 1] = '\0'; /* if name was too long, it may not have termination */
	btdev->my_role = my_role;

	btdev->hci = _hci_open(btdev->name); /* returns NULL on error */
	if(!btdev->hci) {
		ERR("cannot open BT device %s\n", btdev->name);
		goto fail;
	}

	if( _HCI_ReadBDAddr(btdev->hci, &btdev->bdaddr) != 0) { /* error */
		DBG_F(DBG_MISC, "error reading BD_ADDR\n");
		goto fail_hci_open;
	}
	INFO("PAN BD_ADDR: %s\n", BD_ADDR2str(&btdev->bdaddr));

	/* create net device "ethx" */
	//strcpy(btdev->net_dev.name, "eth%d"); /* kernel will assign a number */
	if (dev_alloc_name(&btdev->net_dev, "eth%d") < 0)
		goto fail_hci_open;

	btdev->net_dev.init = bt_init;
	btdev->net_dev.priv = (void*) btdev;

	_BD_ADDR2ETH_ADDR(&btdev->net_dev.dev_addr, &btdev->bdaddr);

        if (register_netdevice(&btdev->net_dev) != 0) { /* register network device */
		ERR("unable to register network device %s\n", btdev->net_dev.name);
		goto fail_hci_open;
	}
	INFO("PAN net device: %s\n", btdev->net_dev.name);

	return btdev;

fail_hci_open:
	_hci_close(btdev->hci);
fail:
	vfree(btdev);
	return NULL;
}

/* frees memory of btdev and closes hci handle, unregister netdev */
void btd_close(struct bt_dev* btdev) {
	if(!btdev)
		return;

	if(btdev->hci)
		_hci_close(btdev->hci);

	if(btdev->net_dev.name) /* net_dev is valid */
		unregister_netdevice(&btdev->net_dev);

	vfree(btdev);
}

/* add new connection to btdev->connections
   forward connection to bnep layer; when running as a gan/nap, update class of device,
   returns 0 on success */
int btd_connection_add(struct bt_dev *btdev, _L2CAP_con *con, State state) {
	struct con_list *cl = (struct con_list*) vmalloc(sizeof(struct con_list));

	DBG_ENTER(DBG_MISC);
	if(!cl) return -ENOMEM; /* out of memory */
	if(btdev->connections_counter >= 7) { /* should not be possible */
		DBG_F(DBG_MISC, "too many connections!\n");
		vfree(cl);
		return -EBUSY;
	}

	memset(cl, 0, sizeof(struct con_list));

	con->priv = (void*) cl;
	cl->con = con;
	cl->btdev = btdev;
	cl->state = state;

	/* add to head of connection list */
	btdev->connections_counter++;
	cl->prev = NULL;
	cl->next = btdev->connections;
	btdev->connections = cl;
	if(cl->next)
		cl->next->prev = cl;
/* connection is always configured now (new Affix L2CAP API) */
//	if(state == configured)
		bnep_init(cl);

	DBG_EXIT(DBG_MISC);
	return 0;
}

/* change state of connection,
   forward connection to bnep layer */
void btd_connection_change_state(_L2CAP_con *con, State state) {
	struct con_list *cl = (struct con_list*) con->priv;

	DBG_ENTER(DBG_MISC);
	if(!cl || cl->state == state)
		return;

	cl->state = state;

	/* we dont expect to change state from configured to unconfigured */
	if(state == configured)
		bnep_init(cl);

	DBG_EXIT(DBG_MISC);
}

/* remove connection from btdev->connections
   forward connection to bnep layer; when running as a gan/nap, update class of device */
void btd_connection_remove(_L2CAP_con *con) {
	struct con_list *cl = (struct con_list*) con->priv;
	struct bt_dev *btdev = cl->btdev;

	DBG_ENTER(DBG_MISC);
	if(!cl)
		return;
	con->priv = NULL;

	bnep_close(cl);

	if (cl->next)
		cl->next->prev = cl->prev;
	if (cl->prev)
		cl->prev->next = cl->next;
	else
		btdev->connections = cl->next; /* this was the first entry */
	vfree(cl);

	btdev->connections_counter--;

	DBG_EXIT(DBG_MISC);
}

/* close all open connections of given device */
void btd_close_connections(struct bt_dev *btdev) {
	while(btdev->connections_counter) {
		DECLARE_WAITQUEUE(wait, current);
		_L2CAP_con *con = btdev->connections->con;

		/* it may happen that we receive a DisconnectInd while waiting
		   for DisconnectCfm, possibly of another connection.
		   in this case we should avoid to send DisconnectReq again.
		   This can be checked with con->state == _L2CAP_CON_DISCONNECT
		*/

		DBG_F(DBG_MISC, "closing connection to %s\n", BD_ADDR2str(&con->bda));
		SigDisconnectCfm = 0;

		if( (con->state & _L2CAP_CON_DISCONNECT) == 0)
			if(_L2CA_DisconnectReq(con)) { /* disconnect fails -> connection not available */
				btd_connection_remove(con);
				continue; /* dont wait for DisconnectCfm */
			}

		add_wait_queue(&SigDisconnectCfm_wait, &wait);
		current->state = TASK_INTERRUPTIBLE;

		DBG_F(DBG_MISC, "waiting for SigDisconnectCfm\n");
		while(!SigDisconnectCfm)
		{
			current->state = TASK_RUNNING; // unecessary???
			schedule();
		}

		current->state = TASK_RUNNING;
		remove_wait_queue(&SigDisconnectCfm_wait, &wait);
	}
	DBG_EXIT(DBG_MISC);
}

/* this function set the class of device/scan activity according to the current number of connections */
void btd_connections_update(struct bt_dev *btdev) {
	int counter = btdev->connections_counter;

	if(counter > 7)
		counter = 7; /* class of device is defined for up to 7 connections */

	if(counter == btdev->connections_counter_old)
		return;

	if(btdev->my_role != panu_role) { /* switch */
		DBG_F(DBG_MISC, "setting CoD to load %d\n", counter);
		_HCI_WriteClassOfDevice(btdev->hci, &CLASS_OF_DEVICE_LAN_ACCESS_POINT_LOAD[counter]);
		if(counter == 7) /* maximum number of slaves reached */
			_HCI_WriteScanEnable(btdev->hci, _HCI_Scan_off); /* disable page/inquiry scan */
		else if(btdev->connections_counter_old == 7) /* maximum number of slaves left */
			_HCI_WriteScanEnable(btdev->hci, _HCI_Scan_Inquiry | _HCI_Scan_Page);
	}

	btdev->connections_counter_old = counter;
}

/************************************************************************************************/
/* connection initiation */

/* thread that connects to other devices until thread_stop != 0 */
int connect_active_thread(void *p) {
 #define Connections_wanted 1 /* a pan user only wants a connection to one gan/nap */
#define Max_Num_Responses 100
#define Max_Connection_Trys 5
	struct bt_dev		*btdev = p;
	_HCI_Inquiry_Item	inquiry_list[Max_Num_Responses];	/* list of inquiry results */
	int			connect_counter[Max_Num_Responses];	/* list of connection trys to each inquiry device */

	DBG_ENTER(DBG_THREAD);
	while (!btdev->thread_stop) { /* loop until thread_stop is set */
		DECLARE_WAITQUEUE(wait, current);
		__u8 Num_Responses_real;
		int i, ret, counter, Num_Responses;

		INFO("PAN: inquiry...\n");
		ret = _HCI_Inquiry(btdev->hci, 10, Max_Num_Responses, inquiry_list, &Num_Responses_real);

		if (ret != 0) {
			ERR("HCI inquiry error\n");
			btdev->thread_stop = 1;
			break;
		}

		/* filter inquiry result with CoD */
		i = 0;
		Num_Responses = Num_Responses_real;
		while( i < Num_Responses) {
			int ok = 0;
			int j;
			for (j = 0; j < 8; j++)
				if (!memcmp(&inquiry_list[i].bdc, CLASS_OF_DEVICE_LAN_ACCESS_POINT_LOAD[j], 3))
					ok = 1;
			if(!ok) {
				DBG_F(DBG_MISC, "skipping %s/%s\n", BD_ADDR2str(&inquiry_list[i].bda), BD_CLASS2str(&inquiry_list[i].bdc));
				memcpy(&inquiry_list[i], &inquiry_list[i + 1], (Num_Responses - i) * sizeof(_HCI_Inquiry_Item));
				Num_Responses--;
			} else {
				i++;
			}
		}

		INFO("PAN: found %d devices (%d total)\n", Num_Responses, Num_Responses_real);
		memset(&connect_counter, 0, sizeof(connect_counter));

		/* now try to connect to up to Connections_wanted devices */
		i = 0;
		counter = 0; /* counts connection trys */
		while (!btdev->thread_stop && (i < Num_Responses ) && (btdev->connections_counter < Connections_wanted)) {
			DBG_F(DBG_THREAD, "connecting to %s/%s (device %d/%d, try %d/%d)\n", BD_ADDR2str(&inquiry_list[i].bda),
				BD_CLASS2str(&inquiry_list[i].bdc), i+1, Num_Responses, connect_counter[i] + 1, Max_Connection_Trys);

			SigConnectCfm = 0;
			ret = _L2CA_ConnectReq(&inquiry_list[i].bda, BNEP_PSM);
			if(ret == 0) { /* now wait until connection is confirmed or cancelled */
				DBG_F(DBG_THREAD, "waiting for SigConnectCfm\n");
				add_wait_queue(&SigConnectCfm_wait, &wait);
				current->state = TASK_INTERRUPTIBLE;

				while(!btdev->thread_stop && !SigConnectCfm) {
					if (signal_pending(current))
						btdev->thread_stop = 1;
					
		        		current->state = TASK_RUNNING;  // unecessary?
					schedule();
				}

		        	current->state = TASK_RUNNING;
				remove_wait_queue(&SigConnectCfm_wait, &wait);
				DBG_F(DBG_THREAD, "waiting finished\n");
			}
			btd_connections_update(btdev);

			/* determine next connection partner */
			if(SigConnectCfm == SigConnectCfm_Pos)
				connect_counter[i] = Max_Connection_Trys;
			else
				connect_counter[i]++;

			do {
				i++;
			} while ( i < Num_Responses && connect_counter[i] == Max_Connection_Trys );

			if(i == Num_Responses) {
				i = 0;
				while ( i < Num_Responses && connect_counter[i] == Max_Connection_Trys )
					i++;
			}
		}

		/* Now we have tried to connect to all devices from last inquiry OR have connected to Connections_wanted devices.
		   Before we start the next inquiry, we wait until all connections are down. */
		DBG_F(DBG_THREAD, "waiting for all connections to close\n");
		add_wait_queue(&SigDisconnectCfm_wait, &wait);
		current->state = TASK_INTERRUPTIBLE;

		while(!btdev->thread_stop && (btdev->connections_counter)) {
			if (signal_pending(current))
				btdev->thread_stop = 1;

			current->state = TASK_RUNNING;
			schedule();
			btd_connections_update(btdev);
		}
		current->state = TASK_RUNNING;
		remove_wait_queue(&SigDisconnectCfm_wait, &wait);
		DBG_F(DBG_THREAD, "no connections available\n");
	}

	up(&btdev->thread_sem); /* thread going down */
	DBG_EXIT(DBG_THREAD);
	return 0;
}

/* thread that updates CoD/scan activity until thread_stop != 0 */
int connect_passive_thread(void *p) {
	struct bt_dev		*btdev = p;
	DECLARE_WAITQUEUE(wait1, current);
	DECLARE_WAITQUEUE(wait2, current);

	DBG_ENTER(DBG_THREAD);



	add_wait_queue(&SigConnectCfm_wait, &wait1);
	add_wait_queue(&SigDisconnectCfm_wait, &wait2);
	current->state = TASK_INTERRUPTIBLE;

	while(!btdev->thread_stop) {	
		if (signal_pending(current))
			{
				btdev->thread_stop = 1;				
				current->state = TASK_RUNNING;  //fixed!!
			}
	
		schedule();
		btd_connections_update(btdev);
		current->state = TASK_INTERRUPTIBLE;		//fixed!!
	}

     	current->state = TASK_RUNNING;
	remove_wait_queue(&SigConnectCfm_wait, &wait1);
	remove_wait_queue(&SigDisconnectCfm_wait, &wait2);

	up(&btdev->thread_sem); /* thread going down */
	DBG_EXIT(DBG_THREAD);
	return 0;
}

/************************************************************************************************/
/* start + stop panu/access point */

void access_point_start(struct bt_dev *btdev) {
	/* enable Inquiry Scan + Page Scan and wait for panu to connect */
	_HCI_WriteScanEnable(btdev->hci, _HCI_Scan_Inquiry | _HCI_Scan_Page);

	btdev->thread_stop = 0;
	down(&btdev->thread_sem); /* assume thread is running: set thread running lock */
	if (kernel_thread(connect_passive_thread, btdev, 0) < 0) {
		up(&btdev->thread_sem); /* unlock */
		ERR("PAN: unable to start kernel thread\n");
	}
}

void panu_start(struct bt_dev *btdev) {
	btdev->thread_stop = 0;
	down(&btdev->thread_sem); /* assume thread is running: set thread running lock */
	if (kernel_thread(connect_active_thread, btdev, 0) < 0) {
		up(&btdev->thread_sem); /* unlock */
		ERR("PAN: unable to start kernel thread\n");
	}
}

void access_point_stop(struct bt_dev *btdev) {
	_HCI_WriteScanEnable(btdev->hci, _HCI_Scan_off);
	/* stop thread */
	btdev->thread_stop = 1;				/* signal thread to stop */
	wake_up_interruptible(&SigConnectCfm_wait);	/* wake thread if it is sleeping */
	wake_up_interruptible(&SigDisconnectCfm_wait);	/* wake thread if it is still sleeping */
	down(&btdev->thread_sem);			/* wait for thread to finish */
	up(&btdev->thread_sem);				/* but dont block restart */

	/* close all open connections */
	btd_close_connections(btdev);
}

void panu_stop(struct bt_dev *btdev) {
	/* stop thread */
	btdev->thread_stop = 1;				/* signal thread to stop */
	wake_up_interruptible(&SigConnectCfm_wait);	/* wake thread if it is sleeping */
	wake_up_interruptible(&SigDisconnectCfm_wait);	/* wake thread if it is still sleeping */
	down(&btdev->thread_sem);			/* wait for thread to finish */
	up(&btdev->thread_sem);				/* but dont block restart */

	/* close all open connections */
	btd_close_connections(btdev);
}

/************************************************************************************************/
/* net device functions: called from network layer */

int bt_open(struct net_device *dev) {
	struct bt_dev *btdev = (struct bt_dev*) dev->priv;

	DBG_ENTER(DBG_NET);
	MOD_INC_USE_COUNT; /* module is now in use, not removable */

	/* set device flag 'up' */
	netif_start_queue(dev);

	/* prepare bluetooth connection */
	if(btdev->my_role == panu_role)
		panu_start(btdev);
	else
		access_point_start(btdev);

	DBG_EXIT(DBG_NET);
	return 0;
}

int bt_close(struct net_device *dev) {
	struct bt_dev *btdev = (struct bt_dev*) dev->priv;

	DBG_ENTER(DBG_NET);

	/* set device flag 'down' */
	netif_stop_queue(dev);

	/* prepare bluetooth connection */
	if(btdev->my_role == panu_role)
		panu_stop(btdev);
	else
		access_point_stop(btdev);

	MOD_DEC_USE_COUNT; /* module can now be removed */

	DBG_EXIT(DBG_NET);
	return 0;
}

int bt_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd) {
	struct bt_dev *btdev = (struct bt_dev*) dev->priv;

	switch (cmd) {
	case SIOCSFILTERPROTOCOL: { /* set protocol filter */
		protocol_filter pf;
		if(copy_from_user(&pf, ifr->ifr_ifru.ifru_data, sizeof(pf)))
			return -EFAULT;
		if(pf.count > PROTOCOL_FILTER_MAX)
			return -ENOMEM;
		bnep_set_filter_protocol(btdev, &pf);
		return 0; }

	case SIOCSFILTERMULTICAST: { /* set multicast filter */
		multicast_filter mf;
		if(copy_from_user(&mf, ifr->ifr_ifru.ifru_data, sizeof(mf)))
			return -EFAULT;
		if(mf.count > MULTICAST_FILTER_MAX)
			return -ENOMEM;
		bnep_set_filter_multicast(btdev, &mf);
		return 0; }

	case SIOCGFILTERPROTOCOL: /* get protocol filter */
		if(copy_to_user(ifr->ifr_ifru.ifru_data, &btdev->pf, sizeof(protocol_filter)))
			return -EFAULT;
		return 0;

	case SIOCGFILTERMULTICAST: /* get multicast filter */
		if(copy_to_user(ifr->ifr_ifru.ifru_data, &btdev->mf, sizeof(multicast_filter)))
			return -EFAULT;
		return 0;
	}

	return -EOPNOTSUPP;
}

/* Set or clear the multicast filter for this device */
void bt_set_multicast_list(struct net_device *dev) {
	struct bt_dev *btdev = (struct bt_dev*) dev->priv;
	multicast_filter mf;
	struct dev_mc_list *dmi = dev->mc_list;
	ETH_ADDR *addr;
	int i;

	mf.count = 0;
	if ((dev->flags & IFF_PROMISC) || (dev->flags & IFF_ALLMULTI) || (dev->mc_count > MULTICAST_FILTER_MAX)) {
		DBG_F(DBG_MISC, "all multicast mode\n");
		bnep_set_filter_multicast(btdev, &mf);
		return;
	}

	/* Add addresses */
	DBG_F(DBG_MISC, "%d multicast addresses:\n", dev->mc_count);
	for (i = 0; i < dev->mc_count; i++) {
		addr = (ETH_ADDR*) dmi->dmi_addr;
		DBG_F(DBG_MISC, "- %s\n", ETH_ADDR2str(addr));

		memcpy(&mf.multicast[i][0], addr, 6);
		memcpy(&mf.multicast[i][1], addr, 6); /* start range == stop range */
		dmi   = dmi->next;
	}
	mf.count = dev->mc_count;
	bnep_set_filter_multicast(btdev, &mf);
}

struct net_device_stats *bt_get_stats(struct net_device *dev) {
	struct bt_dev *btdev = (struct bt_dev*) dev->priv;

	return &btdev->stats;
}

/* receive a packet from upper layer */
int bt_xmit(struct sk_buff *skb, struct net_device *dev) {
	struct bt_dev *btdev = (struct bt_dev*) dev->priv;
	int result = -1, len;

	len = skb->len;
	result = bnep_process_upper(btdev, skb);

	if (result == 0) {
		btdev->stats.tx_packets++;
		btdev->stats.tx_bytes += len;
		btdev->net_dev.trans_start = jiffies;
	} else {
		btdev->stats.tx_dropped++;
	}

	DBG_EXIT(DBG_TX_DATA);
	return 0;
}

/* called from ourself to transmit packet to net device */
void bt_receive(struct sk_buff *skb, struct bt_dev *btdev) {
	DBG_ENTER(DBG_NET);

	skb->dev = &btdev->net_dev;

	btdev->stats.rx_packets++;
	btdev->stats.rx_bytes += skb->len;
	skb->dev->last_rx = jiffies;

	DBG_F(DBG_NET, "received %d \n", skb->len);
	netif_rx(skb);   /* Eat it! */

	DBG_EXIT(DBG_NET);
}

/* initialize virtual network device 'ethx' */
int __init bt_init(struct net_device *dev) {
	struct bt_dev *btdev = (struct bt_dev*) dev->priv;

	dev->open		= bt_open;
	dev->stop		= bt_close;
	dev->do_ioctl		= (void*)bt_ioctl;
	dev->get_stats		= bt_get_stats;
	dev->hard_start_xmit	= bt_xmit;

	if(btdev->my_role == panu_role) /* get multicast filter from net device */
		dev->set_multicast_list = bt_set_multicast_list;

	ether_setup(dev); 	/* register some kernel internal default ethernet functions (net_init.c) */
	dev->set_mac_address 	= NULL;	/* overwrite default ethernet function */

	dev->hard_header_len	= BNEP_hard_header_len + _L2CAP_hard_header_len;
	dev->tx_queue_len	= 25; /* maximum number of frames the kernel should queue for the device.
					 Typically this is around 100 for Ethernet */
	dev_init_buffers(dev); // TODO: necessary?

	return 0;
}

/************************************************************************************************/
/* PAN L2CAP upper protocol layer */

/* connection establishment, configuration and disconnection are done with four signals each,
   according to BT spec.:

              initiator                         responder
               |   A                             |   A
               |   | confirmation      response  |   |
       request |   |                             V   | indication
               |   -------------------------------   |
               V                                     |
               ---------------------------------------

   After a connection is established, it must be configured to change the L2CAP MTU from 672 to
   1691 (BNEP MTU) octets. The configuration only applys to the direction of its request
   (MTU: max. size the initiator of a ConfReq can accept).
   The initiator will start ConfigReq as soon as he receives ConnectCfm. The responder will start
   ConfigReq after ConfigInd. The state of a L2CAP channel is saved in its corresponding
   struct con_list. Data may be sent only if state == configured.
*/

void pan_ConnectInd (_L2CAP_con *con) {
	struct bt_dev *btdev = btdev0; /* currently, only one device supported */

	INFO("PAN: connection request from %s\n", BD_ADDR2str(&con->bda));

	if(btdev->my_role == panu_role) /* I am not a gan/nap */
		goto fail;

	/* if gan/nap is not "up", refuse connection */
	if ((test_bit(__LINK_STATE_START, &btdev->net_dev.state) == 0)) {
		DBG_F(DBG_MISC, "connection refused - net device is not running\n");
		goto fail;
	}
	
	/* if master-slave swich fails, refuse connection */
	/* thread problem, now use the link policy of L2CAP in Affix stack! (added Nicolas Delest) */

/*	if (_HCI_SwitchRole(btdev->hci, &con->bda, _HCI_Role_Master) != 0) {
		WARN("PAN: role switch failed, running as slave\n");
		//goto fail;
	}
*/
	/* accept connection */
	/* not needed anymore */
/*	if (btd_connection_add(btdev, con, unconfigured) != 0)
		goto fail;
*/
	/* accept connection but do the channel configuration before */
	con->mtu_out=PAN_MTU;

	_L2CA_ConnectRsp(con, _L2CAP_ConnectRsp_Res_Success, 0);
	/* signal connect thread  */
	wake_up_interruptible(&SigConnectCfm_wait); /* waked up when ConnectionEvent is set */
	return;

fail:
	_L2CA_ConnectRsp(con, _L2CAP_ConnectRsp_Res_NoSrc, 0);
}

void pan_ConnectCfm (_L2CAP_con *con, __u16 result, __u16 status) {
	struct bt_dev *btdev = btdev0; /* currently, only one device supported */

	DBG_ENTER(DBG_NET);

	switch(result) {
	case _L2CAP_ConnectCfm_Res_Success:
		INFO("PAN: connection to %s successful\n", BD_ADDR2str(&con->bda));

/* 		state is configured now because of new Affix L2CAP API 		*/
		
		con->state=_L2CAP_CON_CONFIGURED;

/*		if (btd_connection_add(btdev, con, unconfigured) != 0) {	*/
/* 		state is configured now because of new Affix L2CAP API 		*/
		if (btd_connection_add(btdev, con, configured) != 0) {
			DBG_F(DBG_MISC, "internal error, closing\n");
			_L2CA_DisconnectReq(con);
			SigConnectCfm = SigConnectCfm_Neg;
			break;
		}
		SigConnectCfm = SigConnectCfm_Pos;
		break;

	case _L2CAP_ConnectCfm_Res_Pending:
		DBG_F(DBG_NET, "connection to %s is pending\n", BD_ADDR2str(&con->bda));
		return;
		break;

	default:
		ERR("PAN: connection to %s failed\n", BD_ADDR2str(&con->bda));
		SigConnectCfm = SigConnectCfm_Neg;
	}

	/* signal connect thread  */
	wake_up_interruptible(&SigConnectCfm_wait); /* waked up when ConnectionEvent is set */
	DBG_EXIT(DBG_NET);
}

void pan_ConfigInd (_L2CAP_con *con) {
	struct con_list *cl = (struct con_list*) con->priv;
	struct bt_dev *btdev = cl->btdev;

	DBG_ENTER(DBG_NET);

	if(con->mtu_out < PAN_MTU) { /* wrong MTU */
		DBG_F(DBG_MISC, "wrong MTU (%d instead %d), sending negative ConfigRsp\n", con->mtu_out, PAN_MTU);
		_L2CA_ConfigRsp(con, PAN_MTU, 0, NULL, _L2CAP_ConfigRsp_Res_Failure_unacceptable);
		return;
	}

	DBG_F(DBG_NET, "sending positive ConfigRsp\n");
	/* the responder of a positive ConfigReq must always return the MTU it will use */
	_L2CA_ConfigRsp(con, PAN_MTU, 0, NULL, _L2CAP_ConfigRsp_Res_Success);

	/* now connection responder configures channel */
	if(btdev->my_role != panu_role) {
		DBG_F(DBG_NET, "sending ConfigReq\n");
		_L2CA_ConfigReq(con, PAN_MTU, 0, NULL, PAN_LINK_TIMEOUT);
	}

	if((con->state & _L2CAP_CON_CONFIGURED) == _L2CAP_CON_CONFIGURED) { /* connection is now configured */
		DBG_F(DBG_NET, "connection configured\n");
		btd_connection_change_state(con, configured);
	}

	DBG_EXIT(DBG_NET);
}

void pan_ConfigCfm (_L2CAP_con *con, __u16 result) {
	if(result != _L2CAP_ConfigCfm_Res_Success)
		goto fail;

	if((con->state & _L2CAP_CON_CONFIGURED) == _L2CAP_CON_CONFIGURED) { /* connection is now configured */
		INFO("PAN: connection configured\n");
		btd_connection_change_state(con, configured);
	}

	return;

fail: /* cannot configure channel, so close it */
	DBG_F(DBG_MISC, "connection configuration failed, closing\n");
	_L2CA_DisconnectReq(con);
}

void pan_DisconnectInd (_L2CAP_con *con) {
	DBG_ENTER(DBG_NET);
	INFO("PAN: connection to %s closed\n", BD_ADDR2str(&con->bda));

	btd_connection_remove(con);
	_L2CA_DisconnectRsp(con);

	/* signal disconnect task that connection is closed, se comment in btd_close_connections */
	DBG_F(DBG_THREAD, "sending SigDisconnectCfm\n");
	SigDisconnectCfm = 1;
	wake_up_interruptible(&SigDisconnectCfm_wait);

	DBG_EXIT(DBG_NET);
}

void pan_DisconnectCfm (_L2CAP_con *con) {
	DBG_ENTER(DBG_NET);
	INFO("PAN: connection to %s closed\n", BD_ADDR2str(&con->bda));

	btd_connection_remove(con);

	/* signal disconnect task that connection is closed */
	DBG_F(DBG_THREAD, "sending SigDisconnectCfm\n");
	SigDisconnectCfm = 1;
	wake_up_interruptible(&SigDisconnectCfm_wait);

	DBG_EXIT(DBG_NET);
}

void pan_ReceiveData (_L2CAP_con *con, struct sk_buff *skb) {
	struct con_list *cl = (struct con_list*) con->priv;
	bnep_process_lower(cl, skb);
}

_L2CAP_layer		l2cap_layer = { pan_ConnectInd, pan_ConnectCfm, pan_ConfigInd, pan_ConfigCfm,
				pan_DisconnectInd, pan_DisconnectCfm, pan_ReceiveData, BNEP_PSM };

/************************************************************************************************/

/*
 * Device registration/unregistration notifier
 */

static Role my_role = default_role;

int pan_init_device(char *name)
{
	int	err = 0;

	DBFENTER;
	
	if (btdev0)
		return 0;

	DBPRT(DBINIT, "init device: %s\n", name);
	btdev0 = btd_init(my_role, name); /* the default device */
	if(!btdev0) {
		ERR("No Bluetooth devices found\n");
		err = -ENODEV;
		goto fail_init;
	}

	/* set class of device and disable page/inquiry scan until net device is "up" */
	_HCI_WriteScanEnable(btdev0->hci, _HCI_Scan_off);
	if(btdev0->my_role == panu_role)
		_HCI_WriteClassOfDevice(btdev0->hci, &CLASS_OF_DEVICE_COMPUTER);
	else
		_HCI_WriteClassOfDevice(btdev0->hci, &CLASS_OF_DEVICE_LAN_ACCESS_POINT_LOAD[0]); /* no connections */

	DBFEXIT;
	return 0;
fail_init:
	btd_close(btdev0);
	btdev0 = NULL;
	return err;

}

int pan_exit_device(char *name)
{

	if (strncmp(btdev0->name, name, IFNAMSIZ) == 0) {
		btd_close(btdev0);
		btdev0 = NULL;
	}

	return 0;
}

int pan_event(struct notifier_block *nb, unsigned long event, void *arg)
{
	hci_struct	*hci = arg;

	switch (event) {
	case NETDEV_REGISTER:
		break;

	case NETDEV_UNREGISTER:
		break;

	case NETDEV_UP:
		pan_init_device(hci->name);
		break;

	case NETDEV_DOWN:
		pan_exit_device(hci->name);
		break;

	default:
		break;
	}

	return NOTIFY_DONE;
}

struct notifier_block pan_notifier_block = {
	notifier_call:	pan_event,
};


/* init + close module */
extern struct proc_dir_entry	*proc_affix;


int init_pan(void)
{
	int		err = -EINVAL;
	hci_struct	*hci = NULL;
	struct net_device	*dev;

	DBFENTER;

	/* examine module parameters */
	btdev0 = NULL;

	/* evaluate "role" */
	if(role) {
		if (!strcmp(role, "panu")) {
			my_role = panu_role;
		} else if (!strcmp(role, "nap")) {
			my_role = nap_role;
		} else if (!strcmp(role, "gan")) {
			my_role = gan_role;
		} else {
			ERR("unknown role '%s':\n", role);
			ERR("role must be 'panu', 'nap' or 'gan'\n");
			goto fail_init;
		}
	}
	affix_register_notifier(&pan_notifier_block);

	/* register upper protocol to L2CAP */
	if (_L2CA_register_protocol(&l2cap_layer)) {
		ERR("PAN: error registering L2CAP protocol");
		goto fail_init;
	}

#ifdef CONFIG_PROC_FS
	/* install proc file entry */
	if ((proc_entry = create_proc_entry(proc_name, S_IFREG | S_IRUGO, proc_affix)) != NULL) {
		proc_entry->read_proc = proc_read;
		proc_entry->write_proc = NULL;
	}
#endif

	/* try to enable existing devices */
	if (device) {
		dev = dev_get_by_name(device);
		if (dev)
			hci = DEV_HCI(dev);
	} else {
		hci = hci_select();
	}

	if (hci) {
		pan_init_device(hci->name);
		hci_put(hci);
	}
	
	DBFEXIT;
	return 0; /* success */

fail_init:
	ERR("init_module failed\n");
	return err; /* error */
}

void exit_pan(void)
{
	DBFENTER;

	/* stop devices first */
	if (btdev0) 
		pan_exit_device(btdev0->name);

#ifdef CONFIG_PROC_FS
	if (proc_entry) /* unregister proc entry */
		remove_proc_entry(proc_name, proc_net);
#endif

	_L2CA_unregister_protocol(BNEP_PSM);
	affix_unregister_notifier(&pan_notifier_block);
	DBFEXIT;
}

#ifdef MODULE
module_init(init_pan);
module_exit(exit_pan);
#endif

#ifdef MODULE

MODULE_AUTHOR("Unknown fellow");
MODULE_DESCRIPTION("Bluetooth Personal Area Networking Profile");
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 4, 15)
MODULE_LICENSE("GPL");
#endif

#endif // MODULE

EXPORT_NO_SYMBOLS;

