/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Muller Ulrich <ulrich.muller@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/*
  $Id: pan.h,v 1.3 2002/07/15 14:05:48 kds Exp $
 
  pan.c - PAN kernel module

   This module implements a PAN User, Group Ad-Hoc Network and Network Access Point.
   The role is setup at startup. Functionality:
   PAN User:
   - registers net device "ethx"
   - register BNEP layer as upper layer of L2CAP of Bluetooth device "bt0"
   - create connection to GAN/NAP
   - pass all packets from BNEP to net device and vica versa
   Group Ad-Hoc Network/Network Access Point:
   - registers net device "ethx"
   - register BNEP layer as upper layer of L2CAP of Bluetooth device "bt0"
   - accept connections from PAN Users
   - route packets between net device and all BNEP connections
   A NAP passes all packets to the net device that are not addressed to a member
   of the Piconet; in contract, a GAN dropps these packets.

   The module gets initialized and closed with the init_module() and cleanup_module()
   fuctions at the bottom. It registeres an upper L2CAP layer (pan_* functions) immediatley.
   The interface to the net device are the bt_* functions. btd_* functions are used to
   manage the connections. Connections are initiated by the connect_thread(). The main
   reason to put the connection initiation into a thread is that it runs in normal context,
   whereas so_* and bt_* functions are usually called in restricted context, e.g. interrupts
   disabled. Thus the connection thread can have more functionality, e.g. call HCI functions.

*/

/************************************************************************************************/
/* includes */

#ifndef _PAN_H
#define _PAN_H

#include <linux/if.h>
#include <linux/netdevice.h>

#include <affix/bluetooth.h>

#include "bt_driver.h"
#include "debug.h"

/************************************************************************************************/
/* constants */

#define BNEP_PSM 0x000F		// added according to Bluetooth Assign numbers


#define PAN_MTU		1691

#define PAN_LINK_TIMEOUT	0x7D00 /* default value, 20 secs */

/************************************************************************************************/
/* typedefs */

typedef enum {panu_role, nap_role, gan_role, default_role = panu_role} Role;
/* PAN devices: PAN User, Network Access Point, Group Ad-hoc Network */

typedef enum {unconfigured, configured} State;
/* status of L2CAP connection */

/* This data structure represents a PAN bluetooth interface */
struct bt_dev {
	char			name[IFNAMSIZ];		/* name of bluetooth device, e.g. bt0 */
	_BD_ADDR		bdaddr;			/* hardware device address */
	_HCI_handle		*hci;			/* handle to access device */
	Role			my_role;		/* role in PAN */

	struct net_device	net_dev;		/* network device */
	struct net_device_stats	stats;			/* statistics of network device */

	struct con_list		*connections;		/* list of all active connections */
	int			connections_counter;	/* counter of all active connections */
	int			connections_counter_old;/* counter of all active connections at last CoD/Scan update */

	int			thread_stop;		/* indicates inquiry thread to terminate */
	struct semaphore	thread_sem;		/* indicates if inquiry thread is running */

	protocol_filter		pf;			/* settings of local protocol filter */
	multicast_filter	mf;			/* settings of local multicast filter */
};

/* double linked list of all active connections, required for a clean shutdown */
struct con_list {
	_L2CAP_con		*con;		/* L2CAP connection handle to identify connection */
	struct bt_dev		*btdev;		/* link to device that links to this connection list */
	State			state;		/* connection state */
	void			*priv;		/* internal data of bnep layer */

	struct con_list		*next;
	struct con_list		*prev;
};

/************************************************************************************************/
/* functions */

/* called to transmit packet to net device */
void bt_receive(struct sk_buff *skb, struct bt_dev *btdev);

/* to transmit packet to bluetooth:
int _L2CA_DataWrite(_L2CAP_con, skb);
*/

/************************************************************************************************/

#endif


