/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: bty.c,v 1.5 2002/09/03 09:15:09 kds Exp $

   BTY - RF Virutal terminal Driver for RFCOMM

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
                Imre Deak <ext-imre.deak@nokia.com>
*/		

/* The following prevents "kernel_version" from being set in this file. */
#define __NO_VERSION__

#include <linux/config.h>

/* Module related headers, non-module drivers should not include */
#include <linux/module.h>
#include <linux/init.h>

/* Standard driver includes */
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>

#include <linux/interrupt.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <linux/proc_fs.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/tty.h>
#include <linux/serial.h>

#include <asm/uaccess.h>
#include <asm/bitops.h>

#include <linux/skbuff.h>
#include <net/sock.h>

#define FILEBIT	DBBTY

/* Local Includes */
#include <affix/types.h>
#include <affix/btdebug.h>
#include <affix/btlist.h>

#include <affix/rfcomm.h>
#include "bty.h"


/*================================================================*/
/* "Driver global" definitions */

int			bty_maxdev = BTY_MAX_TTY;
struct bty_table	*bty_table = NULL;

struct tty_driver	bty_drv;	/* our driver structure	*/

static int 			tty_refcount;
static struct tty_struct 	**tty_table;
static struct termios 		**tty_termios;
static struct termios 		**tty_termios_locked;

static char *revision_date = " 14.08.2001 ";

/*
  Memory management
*/


long skb_count = 0;

/* 
 * Write buffer destructor automatically called from kfree_skb. 
 */
void bty_wfree(struct sk_buff *skb)
{
	bty_struct *bty = (void*)skb->sk;

	DBFENTER;

	skb_count--;

	/* In case it might be waiting for more memory. */
	atomic_sub(skb->truesize, &bty->wmem_alloc);
	if (bty_writeable(bty)) {
		if (bty->flags & ASYNC_INITIALIZED) {
			bty_control_ind(bty->con, TX_START);
		}
	}
	bty_put(bty);

	DBFEXIT;
}

void bty_set_owner_w(struct sk_buff *skb, bty_struct *bty)
{
	bty_hold(bty);
	skb->sk = (void*)bty;
	skb->destructor = bty_wfree;
	atomic_add(skb->truesize, &bty->wmem_alloc);
}

void bty_clear_owner_w(struct sk_buff *skb, bty_struct *bty)
{
	skb->sk = NULL;
	skb->destructor = NULL;
	atomic_sub(skb->truesize, &bty->wmem_alloc);
	bty_put(bty);
}

/*
 * Allocate a skb from the socket's send buffer.
 * It's not waiting !!!
 */
struct sk_buff *bty_wmalloc(bty_struct *bty, unsigned long size, int force, int priority)
{
	DBFENTER;

	DBPRT(DBXMIT, "allocated: %ld requested: %ld, wmem_alloc: %d, sndbuf: %d\n", skb_count, size, atomic_read(&bty->wmem_alloc), bty->sndbuf);

	if (size && (force || atomic_read(&bty->wmem_alloc) < bty->sndbuf)) {
		struct sk_buff 	*skb;
		skb = rfcon_alloc_data_skb(bty->con, size, priority);
		if (skb) {
			bty_set_owner_w(skb, bty);
			skb_count++;
			DBFEXIT;
			return skb;
		}
	}
	return NULL;
}


/*
  BTY internal functions
*/
bty_struct *bty_alloc(void)
{
	bty_struct	*bty;

	DBFENTER;

	bty = kmalloc(sizeof(bty_struct), GFP_ATOMIC);
	if (bty == NULL)
		return NULL;
	memset(bty, 0, sizeof(bty_struct));

	bty->magic = TTY_DRIVER_MAGIC;

	init_waitqueue_head(&bty->open_wait);	/* should be waken on ready */
	init_waitqueue_head(&bty->close_wait);
	init_waitqueue_head(&bty->delta_msr_wait);
	init_waitqueue_head(&bty->tx_wait);

	init_timer(&bty->timer);
	bty->timer.data = (unsigned long)bty;
	bty->timer.function = bty_timer;

	tasklet_init(&bty->rx_task, (void*)bty_softint, (unsigned long) bty);

	bty->closing_wait = 10*HZ;
	bty->close_delay = 5*HZ/10;

	/* buffer management */
	bty->sndbuf = BTY_SNDBUF_SIZE;
	
	DBFEXIT;
	return bty;
}

int bty_create(int line, bty_struct **ret_bty)
{
	bty_struct	*bty;

	DBFENTER;

	bty = bty_get(line);
	if (bty) {
		*ret_bty = bty;
		bty_hold(bty);
		return 0;
	}
	
	bty = bty_alloc();
	if (bty == NULL)
		return -ENOMEM;

	bty_table[line].bty = bty;
	atomic_set(&bty->refcnt, 1);
	bty->line = line;

	*ret_bty = bty;
	
	DBFEXIT;
	return 0;
}

void bty_destroy(bty_struct *bty)
{
	DBFENTER;

	DBPRT(DBCTRL, "Trying to destory bty object=%d\n",
			atomic_read(&bty->refcnt));
	
	if (atomic_read(&bty->refcnt) == 0) {
		bty_table[bty->line].bty = NULL;
		kfree(bty);
	}

	DBFEXIT;
}

/*
  For Remote Port
*/
void bty_change_param(bty_struct *bty)
{
	unsigned int	cflag;
	port_param	param;

	DBFENTER;

	memset(&param, 0, sizeof(param));

	cflag = bty->tty->termios->c_cflag;

	/* 
	 * byte size and parity
	 */
	switch (cflag & CSIZE) {
	
	case CS5:
		param.bit_rate = RFCOMM_CS5;
		break;
	case CS6:
		param.bit_rate = RFCOMM_CS6;
		break;
	case CS7:
		param.bit_rate = RFCOMM_CS7;
		break;
	case CS8:
		param.bit_rate = RFCOMM_CS8;
		break;
	default:
		param.bit_rate = RFCOMM_CS8;
	}

	if (cflag & CSTOPB)
		param.stop_bit = 1;

	if (cflag & PARENB)
		param.parity = 1;

	if (!(cflag & PARODD))
		param.parity_type = RFCOMM_PAREVEN;

	if (cflag & CRTSCTS) {
		bty->flags |= ASYNC_CTS_FLOW;
		param.rtr_input = 1;
		param.rtr_output = 1;
	} else {
		bty->flags &= ~ASYNC_CTS_FLOW;
		param.rtr_input = 0;
		param.rtr_output = 0;
	}

	if (cflag & CLOCAL)
		bty->flags &= ~ASYNC_CHECK_CD;
	else
		bty->flags |= ASYNC_CHECK_CD;
	
	/*
	 * Set up parity check flag
	 */

	bty->read_status_mask = BTY_LSR_OE;
	if (I_INPCK(bty->tty))
		bty->read_status_mask |= BTY_LSR_FE | BTY_LSR_PE;
	if (I_BRKINT(bty->tty) || I_PARMRK(bty->tty))
		bty->read_status_mask |= BTY_LSR_BI;
	
	/*
	 * Characters to ignore
	 */
	bty->ignore_status_mask = 0;
	if (I_IGNPAR(bty->tty))
		bty->ignore_status_mask |= BTY_LSR_PE | BTY_LSR_FE;

	if (I_IGNBRK(bty->tty)) {
		bty->ignore_status_mask |= BTY_LSR_BI;
		/*
		 * If we're ignore parity and break indicators, ignore 
		 * overruns too.  (For real raw supcon).
		 */
		if (I_IGNPAR(bty->tty)) 
			bty->ignore_status_mask |= BTY_LSR_OE;
	}

	rfcon_set_param(bty->con, &param);

	DBFEXIT;
}

static inline int bty_init_buffer(bty_struct *bty)
{
	__u32	mtu = rfcon_getmtu(bty->con);

	bty->txbuff = bty_wmalloc(bty, mtu, 1, GFP_ATOMIC);
	if (!bty->txbuff)
		return -ENOMEM;

	return 0;
}

int bty_startup(bty_struct *bty)
{
	int		err;
	rfcomm_con	*con;
	
	DBFENTER;

	if (bty->flags & ASYNC_INITIALIZED)
		return 0;

	con = bty_getcon(bty->line);
	if (!con)
		return -ENODEV;

	/* take receive_queue from sk. con->priv is valid */
	bty->rx_queue = &((struct sock*)con->priv)->receive_queue;
	rfcon_hold(con);
	bty->con = con;

	err = bty_init_buffer(bty);
	if (err)
		return err;

	/* MCR & MSR */
	bty->msr = rfcon_get_msr(bty->con);
	bty->mcr = rfcon_get_mcr(bty->con);
	bty->mcr |= (BTY_MCR_DTR | BTY_MCR_RTS | BTY_MCR_DCD);
	
	rfcon_set_mcr(bty->con, bty->mcr);
	rfcon_set_flow(bty->con, FLOW_START);

	if (bty->tty)
		clear_bit(TTY_IO_ERROR, &bty->tty->flags);

	bty->flags |= ASYNC_INITIALIZED;

	bty_change_param(bty);

	DBFEXIT;
	return 0;
}

void bty_shutdown(bty_struct *bty)
{
	unsigned long	flags;

	DBFENTER;

	if (!(bty->flags & ASYNC_INITIALIZED))
		return;

	bty->flags &= ~ASYNC_INITIALIZED;

	/* wake up ioctl.. */
	wake_up_interruptible(&bty->delta_msr_wait);

	/* clear DTR and RTS */
	if (!bty->tty || (bty->tty->termios->c_cflag & HUPCL))
 		bty->mcr &= ~(BTY_MCR_DTR|BTY_MCR_RTS|BTY_MCR_DCD);

	if (bty->con) {
		rfcon_set_mcr(bty->con, bty->mcr);
		rfcon_set_flow(bty->con, FLOW_STOP);
		rfcon_put(xchg(&bty->con, NULL));
	}

	save_flags(flags); cli(); /* Disable interrupts */

	if (bty->tty)
		set_bit(TTY_IO_ERROR, &bty->tty->flags);
	
	bty_stop_timer(bty);

	if (bty->txbuff) {
		skb_trim(bty->txbuff, 0);
		bty_clear_owner_w(bty->txbuff, bty);
		kfree_skb(bty->txbuff);
		skb_count--;
		bty->txbuff = NULL;
	}

	restore_flags(flags);

	DBFEXIT;
}

int bty_write_to_tty(bty_struct *bty)
{
	rfcomm_con		*con = bty->con;
	struct tty_struct	*tty = bty->tty;
	struct sk_buff		*skb;
	int			room, flag;
	int			status;

	DBFENTER;

	if (con == NULL || tty == NULL)
		return -ENODEV;

	
	/* XXX: check if rx_disabled ?? */

	DBPRT(DBRECV, "Queue length: %d\n", skb_queue_len(bty->rx_queue));
	
	skb = skb_dequeue(bty->rx_queue);
	if (skb == NULL) {
		BTERRORF("Queue is empty - should not happen\n");
		return 0;
	}

	DBPRTPKT(DBRECV, skb->data, skb->len);
	DBDUMPCHAR(DBRECV, skb->data, skb->len);

	status = rfcon_get_line_status(con) & bty->read_status_mask;

	if (status & bty->ignore_status_mask) {
		DBPRT(DBRECV, "some error:ignore characters.\n");
		dev_kfree_skb(skb);
		return 0;
	}

	room = btmin(TTY_FLIPBUF_SIZE-tty->flip.count, skb->len);

	if (con->peer_break_signal.b1) {
		con->peer_break_signal.b1 = 0;
		DBPRT(DBRECV, "handling break....\n");
		
		flag = TTY_BREAK;
		if (bty->flags & ASYNC_SAK)
			do_SAK(tty);

	} else if (status & BTY_LSR_PE) flag = TTY_PARITY;
	else if (status & BTY_LSR_FE) flag = TTY_FRAME;
	else if (status & BTY_LSR_OE) flag = TTY_OVERRUN;
	else flag = TTY_NORMAL;

	DBPRT(DBRECV, "RECV, byte: %#02x, room: %d, flag: %#02x\n", *skb->data, room, flag);

	if (room) {
		memset(tty->flip.flag_buf_ptr, flag, room);
		memcpy(tty->flip.char_buf_ptr, skb->data, room);
		tty->flip.char_buf_ptr += room;
		tty->flip.flag_buf_ptr += room;
		tty->flip.count += room;
		skb_pull(skb, room);
	}

	if (skb->len == 0)
		dev_kfree_skb(skb);
	else
		skb_queue_head(bty->rx_queue, skb);
	
	if (room) 
		tty_flip_buffer_push(tty);

	/* perform flow control */

	if (skb_queue_empty(bty->rx_queue) && rfcon_disconnect_pend(con)){
		/* disconnect */
		tty_hangup(tty);
	}

	DBFEXIT;
	return room;
}

void bty_timer(unsigned long data)
{
	bty_struct		*bty = (bty_struct*)data;
	struct tty_struct	*tty = bty->tty;
	
	DBFENTER;

	if (!(bty->flags & ASYNC_INITIALIZED) || tty == NULL)
		return;

	while ((TTY_FLIPBUF_SIZE-tty->flip.count) && !skb_queue_empty(bty->rx_queue))
		bty_write_to_tty(bty);

	if (!skb_queue_empty(bty->rx_queue))
		bty_start_timer(bty, 1);

	DBFEXIT;
}

/*
  recive procedure
  called form the event.. no races with timer
*/
int bty_data_ind(rfcomm_con *con)
{
	bty_struct	*bty = bty_get(con->line);

	DBFENTER;

	if (bty == NULL)
		return -ENODEV;

	bty_stop_timer(bty);

	if (!(bty->flags & ASYNC_INITIALIZED))
		return -ENODEV;
	
	bty_write_to_tty(bty);
	if (!skb_queue_empty(bty->rx_queue))
		bty_start_timer(bty, 1);

	DBFEXIT;
	return 0;
}

int bty_send_data(bty_struct *bty)
{
	rfcomm_con	*con = bty->con;
	struct sk_buff	*skb = bty->txbuff;

	DBFENTER;

	if (skb == NULL || skb->len == 0)	/* nothing to send */
		return 0;
	
	/*  bty_send_data can be reentered after rfcomm_send_data so reset
	 *  txbuff here.
	 */
	bty->txbuff = NULL;
	rfcomm_send_data(con, skb);
	/*
	  FIXME: may be initialize before rfcomm_send_data
	*/
	bty_init_buffer(bty);

	DBFEXIT;
	return 0;
}

void bty_control_ind(rfcomm_con *con, int event)
{
	bty_struct		*bty = bty_get(con->line);
	struct tty_struct	*tty;
	__u8			delta_msr;

	DBFENTER;

	if (bty == NULL)
		return;

	tty = bty->tty;

	if (!(bty->flags & ASYNC_INITIALIZED) || tty == NULL)
		return;

	switch (event) {
	case TX_START:
		tty->hw_stopped = rfcon_disabled(con);
		if (tty->hw_stopped)
			return;
		bty_send_data(bty);
		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
			tty->ldisc.write_wakeup(tty);
		wake_up_interruptible(&tty->write_wait);
		break;

	case TX_STOP:
		tty->hw_stopped = 1;
		break;

	case LINE_STATE:
		delta_msr = bty->msr;
		bty->msr = rfcon_get_msr(bty->con);
		delta_msr ^= bty->msr;

		if (delta_msr) {
			if(bty->msr & BTY_MSR_CTS) bty->icount.cts++;
			if(bty->msr & BTY_MSR_DSR) bty->icount.dsr++;
			if(bty->msr & BTY_MSR_RI)  bty->icount.rng++;
			if(bty->msr & BTY_MSR_DCD) bty->icount.dcd++;
			wake_up_interruptible(&bty->delta_msr_wait);
		}

		if ((bty->flags & ASYNC_CHECK_CD) && (delta_msr & BTY_MSR_DCD)) {
			DBPRT(DBCTRL ,"CD now %s...\n", (bty->msr & BTY_MSR_DCD) ? "on" : "off");
			if (bty->msr & BTY_MSR_DCD) {
				/* DCD raised! */
				wake_up_interruptible(&bty->open_wait);
			} else {
				/* DCD falled */
				DBPRT(DBCTRL ,"hangup..\n");
				tty_hangup(tty);
			}
		}
	
		/* like ASYNC_CTS_FLOW */
		if (bty->flags & ASYNC_CTS_FLOW) {
			if (tty->hw_stopped) {
				if (bty->msr & BTY_MSR_CTS) {
					DBPRT(DBCTRL ,"CTS tx start...\n");
					tty->hw_stopped = rfcon_disabled(con);
					if (tty->hw_stopped == 0) {
						bty_send_data(bty);
						if( (tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) 
						    && tty->ldisc.write_wakeup )
							tty->ldisc.write_wakeup(tty);
						wake_up_interruptible(&tty->write_wait);
					}
					return;
				}
			} else {
				if (!(bty->msr & BTY_MSR_CTS)) {
					DBPRT(DBCTRL ,"CTS tx stop...");
					tty->hw_stopped = 1;
				}
			}
		}
		break;
	}

	DBFEXIT;
}

void bty_softint(void *data)
{
	bty_struct		*bty = data;
	struct tty_struct	*tty = bty->tty;

	DBFENTER;

	if (!tty)
		return;

	if (test_and_clear_bit(BTY_EVENT_WRITE_WAKEUP, &bty->event)) {
		if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
			tty->ldisc.write_wakeup(tty);
		wake_up_interruptible(&tty->write_wait);
	}
	
	DBFEXIT;
}

/*
  it seems it is called from BH
*/
int bty_register_con(rfcomm_con *con, int line)
{
	DBFENTER;

	DBPRT(DBINIT, "Register line: %d\n", line);

	if( line == -1) {
		for (line = 0; line < bty_maxdev; line++)
			if (bty_table[line].con == NULL) {
				break;
			}
		if (line >= bty_maxdev)
			return -EBUSY;
	} else {
		if (line < 0 || line >= bty_maxdev)
			return -EINVAL;
		else if (bty_table[line].con)
			return -EBUSY;
	}

	bty_table[line].con = con;
	con->line = line;
	rfcon_hold(con);	/* hold it */
	
	DBPRT(DBINIT, "Line registered: %d\n", line);

	MOD_INC_USE_COUNT;
	DBFEXIT;
	return line;
}

int bty_unregister_con(int line)
{
	rfcomm_con	*con;
	bty_struct	*bty;

	DBFENTER;

	con = bty_getcon(line);
	if (con == NULL)
		return -ENODEV;

	bty_table[line].con = NULL;
	con->line = -1;

	bty = bty_get(line);
	if (bty && bty->tty)
		tty_hangup(bty->tty);

	rfcon_put(con);

	MOD_DEC_USE_COUNT;
	DBFEXIT;
	return 0;
}


/*********************   CALLBACKS   *******************/

/*
  open and friends
*/
int bty_open(struct tty_struct * tty, struct file * filp)
{
	bty_struct	*bty;
	__u32 		line;
	int		err;
  
	DBFENTER;

	MOD_INC_USE_COUNT;

	line = get_line(tty);
	if (line >= bty_maxdev) {
		MOD_DEC_USE_COUNT;
		return -ENODEV;
	}

	err = bty_create(line, &bty);
	if (err) {
		MOD_DEC_USE_COUNT;
		BTERROR("bty structure create err: %d\n", err);
		return err;
	}
	bty->count++;
	DBPRT(DBINIT, "%s[%d] opened, count %d\n", tty->driver.name, line, bty->count);

	/* set pointer to driver structures */
	DBPRT(DBINIT, "tty: %p, data: %p\n", tty, bty->tty);
	tty->driver_data = bty;
	bty->tty = tty;
	tty->low_latency = (bty->flags & ASYNC_LOW_LATENCY) ? 1 : 0;

	if (tty_hung_up_p(filp) || (bty->flags & ASYNC_CLOSING)) {
		DBPRT(DBINIT, "Hang up...\n");
		if (bty->flags & ASYNC_CLOSING)
			interruptible_sleep_on(&bty->close_wait);
#ifdef DO_RESTART
		if (!(bty->flags & ASYNC_HUP_NOTIFY))
			return -ERESTARTSYS;
#endif
		return -EAGAIN;
	}

	err = bty_startup(bty);
	if (err) {
		return err;
	}

	// block till ready XXX

	bty->session = current->session;
	bty->pgrp = current->pgrp;
	tty->hw_stopped = rfcon_disabled(bty->con);

	DBFEXIT;
	return 0;
}


/*
 * ----------------------------------------------------------------------
 * bty_close() and friends
 *
 * most of this function is stolen from serial.c
 * ----------------------------------------------------------------------
 */

void bty_close(struct tty_struct * tty, struct file * filp)
{
	bty_struct 	*bty = (bty_struct*)tty->driver_data;
	unsigned long	flags;

	DBFENTER;
	
	DBPRT(DBINIT, "tty: %p, data: %p\n", tty, tty->driver_data);

	if (!bty) {
		BTERROR("bty == NULL\n");
		return;
	}

	save_flags(flags);cli();

	if (--bty->count < 0) {
		BTERROR("driver count is negative: %d\n", bty->count);
		bty->count = 0;
		restore_flags(flags);
		return;
	}		

	if (tty_hung_up_p(filp)) {
		/*
		 * upper tty layer caught a HUP signal and called bty_hangup()
		 * before. so we do nothing here.
		 */
		DBPRT(DBINIT, "Hang up...\n");
		goto exit;
	}

	if (bty->count) { 	/* do nothing */
		DBPRT(DBINIT, "driver still in use: %d\n", bty->count);
		goto exit;
	}

	bty->flags |= ASYNC_CLOSING;

	/*
	 * Now we wait for the transmit buffer to clear; and we notify 
	 * the line discipline to only process XON/XOFF characters.
	 */
	tty->closing = 1;
	if (bty->closing_wait != ASYNC_CLOSING_WAIT_NONE){
		DBPRT(DBINIT, "calling tty_wait_until_sent()\n");
		tty_wait_until_sent(tty, bty->closing_wait);
	}

	bty_shutdown(bty);

	if (tty->ldisc.flush_buffer)
		tty->ldisc.flush_buffer(tty);
 	if (tty->driver.flush_buffer) 
 		tty->driver.flush_buffer(tty);  

	tty->closing = 0;
	bty->tty = NULL;

	if (bty->blocked_open) {/* used to block till ready */
		if (bty->close_delay) {
			/* kill time */
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(bty->close_delay);
		}
		wake_up_interruptible(&bty->open_wait);
	}

	bty->flags &= ~ASYNC_CLOSING;
        wake_up_interruptible(&bty->close_wait); 
	
 exit:
	bty_put(bty);
	MOD_DEC_USE_COUNT;
        restore_flags(flags);
	DBFEXIT;
}

/*
 * ------------------------------------------------------------
 * bty_hangup()
 * This routine notifies that tty layer have got HUP signal
 * ------------------------------------------------------------
 */

void bty_hangup(struct tty_struct *tty)
{
	bty_struct	*bty = tty->driver_data;

	DBFENTER;

	bty_flush_buffer(tty);
	if (bty->flags & ASYNC_CLOSING)
		return;
	bty_shutdown(bty);
	bty->tty = NULL;
	wake_up_interruptible(&bty->open_wait);

	DBPRT(DBCTRL, "bty->count: %d, refcnt: %d\n", bty->count, atomic_read(&bty->refcnt));

	DBFEXIT;
}



/*
 * ----------------------------------------------------------------------
 * bty_write() and friends
 * This routine will be called when something data are passed from
 * kernel or user.
 * ----------------------------------------------------------------------
 */

int bty_write(struct tty_struct * tty, int from_user, const unsigned char *buf, int count)
{
	bty_struct	*bty = (bty_struct*)tty->driver_data;
	rfcomm_con	*con;
	struct sk_buff	*skb, *oskb = NULL;
	int		room, wrote = 0;
	__u8		*data;

	DBFENTER;
	
	DBPRT(DBXMIT, "Transmit: %d bytes\n", count);
	DBPRTPKT(DBXMIT, (void*)buf, count);
	DBDUMPCHAR(DBXMIT, buf, count);

	if (tty->hw_stopped || count == 0)
		return 0;

	if (bty == 0) {
		DBPRT(DBXMIT, "Driver structure is not exists yet\n");
		return -ENODEV;
	}

	con = bty->con;
	if (con == NULL || STATE(con) != CON_OPEN) {
		DBPRT(DBXMITE, "Port is not connected\n");
		return 0;
	}

	skb = bty->txbuff;
	if (skb != NULL) {
		bty->txbuff = NULL;
		if (skb->len)
			rfcomm_send_data(bty->con, skb);
		else {
			oskb = skb;
			bty_clear_owner_w(oskb, bty);
		}
	}

	while (count) {
		int mtu = rfcon_getmtu(bty->con);
		
		if (tty->hw_stopped)
			break;

		room = btmin(count, mtu);
		skb = bty_wmalloc(bty, room, 0, GFP_ATOMIC);
		if (skb == NULL)
			break;

		DBPRT(DBXMIT, "room: %d\n", room);
		data = skb_put(skb, room);
		if (from_user)
			copy_from_user(data, buf, room);
		else
			memcpy(data, buf, room);

		if (rfcomm_send_data(bty->con, skb) < 0)
		{
			DBPRT(DBXMITE, "!!!!rfcomm_send_failed!!!!\n");
		}
		
		buf += room;
		wrote += room;
		count -= room;
	}

	if (oskb) {
		bty->txbuff = oskb;
		bty_set_owner_w(oskb, bty);
	} else
		bty_init_buffer(bty);

	if (count == 0) {
		set_bit(BTY_EVENT_WRITE_WAKEUP, &bty->event);
		tasklet_schedule(&bty->rx_task);
	}

	DBPRT(DBXMIT, "wrote: %d\n", wrote);
	DBFEXIT;
	return wrote;
}

/*
 * Function bty_put_char (tty, ch)
 *
 *    This routine is called by the kernel to pass a single character.
 *    If we exausted our buffer,we can ignore the character!
 *
 */
void bty_put_char(struct tty_struct *tty, unsigned char ch)
{
	bty_struct	*bty = tty->driver_data;
	struct sk_buff	*skb = bty->txbuff;
	__u8		*ptr;

	DBFENTER;

	if (skb == NULL)
		return;

	if (rfcon_skb_tailroom(skb) == 0) {
		if (tty->hw_stopped)
			return;
		bty_send_data(bty);
	}

	skb = bty->txbuff;
	if (skb == NULL)
		return;

	ptr = skb_put(skb, 1);
	*ptr = ch;

	DBFEXIT;
}


void bty_flush_chars(struct tty_struct *tty)
{
	bty_struct	*bty = tty->driver_data;

	DBFENTER;

	/*
	 * FIXME: What is necessary here?? 
	 * What should be done if hw_stopped ??
	 */
	if (tty->hw_stopped)
		return;
	bty_send_data(bty);

	DBFEXIT;
}


/*
 * Function bty_write_room (tty)
 *
 *    This routine returns the room that our buffer has now.
 *
 */
int bty_write_room(struct tty_struct *tty)
{
	bty_struct	*bty = tty->driver_data;
	struct sk_buff	*skb = bty->txbuff;
	int		ret;

	DBFENTER;

	if (skb == NULL)
		return 0;

	ret = rfcon_skb_tailroom(skb);

	DBFEXIT;
	return ret;
}


/*
 * Function bty_chars_in_buffer (tty)
 *
 *    This function returns how many characters which have not been sent yet 
 *    are still in buffer.
 *
 */
int bty_chars_in_buffer(struct tty_struct *tty)
{
	bty_struct	*bty = tty->driver_data;
	struct sk_buff	*skb = bty->txbuff;
	int		len;

	//DBFENTER;

	if (skb == NULL)
		return 0;

	len = skb->len;

	//DBFEXIT;
	return len;
}

void bty_flush_buffer(struct tty_struct *tty)
{
	bty_struct	*bty = tty->driver_data;

	DBFENTER;

	if (bty->txbuff)
		skb_trim(bty->txbuff, 0);	/* XXX */

	wake_up_interruptible(&tty->write_wait);

	if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
		tty->ldisc.write_wakeup(tty);

	DBFEXIT;
}

int get_modem_info(bty_struct *bty, unsigned int *value)
{
	unsigned int 	result;

	result =  ((bty->mcr & BTY_MCR_RTS) ? TIOCM_RTS : 0)
		| ((bty->mcr & BTY_MCR_DTR) ? TIOCM_DTR : 0)

		| ((bty->msr & BTY_MSR_DCD) ? TIOCM_CAR : 0)
		| ((bty->msr & BTY_MSR_RI) ? TIOCM_RNG : 0)
		| ((bty->msr & BTY_MSR_DSR) ? TIOCM_DSR : 0)
		| ((bty->msr & BTY_MSR_CTS) ? TIOCM_CTS : 0);

	return put_user(result, value);
}

int set_modem_info(bty_struct *bty,  unsigned int cmd, unsigned int *value)
{ 
	rfcomm_con 	*con = bty->con;
	int		err;
	unsigned int 	arg;

	DBFENTER;

	err = get_user(arg, value);
	if(err)
		return err;

	switch (cmd) {
	case TIOCMBIS: 
		if (arg & TIOCM_RTS) 
			bty->mcr |= BTY_MCR_RTS;
		if (arg & TIOCM_DTR)
			bty->mcr |= BTY_MCR_DTR;
		if (arg & TIOCM_CAR)
			bty->mcr |= BTY_MCR_DCD;
		if (arg & TIOCM_RNG)
			bty->mcr |= BTY_MCR_RI;
		break;
		
	case TIOCMBIC:
		if (arg & TIOCM_RTS)
			bty->mcr &= ~BTY_MCR_RTS;
		if (arg & TIOCM_DTR)
 			bty->mcr &= ~BTY_MCR_DTR;
		if (arg & TIOCM_CAR)
 			bty->mcr &= ~BTY_MCR_DCD;
		if (arg & TIOCM_RNG)
 			bty->mcr &= ~BTY_MCR_RI;
 		break;
		
	case TIOCMSET:
 		bty->mcr = ((bty->mcr & ~(BTY_MCR_RTS | BTY_MCR_DTR | BTY_MCR_DCD | BTY_MCR_RI))
				| ((arg & TIOCM_RTS) ? BTY_MCR_RTS : 0)
				| ((arg & TIOCM_DTR) ? BTY_MCR_DTR : 0)
				| ((arg & TIOCM_CAR) ? BTY_MCR_DCD : 0)
				| ((arg & TIOCM_RNG) ? BTY_MCR_RI : 0));
		break;
		
	default:
		return -EINVAL;
	}
	

	rfcon_set_mcr(con, bty->mcr);

	DBFEXIT;
	return 0;
}

/*
 * ----------------------------------------------------------------------
 * bty_ioctl() and friends
 * This routine allows us to implement device-specific ioctl's.
 * If passed ioctl number (i.e.cmd) is unknown one, we should return 
 * ENOIOCTLCMD.
 *
 * TODO: we can't use setserial on tty because some ioctls are not implemented.
 * we should add some ioctls and make some tool which is resemble to setserial.
 * ----------------------------------------------------------------------
 */

int bty_ioctl(struct tty_struct *tty, struct file * file, unsigned int cmd, unsigned long arg)
{
	bty_struct	*bty = tty->driver_data;
	rfcomm_con	*con;
	unsigned long	flags;
	int 		err = 0;

	struct serial_icounter_struct	cnow,cprev;
	struct serial_icounter_struct	*p_cuser;	/* user space */


	//DBFENTER;
  
	con = bty->con;

	switch (cmd) {

	case TIOCMGET:
		return get_modem_info(bty, (unsigned int *) arg);

	case TIOCMBIS:
	case TIOCMBIC:
	case TIOCMSET:
		return set_modem_info(bty, cmd, (unsigned int *) arg);

	case TIOCMIWAIT:
		save_flags(flags); cli();
		/* note the counters on entry */
		cprev = bty->icount;
		restore_flags(flags);
		while (1) {
			interruptible_sleep_on(&bty->delta_msr_wait);
				/* see if a signal did it */
			if (signal_pending(current))
				return -ERESTARTSYS;
			save_flags(flags); cli();
			cnow = bty->icount; /* atomic copy */
			restore_flags(flags);
			if( cnow.rng == cprev.rng && cnow.dsr == cprev.dsr && 
			    cnow.dcd == cprev.dcd && cnow.cts == cprev.cts )
				return -EIO; /* no change => error */
			if( ((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
			    ((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
			    ((arg & TIOCM_CD)  && (cnow.dcd != cprev.dcd)) ||
			    ((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)) ) {
				return 0;
			}
			cprev = cnow;
		}
		/* NOTREACHED */
		
	case TIOCGICOUNT:
		save_flags(flags); cli();
		cnow = bty->icount;
		restore_flags(flags);
		p_cuser = (struct serial_icounter_struct *) arg;
		err = put_user(cnow.cts, &p_cuser->cts);
		if (err) return err;
		err = put_user(cnow.dsr, &p_cuser->dsr);
		if (err) return err;
		err = put_user(cnow.rng, &p_cuser->rng);
		if (err) return err;
		err = put_user(cnow.dcd, &p_cuser->dcd);
		if (err) return err;
		err = put_user(cnow.rx, &p_cuser->rx);
		if (err) return err;
		err = put_user(cnow.tx, &p_cuser->tx);
		if (err) return err;
		err = put_user(cnow.frame, &p_cuser->frame);
		if (err) return err;
		err = put_user(cnow.overrun, &p_cuser->overrun);
		if (err) return err;
		err = put_user(cnow.parity, &p_cuser->parity);
		if (err) return err;
		err = put_user(cnow.brk, &p_cuser->brk);
		if (err) return err;
		err = put_user(cnow.buf_overrun, &p_cuser->buf_overrun);
		if (err) return err;			

		return 0;
		
		
		/* ioctls which are imcompatible with serial.c */

	case TIOCGSERIAL:
		DBPRT(DBCTRL, "TIOCGSERIAL is not supported\n");
		return -ENOIOCTLCMD;  
		//return get_serial_info(driver, (struct serial_struct *) arg);

	case TIOCSSERIAL:
		DBPRT(DBCTRL, "TIOCSSERIAL is not supported\n");
		return -ENOIOCTLCMD;  
		//return set_serial_info(driver, (struct serial_struct *) arg);

	case TIOCSERGSTRUCT:
		DBPRT(DBCTRL, "TIOCSERGSTRUCT is not supported\n");
		return -ENOIOCTLCMD;  
	case TIOCSERGETLSR:
		DBPRT(DBCTRL, "TIOCSERGETLSR is not supported\n");
		return -ENOIOCTLCMD;  
 	case TIOCSERCONFIG:
		DBPRT(DBCTRL, "TIOCSERCONFIG is not supported\n");
		return -ENOIOCTLCMD;  

	default:
		//DBPRT(DBCTRL, "Command ignored: %#04x\n", cmd);
		return -ENOIOCTLCMD;  /* ioctls which we must ignore */
	}

	//DBFEXIT;
	return err;
}

/*
 * ----------------------------------------------------------------------
 * bty_throttle,bty_unthrottle
 *   These routines will be called when we have to pause sending up data to tty.
 *   We use RTS virtual signal when servicetype is NINE_WIRE
 * ----------------------------------------------------------------------
 */

void bty_throttle(struct tty_struct *tty)
{
	bty_struct	*bty = tty->driver_data;
	rfcomm_con	*con = bty->con;

	DBFENTER;

	if (I_IXOFF(tty))
		bty_send_xchar(tty, STOP_CHAR(tty));

	if (tty->termios->c_cflag & CRTSCTS) {
		bty->mcr &= ~BTY_MCR_RTS; 
		rfcon_set_mcr(con, bty->mcr);
	}

	DBFEXIT;
}


void bty_unthrottle(struct tty_struct *tty)
{
	bty_struct	*bty = tty->driver_data;
	rfcomm_con	*con = bty->con;

	DBFENTER;

	if (I_IXOFF(tty))
		bty_send_xchar(tty, START_CHAR(tty));

	if (tty->termios->c_cflag & CRTSCTS) {
		bty->mcr |= BTY_MCR_RTS;
		rfcon_set_mcr(con, bty->mcr);
	}

	DBFEXIT;
}


/*
 * ----------------------------------------------------------------------
 * bty_set_termios()
 * This is called when termios is changed.
 * If things that changed is significant for us,(i.e. changing baud rate etc.)
 * send something to peer device.
 * ----------------------------------------------------------------------
 */

void bty_set_termios(struct tty_struct *tty, struct termios * old_termios)
{
	bty_struct	*bty = tty->driver_data;

	DBFENTER;

	if( (tty->termios->c_cflag == old_termios->c_cflag) &&
	   (RELEVANT_IFLAG(tty->termios->c_iflag) == RELEVANT_IFLAG(old_termios->c_iflag)) )
		return;

	bty_change_param(bty);

	/* handle turning off CRTSCTS */
	if ((old_termios->c_cflag & CRTSCTS) && !(tty->termios->c_cflag & CRTSCTS)) {

		tty->hw_stopped = rfcon_disabled(bty->con);
		/* bty_start(tty); */       /* we don't need this */
	}


	DBFEXIT;
}

/*
 * ------------------------------------------------------------
 * bty_stop() and bty_start()
 *
 * This routines are called before setting or resetting tty->stopped.
 * They enable or disable an interrupt which means "transmitter-is-ready"
 * in serial.c, but  I think these routine are not necessary for us. 
 * ------------------------------------------------------------
 */

#if 0
void bty_stop(struct tty_struct *tty)
{
	bty_struct *bty = (bty_struct *)tty->driver_data; 

	DBFENTER;

	DBFEXIT;
}

void bty_start(struct tty_struct *tty)
{
	bty_struct *bty = (bty_struct *)tty->driver_data;

	DBFENTER;

	DBFEXIT;
}
#endif


void bty_send_xchar(struct tty_struct *tty, char ch)
{
	DBFENTER;
	bty_put_char(tty, ch);
	DBFEXIT;
}

/*
 * Function bty_break (tty, break_state)
 *
 *    Routine which turns the break handling on or off
 *
 */
void bty_break(struct tty_struct *tty, int break_state)
{
	bty_struct	*bty = tty->driver_data;
	rfcomm_con	*con = bty->con;
	unsigned long	flags;

	DBFENTER;

	save_flags(flags);cli();
	if (break_state == -1)
		rfcon_set_break(con, 1);
	else
		rfcon_set_break(con, 0);

	restore_flags(flags);

	DBFEXIT;
}


int bty_proc_read(char *buf, char **start, off_t offset, int len);

int bty_read_proc(char *buf, char **start, off_t offset, int len, int *eof, void *unused)
{
	int count = 0;

	DBFENTER;

	count += sprintf(buf, "Revision Date: %s\n", revision_date);

	DBFEXIT;
	return count+bty_proc_read(buf+count, start, offset, len);
}


/*
 * Function bty_wait_until_sent (tty, timeout)
 *
 *    wait until Tx queue of BTY is empty 
 *
 */
void bty_wait_until_sent(struct tty_struct *tty, int timeout)
{
	bty_struct	*bty = tty->driver_data;
	unsigned long	orig_jiffies;

	DBFENTER;

	if (!tty->closing || !bty->con)
		return;   /* nothing to do */

	/* 
	 * at disconnection, we should wait until Tx queue of RFCOMM is
	 * flushed
	 */

	orig_jiffies = jiffies;

	while (rfcon_disconnect_pend(bty->con)) {

		DBPRT(DBCTRL, "wait..\n");
		current->state = TASK_INTERRUPTIBLE;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
		current->counter = 0;	/* make us low-priority */
#else
#endif
		schedule_timeout(HZ);    /* 1sec */
		if (signal_pending(current))
			break;
		if (timeout && time_after(jiffies, orig_jiffies + timeout))
			break;
	}
	current->state = TASK_RUNNING;

	DBFEXIT;
}

#if defined(CONFIG_DEVFS_FS)

#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
static inline _syscall2(int,chmod,char*,name,mode_t,mode);
static int	errno;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 17)
#define devfs_get_handle	devfs_find_handle
#define devfs_put(a)
#endif

#define DEVPATHLEN	256

void bty_change_mode(struct tty_driver *driver)
{
	char		path[DEVPATHLEN];
	char		*name;
	devfs_handle_t	handle;
	umode_t 	mode = S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
	int		i, pos;
	mm_segment_t	old_fs;

	for(i = 0; i < driver->num; i++) {
		handle = devfs_get_handle (NULL, NULL, driver->major, driver->minor_start + i,
				DEVFS_SPECIAL_CHR, 0);
		if (handle == NULL) {
			BTERRORF("unable to get device handle\n");
			break;
		}
		pos = devfs_generate_path(handle, path, DEVPATHLEN);
		devfs_put(handle);
		if (pos < 0) {
			BTERRORF("unable to generate a path\n");
			break;
		}
		/* add /dev/ */
		name = path + pos - 5;
		memcpy(name, "/dev/", 5);

		old_fs = get_fs(); set_fs(KERNEL_DS);
		chmod(name, mode);
		set_fs(old_fs);
	}
}

#endif

/*
 * Function bty_register_ttydriver(void)
 *
 *   we register "port emulation entity"(see Bty specification) here
 *   as a tty device.
 */

int bty_register_ttydriver(void)
{
	int	err;
	DBFENTER;
  

	/* allocate memory for data structures */
	err = -ENOMEM;
	tty_table = (struct tty_struct**)kmalloc(sizeof(void*)*bty_maxdev, GFP_KERNEL);
	if (tty_table == NULL) {
		BTERROR("Unable to allocate memory for data structures\n");
		goto err1;
	}
	memset(tty_table, 0, sizeof(void*)*bty_maxdev);

	tty_termios = (struct termios**)kmalloc(sizeof(void*)*bty_maxdev, GFP_KERNEL);
	if (tty_termios == NULL) {
		BTERROR("Unable to allocate memory for data structures\n");
		goto err2;
	}
	memset(tty_termios, 0, sizeof(void*)*bty_maxdev);

	tty_termios_locked = (struct termios**)kmalloc(sizeof(void*)*bty_maxdev, GFP_KERNEL);
	if (tty_termios_locked == NULL) {
		BTERROR("Unable to allocate memory for data structures\n");
		goto err3;
	}
	memset(tty_termios_locked, 0, sizeof(void*)*bty_maxdev);

	/* setup virtual serial port device */

	/* Initialize the tty_driver structure ,which is defined in 
	   tty_driver.h */
        
	memset(&bty_drv, 0, sizeof(struct tty_driver));
	bty_drv.magic = TTY_DRIVER_MAGIC;
	bty_drv.driver_name = "bty_drv";
#ifdef CONFIG_DEVFS_FS
	bty_drv.name = "bty%d";
#else
	bty_drv.name = "bty";
#endif
	bty_drv.major = BTY_MAJOR;
	bty_drv.minor_start = BTY_MINOR;
	bty_drv.num = bty_maxdev;
	bty_drv.type = TTY_DRIVER_TYPE_SERIAL;  /* see tty_driver.h */
	bty_drv.subtype = SERIAL_TYPE_NORMAL;  /* see tty_driver.h */


	/*
	 * see drivers/char/tty_io.c and termios(3)
	 */

	bty_drv.init_termios = tty_std_termios;
	bty_drv.init_termios.c_cflag = B115200 | CS8 | CREAD | HUPCL | CLOCAL;
	bty_drv.flags = TTY_DRIVER_REAL_RAW;   /* see tty_driver.h */
	bty_drv.refcount = &tty_refcount;

	/* pointer to the tty data structures */
  
	bty_drv.table = tty_table;  
	bty_drv.termios = tty_termios;
	bty_drv.termios_locked = tty_termios_locked;
  
	/*
	 * Interface table from the kernel(tty driver) to the bty
	 * layer
	 */

	bty_drv.open = bty_open;
	bty_drv.close = bty_close;
	bty_drv.write = bty_write;
	bty_drv.put_char = bty_put_char;
	bty_drv.flush_chars = bty_flush_chars;
	bty_drv.write_room = bty_write_room;
	bty_drv.chars_in_buffer = bty_chars_in_buffer; 
	bty_drv.flush_buffer = bty_flush_buffer;
	bty_drv.ioctl = bty_ioctl; 
	bty_drv.throttle = bty_throttle;
	bty_drv.unthrottle = bty_unthrottle;
	bty_drv.set_termios = bty_set_termios;
	bty_drv.stop = NULL;				/* bty_stop */
	bty_drv.start = NULL;				/* bty_start */
	bty_drv.hangup = bty_hangup;
	bty_drv.send_xchar = bty_send_xchar;
	bty_drv.break_ctl = bty_break;
	bty_drv.read_proc = bty_read_proc;
	bty_drv.wait_until_sent = bty_wait_until_sent;

	if ((err = tty_register_driver(&bty_drv))) {
		goto err4;
	}
#if defined(CONFIG_DEVFS_FS)
	//bty_change_mode(&bty_drv);
#endif
	DBPRT(DBINIT, "done.\n");

	DBFEXIT;

	return 0;
 err4:
	kfree(tty_termios_locked);
 err3:
	kfree(tty_termios);
 err2:
	kfree(tty_table);
 err1:
	return err;
}


/*
 * Function bty_unregister_ttydriver(void) 
 *   it will be called when you rmmod
 */

int bty_unregister_ttydriver(void)
{
	int err;	
	DBFENTER;
  
	/* unregister tty device   */
  
	err = tty_unregister_driver(&bty_drv);

	if (err)
		BTERRORF("BTY: failed to unregister vtd driver(%d)\n", err);

	kfree(tty_termios_locked);
	kfree(tty_termios);
	kfree(tty_table);

	return err;
}
	
/* Proc file system services */

int bty_proc_read(char *buf, char **start, off_t offset, int len)
{
	rfcomm_con	*con;
	bty_struct	*bty;
	int count = 0, i;

	DBFENTER;

	count += sprintf(buf+count, "skb allocated: %ld\n", skb_count);

	for (i=0;i<bty_maxdev;i++) {
		count += sprintf(buf+count, "line: %d\n", i);
		con = bty_getcon(i);
		if (con) {
			count += sprintf(buf+count, "\tMSR: credit: %d, fc: %d, use rtr: %d, rtr: %d, rtc: %d, dv: %d, ic: %d\n", 
					con->remote_credit, con->peer_modem.fc, con->peer_param.rtr_output,
					LV(con->peer_modem.mr&RFCOMM_RTR), LV(con->peer_modem.mr&RFCOMM_RTC),
					LV(con->peer_modem.mr&RFCOMM_DV), LV(con->peer_modem.mr&RFCOMM_IC));
			count += sprintf(buf+count, "\tMCR: credit: %d, fc: %d, use rtr: %d, rtr: %d, rtc: %d, dv: %d, ic: %d\n", 
					con->local_credit, con->modem.fc, con->param.rtr_output,
					LV(con->modem.mr&RFCOMM_RTR), LV(con->modem.mr&RFCOMM_RTC),
					LV(con->modem.mr&RFCOMM_DV), LV(con->modem.mr&RFCOMM_IC));
		}
		bty = bty_get(i);
		if (bty) {
			count += sprintf(buf+count, "\n\tbty count: %d, bty refcnt: %d\n\n", 
					bty->count, atomic_read(&bty->refcnt));
		}
	}

	DBFEXIT;
	return count;
}

struct proc_dir_entry	*bty_proc;

int __init bty_init(void)
{
	int	err = 0;

	DBFENTER;

	err = -ENOMEM;
	bty_table = (struct bty_table*)kmalloc(sizeof(struct bty_table) * bty_maxdev, GFP_KERNEL);
	if (bty_table == NULL) {
		BTERRORF("kmalloc failed!\n");
		goto err1;
	}
	memset(bty_table, 0, sizeof(struct bty_table) * bty_maxdev);

	bty_proc = create_proc_info_entry("bty", 0, proc_affix, bty_proc_read);
	if( bty_proc == 0) {
		BTERROR("Unable to register proc fs entry\n");
		goto err2;
	}

	if ((err = bty_register_ttydriver())) {
		BTERRORF("Error in rdvtd_register_device\n");
		goto err3;
	}


	return 0;
 err3:
	remove_proc_entry("bty", proc_affix);
 err2:
	kfree(bty_table);
 err1:
	return err;
}

void bty_exit(void)
{
	DBFENTER;

	bty_unregister_ttydriver();
	remove_proc_entry("bty", proc_affix);

	if(bty_table) {
		kfree(bty_table);
		bty_table = NULL;
	}

}

